;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

(defparameter *sys* nil) ;; bind me early to an instance of the following

(defparameter *first-kill-all-the-windows* nil)

(defun cello-reset (&optional (systemType 'MGSystem))
  (let ((*first-kill-all-the-windows* t))
    (sleep 1)) ;; get open windows get events

  (cell-reset)
  (ogb-glut-init
   (lambda ()
     (when (and *sys* (kids *sys*))
       (error "prior windows extant. close now, or evaluate: (setf (kids *sys*) nil)"))))
  #+celloglut (ftgl-init)
  #+celloglut (imagick-init)

  (setf ogb::*gl-begun* nil)
;;;  (styles-default-clear)
  (when systemType
    (setf *sys* (to-be (make-instance systemType :md-name 'mgsys))))
  (values)
  )

(defmodel MGSystem (Family)
   (
    (mainWindow :initarg :mainWindow :initform (cv nil) :accessor mainWindow)
    (mouse :cell nil :initarg :mouse :initform nil :accessor mouse)
    (userPreferences :initarg :userPreferences :initform (cv nil) :accessor userPreferences)
    )
   (:default-initargs
     :kids (cv nil)
    ))

(defmethod initialize-instance :after ((self MGSystem) &key)
  (setf (mouse self) (cells::md-make 'Mouse)))

(defmethod sys-close (other)
  (declare (ignore other)))

(defun user-pref-set (key value)
  ;; weird sequence necessary to trigger cell ///
  (setf (userPreferences *sys*)
          (cons (cons key value)
                (delete (assoc key (userPreferences *sys*))
                        (userPreferences *sys*)))))

(defun user-pref-toggle (key)
  (user-pref-set key (not (user-pref key))))

(defun user-pref (key)
  (cdr (assoc key (userPreferences *sys*))))
    
(defmacro ^user-pref (key)
  `(bwhen (ups (^userPreferences *sys*))
      (user-pref ,key)))

;---------------------------------------------------

(defun current-application-time ()
  (ymdhmsh (current-app-universal-time  *sys*)))

(defmethod current-app-universal-time (system)
  (declare (ignore system))
  (get-universal-time))

;may write globals to file when set, then read in at startup time for continuity

(defun fm-find-system (md)
   (upper md MGSystem))


(defmodel GlutWindow (IXFamily)
  ((glutw :cell nil :initarg :glutw :initform nil :accessor glutw :type integer)
   ))

