;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package #:cello)

(defparameter *button-front-back*
  '(( ( 2.0f0  1.0f0 0.0f0)			
     (-2.0f0  1.0f0 0.0f0)			
     (-2.0f0  -1.0f0 0.0f0)			
     ( 2.0f0  -1.0f0 0.0f0))
    (( 2.4f0  1.2f0 -0.2f0)			
     (-2.4f0  1.2f0 -0.2f0)			
     (-2.4f0  -1.2f0 -0.2f0)			
     ( 2.4f0  -1.2f0 -0.2f0))
    )
  "Front 4, then back 4, CCW from top-right")


(defun gen-quad (front-back)
  (flet ((v (l v) (mapcar (lambda (f) (* 1.0f0 f))
                    (elt (elt front-back l) v))))
    (list (v 1 0)(v 1 1)(v 0 1)(v 0 0)
      (v 0 3)(v 1 3)(v 1 2)(v 0 2)
      (v 0 0)(v 0 1)(v 0 2)(v 0 3)
      (v 0 0)(v 0 1)(v 0 2)(v 0 3)
      ;;(v 1 1)(v 1 0)(v 1 3)(v 1 2)
      (v 1 0)(v 0 0)(v 0 3)(v 1 3)
      (v 0 1)(v 1 1)(v 1 2)(v 0 2)
      )))

#+test
(gen-quad *button-front-back*)

(defun draw-button ()
  (draw-quad *button-front-back*))

(defun 5cdr (x)
  (nthcdr 5 x))

(defun 4cdr (x)
  (nthcdr 4 x))

(defun draw-quad (front-back &optional (shade 0.2f0))
  (let ((quad (gen-quad front-back))
        )
    (gl-pushm)
    (gl-begin gl_quads )
    (loop for (v1 v2 v3 v4) on quad by #'4cdr
        until (or (null v1) (null v2) (null v3) (null v4)) do
          (incf shade 0.1f0)
          (gl-color3f shade shade shade)
          (apply #'glvertex3f v1)
          (apply #'glvertex3f v2)
          (apply #'glvertex3f v3)
          (apply #'glvertex3f v4)
          )
    (gl-end)
    (gl-popm)))







