;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)
           
(defvar *render-clip-l*)
(defvar *render-clip-r*)
(defvar *render-clip-t*)
(defvar *render-clip-b*)

(defun call-with-clipping (self clipped-fn)
  (if (not (clipped self))
      (macrolet ((adj-unclipped (g p)
                   `(when ,g (- ,g (,p self)))))
        (let ((*render-clip-l* (adj-unclipped *render-clip-l* px))
              (*render-clip-r* (adj-unclipped *render-clip-r* px))
              (*render-clip-t* (adj-unclipped *render-clip-t* py))
              (*render-clip-b* (adj-unclipped *render-clip-b* py)))
          (funcall clipped-fn)))
    (macrolet ((adj-clipped (mm g l p)
                 `(if ,g
                      (,mm (,l self)
                        (- ,g (,p self)))
                    (,l self))))
      (let ((*render-clip-l* (adj-clipped max *render-clip-l* ll px))
            (*render-clip-r* (adj-clipped min *render-clip-r* lr px))
            (*render-clip-t* (adj-clipped ups-most *render-clip-t* lt py))
            (*render-clip-b* (adj-clipped downs-most *render-clip-b* lb py)))
        (let* ((clip-restore0 (ix-clip self GL_CLIP_PLANE0 :xl (ll self)))
               (clip-restore1 (ix-clip self GL_CLIP_PLANE1 :xr (lr self)))
               (clip-restore2 (ix-clip self GL_CLIP_PLANE2 :yt (lt self)))
               (clip-restore3 (ix-clip self GL_CLIP_PLANE3 :yb (lb self)))
               )
          (trc nil "just clipped" self (ll self)(lr self)(lt self)(lb self))
          (ix-clip-dump "just clipped dumped")
          (let ((result (funcall clipped-fn)))
            (ix-clip-undo self GL_CLIP_PLANE0 clip-restore0)
            (ix-clip-undo self GL_CLIP_PLANE1 clip-restore1)
            (ix-clip-undo self GL_CLIP_PLANE2 clip-restore2)
            (ix-clip-undo self GL_CLIP_PLANE3 clip-restore3)
            result))))))

(defun ix-clip-undo (self p oldeqn)
  (declare (ignorable self))
  (if oldeqn
      (progn
        (trc nil "restoring anothers clip. me:" (md-name self) (eltd oldeqn 0)(eltd oldeqn 1)
          (eltd oldeqn 2)(eltd oldeqn 3))
        (gl-clip-plane p oldeqn)
        (gl-enable p))
    (gl-disable p)))

(defparameter *clipper* (make-ff-array GLDouble 0 0 0 0))

(defun ix-clip-dump (msg)
  msg
  (trc nil "clipdump" msg
    (maptimes (pn 4)
      (when (cello-get-boolean (+ GL_CLIP_PLANE0 pn))
        (gl-get-clip-plane (+ GL_CLIP_PLANE0 pn) *clipper*)
        (floor (eltd *clipper* 3))))))

(defun ix-clip (self p how eqn)
  (declare (ignorable self))
  ;;
  ;; A = y1 (z2 - z3) + y2 (z3 - z1) + y3 (z1 - z2)
  ;; B = z1 (x2 - x3) + z2 (x3 - x1) + z3 (x1 - x2)
  ;; C = x1 (y2 - y3) + x2 (y3 - y1) + x3 (y1 - y2)
  ;; - D = x1 (y2 z3 - y3 z2) + x2 (y3 z1 - y1 z3) + x3 (y1 z2 - y2 z1)        
  ;;
  (let ((x1 0)(y1 0)(z1 0)(x2 0)(y2 0)(z2 0)(x3 0)(y3 0)(z3 0) oldeqn)
    ;
    ; get current clip this plane if any for restore by caller
    ;
    #+no (when (cello-get-boolean p)
      (setf oldeqn (make-ff-array GLDouble 0 0 0 0))
      (gl-get-clip-plane p oldeqn)
      (trc nil "saving anothers clip. me:" self (eltd oldeqn 0)(eltd oldeqn 1)
          (eltd oldeqn 2)(eltd oldeqn 3)))

    (ecase how
     (:xl (setq x1 eqn x2 eqn x3 eqn)
       (setq y2 1 y3 1)
       (setq z3 (ins 1)))
     (:xr (setq x1 eqn x2 eqn x3 eqn)
       (setq y2 1 y3 1)
       (setq z3 (outs 1)))
     (:yt (setq y1 eqn y2 eqn y3 eqn)
       (setq x2 1 x3 1)
       (setq z3 (ins 1)))
     (:yb (setq y1 (+ eqn) y2 (+ eqn) y3 (+ eqn))
       (setq x2 1 x3 1)
       (setq z3 (outs 1))))

    (setf (eltd *clipper* 0)
      (+ (* y1 (- z2 z3)) (* y2 (- z3 z1)) (* y3 (- z1 z2))))

    (setf (eltd *clipper* 1)
      (+ (* z1 (- x2 x3)) (* z2 (- x3 x1)) (* z3 (- x1 x2))))

    (setf (eltd *clipper* 2)
      (+ (* x1 (- y2 y3)) (* x2 (- y3 y1)) (* x3 (- y1 y2))))

  ;; - D = x1 (y2 z3 - y3 z2) + x2 (y3 z1 - y1 z3) + x3 (y1 z2 - y2 z1)     
;;;    (trc "clipping :x1" x1 :y1 y1 :z1 z1)
;;;    (trc "clipping :x2" x2 :y2 y2 :z2 z2)
;;;    (trc "clipping :x3" x3 :y3 y3 :z3 z3)
    (setf (eltd *clipper* 3)
      (- (+ (* x1 (- (* y2 z3) (* y3 z2)))
           (* x2 (- (* y3 z1) (* y1 z3)))
           (* x3 (- (* y1 z2) (* y2 z1))))))

    (trc nil "clipping myself:" self :plane (- p gl_clip_plane0)
      (eltd *clipper* 0)(eltd *clipper* 1)
      (eltd *clipper* 2)(eltd *clipper* 3))

    (gl-clip-plane p *clipper*)
    (progn
      (gl-get-clip-plane p *clipper*)
      (trc nil "ix-clip just set/read" test (floor (eltd *clipper* 3))))
    (gl-enable p)

    oldeqn))
