;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)



(defstruct rgb ;;/// just use ogl native struct?
  (r 0 )
  (g 0 )
  (b 0 ))

(defstruct rgba fo)

(defun mkRGBA (r g b a)
  (let* ((co (uffi:allocate-foreign-object :float 4))
         (c (make-rgba :fo co)))
    (setf (uffi:deref-array co '(:array :float) 0) (/ r 255.0f0))
    (setf (uffi:deref-array co '(:array :float) 1) (/ g 255.0f0))
    (setf (uffi:deref-array co '(:array :float) 2) (/ b 255.0f0))
    (setf (uffi:deref-array co '(:array :float) 3) (/ a 255.0f0))
    c))

(defun wrap-rgba (rgba-foreign)
  (make-rgba :fo rgba-foreign))

(defun make-opengl-rgba (r g b a)
  (let* ((co (uffi:allocate-foreign-object :float 4))
         (c (make-rgba :fo co)))
    (setf (uffi:deref-array co '(:array :float) 0) (* 1.0 r))
    (setf (uffi:deref-array co '(:array :float) 1) (* 1.0 g))
    (setf (uffi:deref-array co '(:array :float) 2) (* 1.0 b))
    (setf (uffi:deref-array co '(:array :float) 3) (* 1.0 a))
    c))

(defun rgba-r (rgba)
  (c-assert (typep rgba 'rgba))
  (uffi:deref-array (rgba-fo rgba) '(:array :float) 0))

(defun rgba-g (rgba)
  (c-assert (typep rgba 'rgba))
  (uffi:deref-array (rgba-fo rgba) '(:array :float) 1))

(defun rgba-b (rgba)
  (c-assert (typep rgba 'rgba))
  (uffi:deref-array (rgba-fo rgba) '(:array :float) 2))

(defun rgba-a (rgba)
  (c-assert (typep rgba 'rgba))
  (uffi:deref-array (rgba-fo rgba) '(:array :float) 3))

(defmethod print-object ((self rgba) s)
  (format s "(r:~a g:~a b:~a a:~a)" (rgba-r self)(rgba-g self)(rgba-b self)(rgba-a self)))

(defmethod ix-render-layer ((self RGBA) gBox)
  (declare (ignore gbox))
  (gl-color4fv (rgba-fo self)))

(defmethod ix-render-layer ((self (eql :fill-lr)) gBox)
  (gl-polygon-mode GL_FRONT_AND_BACK GL_FILL)
  (gl-disable gl_blend)
  (gl-rectf (r-left gbox)(r-top gbox) (r-right gbox) (r-bottom gbox)))

(defmethod ix-render-layer ((self (eql :off)) gBox)
  (declare (ignore gBox))
  (gl-Disable GL_LIGHTING))

(defmethod ix-render-layer ((self (eql :on)) gBox)
  (declare (ignore gBox))
  (gl-Enable GL_LIGHTING))


(defstruct line-width value)

(defmethod ix-render-layer ((self line-width) gbox)
  (declare (ignore gbox))
  (gl-line-width (line-width-value self)))

(defmethod ix-render-layer ((self (eql :line-lr)) gBox)
  (trc nil "boxing" self gbox)
  (gl-polygon-mode GL_FRONT_AND_BACK GL_LINE)
  (gl-rectf (r-left gbox)(r-top gbox) (r-right gbox) (r-bottom gbox)))

(defun rgba-clear-color (rgba &aux (co (rgba-fo rgba)))
  (gl-clear-color
   (uffi:deref-array co '(:array :float) 0)
   (uffi:deref-array co '(:array :float) 1)
   (uffi:deref-array co '(:array :float) 2)
   (uffi:deref-array co '(:array :float) 3)))

(defparameter +white+ (mkRGBA 255 255 255 255))
(defparameter +red+ (mkRGBA 255 0 0 255))
(defparameter +dark-green+ (mkRGBA 0 128 0 255))
(defparameter +green+ (mkRGBA 0 255 0 255))
(defparameter +turqoise+ (mkRGBA 0 255 255 255))
(defparameter +blue+ (mkRGBA 0 0 255 255))
(defparameter +lt-blue+ (mkRGBA 127 127 255 255))
(defparameter +black+ (mkRGBA 0 0 0 255))
(defparameter +yellow+ (mkRGBA 255 255 0 255))
(defparameter +lt-yellow+ (mkRGBA 255 255 127 255))
(defparameter +purple+ (mkRGBA 255 0 255 255))
(defparameter +gray+ (mkRGBA 127 127 127 255))
(defparameter +lt-gray+ (mkRGBA 192 192 192 255))
(defparameter +dk-gray+ (mkRGBA 0.25 0.25 0.25 1))

