;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

(defmodel Control ()
  (
   (title$ :cWhen (:DocDirectory :DocPaint)
           :initarg :title$
           :initform (c? (format nil "~:(~a~)" ;; this is just a poor default-- really s.b. supplied by form author
                            (string-capitalize (substitute #\space #\- (string (md-name self))))))
           :reader title$)
   
   (controlaction :cell nil :initarg :controlaction
                  :initform nil
                  :reader controlaction)
   
   (mouse-up-handler
    :initarg :mouse-up-handler
     :initform nil :reader mouse-up-handler
     :documentation "Menus use this")
   
   (clickEvt :cWhen (:DocEdit)
             :initform (cv nil)
             :initarg :clickEvt
             :accessor clickEvt)

   (click-tolerance :cell nil
     :initform (mkv2 0 0)
     :initarg :click-tolerance
     :reader click-tolerance)
   
   (keyEvt :cell :ephemeral
           :cWhen (:DocEdit)
           :initform nil
           :initarg :keyEvt
           :accessor keyEvt)
   
   (enabled :cWhen (:DocEdit)
            :initform t
            :initarg :enabled
            :accessor enabled)
   
   (accessControl :initform nil
                  :initarg :accessControl
                  :accessor accessControl
                  :documentation
                  "A list of (lock &rest keys)'s. If code tests an undefined lock, access is granted.
 If the tested lock is defined, having any of the keys suffices to gain access.
 Room for refinement here via :and or :or.")
  
   (hilited :cWhen (:DocEdit)
            :initform (c? (bwhen (click (^clickEvt))
                              (^clickOver click)))
            :initarg :hilited
            :accessor hilited)
   
   (kbSelector :cell nil
     :initarg :kbSelector
     :initform nil ;; added 23-04-01 with some trepidation, seem to recall omission deliberate
     :reader kbSelector)
   
   ))

(defmethod do-keydown ((self control) k event)
  (declare (ignorable event))
  (when (control-triggered-by self k event)
    (funcall (controlAction self) self event)
    t)) ;; handled

(defmethod control-triggered-by (control k event)
  (declare (ignorable event))
  (eql k (kbSelector control))) ;; this is lame--to be enhanced

;;;(defmethod character-message ((self CTMarkBox) buttons aChar)
;;;  (declare (ignorable buttons))
;;;  (when (editActive (sFocuser))
;;;    (with-invalidation (self)
;;;      (when (member achar (kbSelector self))
;;;        (Control-do-action self t)))))

(defmethod set-mouseclick ((self Control) click)
  (setf (clickEvt self) click))

(defmethod ix-selectable ((self Control)) t)

(def-c-echo clickEvt ())

(defmethod ctl-disabled (other)
  (declare (ignore other))
  nil)

(defmethod ctl-disabled ((self control))
  (not (enabled self)))


(defmethod fully-enabled (self)
  (declare (ignore self))
  nil)

(defmethod fully-enabled ((self control))
  "Test if self and all ascendant controls are enabled"
  (labels ((no-disabled-up (node)
             (unless (ctl-disabled node)
               (bIf (p (fm-parent node))
                 (no-disabled-up p)
                 t ;; reached top without hitting un-enabled control
                 ))))
    (no-disabled-up self)))

;
; /// m/b odd combo of customizable parameter 'controlAction and
;     generic function 'control-do-action. we like instance-oriented
;     programming, so keep 'controlAction, maybe just coordinate better
;     by establishing a rule: call 'controlAction first, if supplied,
;     and if it returns t that indicates "handled"? ugh
;
(defmethod Control-do-action :around (ct newValue)
  (declare (ignore newvalue))
  (when (fully-enabled ct)
    (call-next-method)))

(defmethod Control-do-action (ct triggerEvt)
   (when (controlAction ct)
       (funcall (controlAction ct) ct triggerEvt)
       ;; (trc "Control-do-action triggering FINIS" ct)
       ))


