;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

(defparameter *markBoxSize* (u96ths 9))

(defmodel CTMarkBox (CTToggle Image)
  ((kbSelector :cell nil :initarg :kbSelector :initform nil :reader kbSelector)
   )
  (:default-initargs
      :lL  (- *markBoxSize*)
    :lT  (ups  *markBoxSize*)
    :lR  *markBoxSize*
    :lB  (downs *markBoxSize*)
    :layers (c? (let ((thick (* 0.15 (lwidth self))))
                  (list (outs thick)
                    (if (^enabled) +white+ +gray+)
                    
                    (make-instance 'Frame3D
                      :3dStyle :EDGE-SUNKEN
                      :of-window .w.
                      :w-box (gbox self)
                      :thickness thick)
                    :off
                    +yellow+
                    (outs thick)
                    (when (^md-value) :x-mark)
                    )))
    ))


(defmethod ix-render-layer ((self (eql :x-mark)) lBox)
  (let* ((thick (ceiling (r-width lbox) 6))
         (br (- (r-right lBox) thick)) ;; /// bogus use of thick to inset "x"
         (bl (+ (r-left lBox) thick))
         (bt (- (r-top lBox) thick))
         (bb (+ (r-bottom lBox) thick))
         )
    ;;(print `(xmark-lbox=> ,lbox thick> ,thick))
    (gl-pushm)
    (gl-line-width (max 4 (log2scr thick)))
    (gl-disable GL_TEXTURE_2D)
    (within-gl-begin-end (GL_LINES)
      (gl-vertex3f bl bt 0)(gl-vertex3f br bb 0)
      (gl-vertex3f bl bb 0)(gl-vertex3f br bt 0))
    (ogb::glec :f3d)
    (gl-popm)
    ))

;----------------------------

(defmethod context-cursor ((self CTMarkBox) kbdModifiers)
   (declare (ignore kbdModifiers))
   (if (enabled self)
     GLUT_CURSOR_CROSSHAIR
     GLUT_CURSOR_DESTROY))

; -----   radios -------------------------------

(defmodel CTRadioItem (CTToggle)
  ((kbSelector :cell nil :initarg :kbSelector :initform nil :reader kbSelector)
    )
  (:default-initargs
    :enabled t
    :md-value (c? (bwhen (radio (upper self CTRadio))
                   (eko (nil "RB watching radio" self .w. radio (windowcache radio))
                     (find (associatedValue self) (^md-value radio)))))
    :controlAction (lambda (self event)
                     (declare (ignorable event))
                     (trc nil "RB acts" self .w.)
                     (unless (md-value self)
                       (let ((r (upper self CTRadio)))
                         (trc nil "here come rb" (associatedValue self) r (windowcache r))
                         (setf (md-value r)
                           (list (associatedValue self))))))))

(defmodel CTRadioButton (CTMarkBox CTRadioItem) ())
(defmodel CTRadio (IXFamily)
  ()
  (:default-initargs
      :md-value (cv nil)))

(defmodel CTRadioRow (IXRow CTRadio)
  ()
  (:default-initargs
      :md-value (cv nil)))

(defmodel CTRadioStack (IXStack CTRadio)
  ()
  (:default-initargs
      :md-value (cv nil)))

(defun radio-on-name (radioValues)
  (some (lambda (rbValue)
            (unless (empty$ (cdr rbValue))
              (car rbValue)))
        radioValues))

;--------------- CTCheckBox --------------------------------------------

(defmodel CTCheckBox (CTMarkBox) 
  ()
  (:default-initargs
      :lighting :on
   :md-value (cv nil))
  )

(defmodel CTCheckText (Control IXRow)
  ()
  (:default-initargs
      :md-value (cv nil)
    :justify :center
    :spacing (u96ths 8)
    :outset (u96ths 2)
    ;;:layers (list +white+ :line-lr)
    :kids (c? (theKids
               (mkPart :checkBox (CTCheckBox)
                 :md-value (c? (md-value .parent))
                 :enabled nil) ;; let parent handle clicks since text is clickable by the rules
               (mkPart :label (IXText)
                 :lighting :off
                 :text$ (c? (^title$ .parent))
                 :style :button
                 :layers (c? (list ;; +white+ :fill-lr
                              (outs 100)
                              (if (^enabled .parent)
                                  +white+
                                +gray+))))))

    :controlAction (lambda (self event)
                     (declare (ignorable event))
                     (trc "checktext bingo" (not (md-value self)))
                     (setf (md-value self) (not (md-value self))))))

(defmodel CTRadioText (IXRow CTRadioItem)
  ()
  (:default-initargs
    :justify :center
    :spacing (u96ths 8)
    :outset (u96ths 2)
    :kids (c? (theKids
               (mkPart :rbutton (CTCheckBox)
                 :md-value (c? (md-value .parent))
                 :enabled nil) ;; let parent handle clicks since text is clickable by the rules
               
               (mkPart :label (IXText)
                 :lighting :off
                 :text$ (c? (^title$ .parent))
                 :style :button
                 :layers (c? (list ;; +white+ :fill-lr
                              (if (^enabled .parent)
                                  +white+
                                +gray+))))))))
