;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

;=========================================
    

(defmodel CTSelector () ;; mixin at any node containing CTSelectable's
  ((selection :accessor selection :initarg :selection)
   (selectionFocus  :initarg :selectionFocus
                    :reader selectionFocus
                    :initform nil)
   (initialSelection :initform nil :reader initialSelection :cell nil :initarg :initialSelection)
   )
   (:default-initargs
    :selection (cv nil)
    ))

(defmethod sm-unchanged-p ((self CTSelector) (slotName (eql 'selection)) newValue oldValue)
  (equal newValue oldValue))

(defun initialSelection-first (self)
  (do-like-fm-parts (it (self CTSelectable))
    (when (enabled it)
      (return-from initialSelection-first (list it)))))

(defmethod md-awaken :after ((self CTSelector))
  (when (initialSelection self)
    (with-metrics (nil nil () "selector initialselct")
      (setf (selection self) (eko ("setting initial selection" self)
                               (funcall (initialSelection self) self))))))

(def-c-echo selection ())

(defmodel CTSelectorStack (CTSelector IXStack)())
(defmodel CTExclusiveStack (CTExclusive IXStack)())
(defmodel CTSelectorRow (CTSelector IXRow)())

;----------

(defmethod selection-set (self x)
   (setf (selection self) x))

(defmethod selection-set1 (self x)
  (selection-set self (list x)))

(defun selection1 (self)
  (car (selection self)))

;----------


(defmodel CTDetails (CTSelector Focus IXDetails)
  ()
  (:default-initargs
      :outset (u8ths 1)))

#+test?
(def-c-echo kids ((self CTDetails))
  ;(trc "ctdetails kids echo" newvalue oldvalue)
  )

(defmodel CTDetailsExclusive (CTExclusive CTDetails)()) ;; go generic with CTSelectorNested?

(defmodel CTSelectable (Control)
  ((selectedp :initarg :selectedp
             :initform (c? (bwhen (selector (selector self))
                               (member self (^selection selector))))
             :reader selectedp))
  (:default-initargs
      :layers (c? (packed-flat!
                   (if (^enabled)
                       (if (^hilited)
                           +blue+
                         (if (^selectedp)
                             +yellow+
                           +white+))
                     +lt-gray+)
                   :fill-lr
                   +black+))))

(defun selector (self)
  (upper self CTSelector))

;=====================================

(defmodel CTExclusive (CTSelectable)
  ()
  (:default-initargs
   :controlAction #'CTExclusive-controlAction))

(defmethod CTExclusive-controlAction (self event)
  (declare (ignorable event))
  
  (with-metrics (nil nil (nil :type :time #+not :count-only #+not :space
                                ;; :count 2000
                                :interpret-closures t
                                ;; :count-list (list #'md::bd-bound-slot-value)
                                ) "CTExclusive-controlAction")
    (selection-set1 (selector self) self)))

;=====================================

(defmodel CTMultiChoice (CTSelectable)
  ()
  (:default-initargs
   :controlAction (lambda (self event
                                  &aux
                                  (buttons (evtButtons (oSEvent event)))
                                  (selector (selector self))
                                  (selection (selection selector)))
                      (selection-set selector
                                        (if (shift-key-down buttons)
                                           (if (member self selection)
                                              (delete self selection)
                                              (cons self selection))
                                           (list self))))))

;=====================================

(defmodel CTToggleChoice (CTToggle CTSelectable)
  ()
  (:default-initargs
   :controlAction #'CTToggleChoice-controlAction))

(defmethod CTToggleChoice-controlAction (self event
                                              &aux
                                              (buttons (evtButtons (oSEvent event)))
                                              (selector (selector self))
                                              (selection (selection selector)))
  (trc "controlaction toggle" self)
  (selection-set selector
                 (if (member self selection)
                     (remove self selection)
                   (if (shift-key-down buttons)
                       (cons self selection)
                     (list self)))))
