;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

(defmodel CTText (Control IXText)
  ()
  (:default-initargs
      :text$ (c? (string (^md-value)))
    :inset (mkv2 (u96ths 2)(u96ths 2))
    :lighting :off
    :layers (c? (list :off +white+ :fill-lr
                  (if (^enabled) 
                      (if (^mouse-over-p)
                          +green+ +black+)
                    +lt-gray+)))))

(defmodel CTButton (Control IXText) ;; same inheritance, but otherwise unrelated to CTText
  ((inset :initform (mkv2 (uPts 4) (uPts 4))))
  (:default-initargs
      :text$ (c? (^title$))
    :justifyHZ :center
    :justifyVT :center
    :style :button
    :layers (c? (let* ((thick (max (u96ths 3) (* 0.08 (lwidth self))))
                       (defl (downs (/ thick 5.0))))
                  (declare (ignorable thick defl))
                  (trc nil "ctbutton" self thick defl (^hilited))
                  (list
                   +gray+ :fill-lr
                   (if (^hilited)
                       (make-3f :x (/ defl 2) :y defl :z (outs (* 0.5 thick)))
                     (make-3f :x 0 :y 0 :z (outs (* 0.8 thick))))
                   +white+
                   :on
                   (make-instance 'Frame3D
                     :3dStyle :EDGE-Raised
                     :of-window .w.
                     :w-box (gbox self)
                     :thickness thick)
                   (if (^mouse-over-p)
                       (if (^enabled) +green+ +red+)
                     +blue+)
                   :off)))))

;;;    :lb (uin -1)
;;;    :lr (u8ths 16)))

#+worked
(c? (let* ((thick (* 0.1 (lwidth self)))
           (defl (downs (/ thick 5.0))))
      (declare (ignorable thick defl))
      (list
       +black+ :fill-lr
       (if (^hilited)
           (make-3f :x (/ defl 2) :y defl :z (outs (* 0.5 thick)))
         (make-3f :x 0 :y 0 :z (outs (* 0.8 thick))))
       +white+
       :on
       (make-instance 'Frame3D
         :3dStyle :EDGE-RAISED
         :of-window .w.
         :w-box (eko ("frame w-box") (gbox self))
         :thickness thick)
       (if (^mouse-over-p) ;;; /// incredibly inefficient
           (if (^enabled) +green+ +red+)
         +blue+)
       :off)))

; ---------------- CT FSM ---------------------
(defmodel CTFSM (Control)
  (
   (transitionFN :cell nil :initarg :transitionFN :reader transitionFN)
   (states :cell nil :initarg :states :reader states)
   )
  (:default-initargs
      :md-value (cv nil)
   :transitionFN (lambda (currentstate statetable)
                     ;(trc "CTFSM :transitionFN  curr,table" currentstate statetable)
                     (or (cadr (member currentstate statetable :test (if (stringp currentState)
                                                                        #'string-equal
                                                                        #'eql)))
                         (car statetable)))

   :controlAction (lambda (self event)
                    (declare (ignorable event))
                    (let ((newv (funcall (transitionFN self) (md-value self) (states self))))
                      (ctfsm-assume-value self newv)))))

(defmethod ctfsm-assume-value (self newValue)
  (setf (md-value self) newValue))

; --------------- CT Toggle -----------------------

(defmodel CTToggle (CTFSM)
  ((associatedValue :initform t :cell nil :initarg :associatedValue :reader associatedValue)
   (associatedShortc :initform t :cell nil :initarg :associatedShortc :reader associatedShortc))
  (:default-initargs
   :states '(nil t)))


;------------------------------------------------------
#+nope
(defmodel CTTwister (CTToggle IXPolygon) ;; convert to IMBitmapMulti??
;
; For twist-down control to open/close details
;
  ()
  (:default-initargs
   :md-value (cv nil) ;;; closed by default
   :polyStyle :fill
   :layers (c? (if (^hilited self)
                      :black :gray))
   :vertices (c? (if (^md-value self)
                     (uCvt '((2 . 4) (7 . 9) (12 . 4)) :96ths)
                     (uCvt '((4 . 2) (9 . 7) (4 . 12)) :96ths)
                     #+big '((0 . 5) (14 . 5) (7 . 12))
                     #+big '((5 . 0) (12 . 7) (5 . 14))
                     ))
   :lL 0 :lT 0 :lR (u96ths 15) :lB (downs (u96ths 15))))

(defmacro mkTwisted (twistedName (labelClass &rest labelArgs)
                                 (twistedClass &rest twistedArgs))
  `(mkPart :twistedGroup (IGZeroTL)
      :showkids (cv nil)
      :ll (minKid-ll)
      :lr (maxKid-lR)
      :kidFactory (lambda (self kidValue)
                    (declare (ignore self kidValue)))
      :kids (c? (let ((thetree self))
                   ;; (trc "making all parts of tree for" (md-value self) rethinker)
                   (theKids
                    (mkpart :header (IXKidSized)
                       :ll (u96ths -20) :px 0
                       :kids (c? (packed-flat!
                                   (mkpart :opener (CTTwister)
                                      :py (u96ths 2)
                                      :px (c? (^px-maintain-pr (u96ths -3)))
                                      :controlaction (lambda (self event &aux (twistGroup (fm-parent .parent)))
                                                       (declare (ignorable self event))
                                                       (with-transaction ()
                                                         (setf (showkids twistGroup) (not (showkids twistGroup))))))
                                   (mkPart ,twistedName (,labelClass)
                                      ,@labelArgs))))
                    (mkpart ,twistedName (,twistedClass)
                       ,@twistedArgs
                       :px 0
                       :py (c? (^priorsib-pb self)))
                    )))))

(defmacro mkTwistedPart (twistedName (label$ &rest labelArgs)
                                                                twistedPart)
   `(mkPart :twistedGroup (IGZeroTL)
           :showkids (cv nil)
           :ll (minKid-ll)
           :lr (maxKid-lR)
           :kidFactory #'null
           :kids (c? (theKids
                                (mkpart :header (IXKidSized)
                                      :ll (u96ths -20) :px 0
                                      :kids (c? (packed-flat!
                                                         (mkpart :opener (CTTwister)
                                                               :py (u96ths 2)
                                                               :px (c? (^px-maintain-pr (u96ths -3))))
                                                         (mkPart ,twistedName (IXText)
                                                               ,@labelArgs
                                                               :text$ ,label$))))
                                ,twistedPart
                              ))))

