;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :Cello)

;------------------------------

(def-c-echo focus ()
  (when old-Value
    (image-set-dirty old-Value nil 'focused-on)
    (setf (focused-on old-Value) nil))

  (when new-Value
    (setf (focused-on new-Value) t)
    (trc nil "focus now>" new-value)
    (image-set-dirty new-Value nil 'focused-on)
    (unless (and old-Value
                 (focus-shared-by old-Value new-Value))
      (turn-editActive self (not (editrequiresactivation new-Value))))
    ))

(defmethod turn-editActive (self new-value)
  (declare (ignorable self new-value)))

(defmethod focus-shared-by (f1 f2)
  (declare (ignore f1 f2))
  nil)


(def-c-echo textualfocus (self newvalue oldvalue)
  (when newValue
    (setf (insertionPt self) 0)
    (setf (selEnd self) (bWhen (t$ (text$ newValue))
                           (length t$)))))


(defmethod focus-starting ((self FocusMinder))
  (or (focusMinded self)
      (focus-find-first self)
      (focus-find-first self :tabStopOnly nil)))

(defmethod focus-on (self &optional focuser)
   (c-assert (or self focuser))
   ;;(trc "focus-on self, focuser" self focuser)
   (setf (focus (or focuser (sFocuser))) self))

(defmethod focus-gain (self)
   (declare (ignore self)))

(defmethod focus-lose (self newFocus)
  (if self
      (focus-lose (fm-parent self) newfocus)
    t) ;; means "yielded"
  )

;________________________________ I d l i n g _______________________
;
(defmethod focus-idle (other) (declare (ignorable other)))

(defmethod focus-idle ((list list))
  (dolist (f list)
    (focus-idle f)))

;_____________________ I n t o - V i e w _____________________
; 
; 990329 /// kt Resurrect eventually
;
(defmethod focus-scroll-into-view ((focus Focus))
  ;;  temp to get going (image-scroll-into-view focus)
  )

(defmethod focus-scroll-into-view (other)
  (declare (ignore other)))

(defmethod focus-scroll-into-view ((focii list))
  (dolist (focus focii)
    (focus-scroll-into-view focus)))

(defun focus-find-first (self &key (test #'true) (tabStopOnly t))
  (let ((finder (lambda (focus)
                    (and (typep focus 'focus)
                         (visible focus)
                         (fully-enabled focus)
                         (or (not tabstopOnly)
                             (tabstopp focus))
                         (funcall test focus)))))
     (declare (dynamic-extent finder))
     (fm-find-if self finder)))


