;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

#| To do

- ix-set-dirty when giving/taking focus

- IXEditSelection needs a home

;;; also got FFComposite rule deciding it was active if any kid was

arrange for Focuser to process clicks and keys first, then mebbe dump into dvk, do-click/do-double-click
bottom up from focus/imageunder

arrange for Controller to process clicks first, then mebbe dump into do-click/do-double-click
bottom up from focus/imageunder

add finalization for radio button (look at others, see if ICR can ne de-celled

when focused and editactive, push caret onto kids of cttext? or work out how to show
it without it being a kid there

|#

(defmodel Focuser (IXCanvas)
  (
   (focus :initarg :focus
          :initform (cv nil)
          :accessor focus)
   
   (textualfocus :initarg :textualfocus
                 :initform nil #+chya (c? (bWhen (focus (and #+runtime-system (^activep (swdw))
                                                   (^focus)))
                                   (when (and (typep focus 'CTText) ;; possibly any 'IXText?
                                              (^editActive))
                                     focus)))
                 :accessor textualfocus)

   (editActive :initarg :editActive
                 :initform (cv nil)
                 :accessor editActive)
     
     (insertionpt :initform (cv 0)
                  :initarg :insertionpt
                  :accessor insertionpt)
     
     (selEnd :initform (cv nil)
             :accessor selEnd)
     
     (selRange :documentation "selEnd identified during drag operation"
       :reader selrange :initarg :selRange
               :initform nil #+chya (c? (bWhen (focus (^textualfocus))
                                 (bwhen (clickevt (^clickEvt focus))
                                    (bwhen (mp (^inDrag clickEvt))
                                      (CTText-find-ip focus mp))))))
     
     (undoData :cell nil :initarg :undoData :accessor undoData
               :initform nil #+hunh (new-undoData)
               :documentation "Data structure holding undo information"
               )
     )
   (:default-initargs
       :kids (c? (theKids (^content)
                   ;        (mkPart :caret (CTEditcaret))
                    ;       (mkPart :selBox (IXEditSelection))
                   ))))

(defun focuser (self)
  (swdw)
  )

(defmacro sFocuser ()
  `(focuser self))
  
(defmethod (setf focus) :around (newFocus self) ;; better be Focuser
  (let ((currFocus (focus self)))
    (trc nil "setf focus sees focuser,new,curr" self newfocus currfocus)
    (unless (eql newFocus currFocus)
      (focus-lose currFocus newFocus)
      (focus-gain newFocus))
    (call-next-method)
    ))



;;;(defmethod character-message :around ((self Focuser) buttons aChar)
;;;  (declare (ignore buttons aChar))
;;;  (when (and (focus self)
;;;             (editActive self))
;;;    (call-next-method)))

;         _____
(defmodel Focus ()
  ((focusthickness :cell nil :initarg :focusthickness
                   :initform (u96ths 3)
                   :accessor focusthickness)

   (tabMode :documentation ":ceiling :stop or nil"
            :cell nil :initarg :tabMode
            :initform :stop
            :accessor tabMode)

   (focused-on :initarg :focused-on
               :initform (cv nil)
               :accessor focused-on)))

(defun tabStopp (self)
  (eql :stop (tabMode self)))

(defmethod tabMode (other)
   (declare (ignore other))
   nil)

(defmethod editRequiresActivation (self)
  (declare (ignore self)))

(defmodel FocusMinder ()
  ;
  ; an entity which remembers which descendant was focused when the
  ; window focus moves outside the FocusMinder. This 'minded' focus
  ; is restored as the window's focus if the FocusMinder itself
  ; becomes the window's focus (if no minded focus, we focus-first)
  ;
  ((focusMinded :accessor focusMinded :initarg :focusMinded
                :cWhen (:DocEdit)
                :initform (c? (let ((focus (^focus .w.)))
                                 (if (fm-includes self focus)
                                     (if (eql self focus)
                                         .cache
                                       focus)
                                   .cache))))))

