;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

(defstruct font)

(defstruct (font-glut (:include font)(:conc-name nil))
  glutID)

(defun font-ffi-glut-id (font)
  (ffi-glut-id (glutid font)))

(defun ffi-glut-id (id)
  (make-ff-pointer id))

(defstruct (font-glut-stroke (:include font-glut)(:conc-name nil))
  (font-size 48))

(defstruct (font-glut-bitmapped (:include font-glut)(:conc-name nil)))
;------------------------------------------------------------

(defmethod font-height ((font font-glut-bitmapped))
  (scr2log
   (glut-bitmap-height (font-ffi-glut-id font))))

(defmethod font-height ((font font-glut-stroke))
  (progn ;; scr2log
   (font-size font)))

(defmethod font-descent ((font font-glut-stroke))
  (- (font-height font)))

(defmethod font-ascent ((font font-glut-stroke))
  0)

;---------------------------------------------------------
(defmethod font-caret-height ((font font-glut))
  (font-height font))

(defmethod font-em-width (font)
  (font-char-width font (char-code #\m)))
;------------------------------------------------------

(defmethod font-char-width ((font font-glut-bitmapped) (char character))
  (font-char-width font (char-code char)))

(defmethod font-char-width ((font font-glut-bitmapped) (char fixnum))
  (glut-bitmap-width (font-ffi-glut-id font) char))

(defmethod font-char-width ((font font-glut-stroke) (char character))
  (font-char-width font (char-code char)))

(defmethod font-char-width ((font font-glut-stroke) (char fixnum))
  (glut-stroke-width (font-ffi-glut-id font) char))

;---------------------------------------------------------

(defmethod font-string-length :around (font string &optional start end)
  (declare (ignorable font start end))
  (if string
      (call-next-method)
    0))

(defmethod font-string-length ((font font-glut-bitmapped) string &optional start end)
  (if start
      (progn
        (unless end
          (setf end (length string)))
        (loop for ix from start to end
            summing (font-char-width font (char-code (elt string ix)))))
    (progn
      (assert (not end))
      (uffi:with-cstring (native$ string)
        (scr2log (glut-bitmap-length (font-ffi-glut-id font) native$))))))

(defparameter *glut-stroke-factor* 1)

(defmethod font-string-length ((font font-glut-stroke) string &optional start end)
  (if start
      (progn
        (unless end
          (setf end (length string)))
        (loop for ix from start to end
            summing (font-char-width font (char-code (elt string ix)))))
    (progn
      (assert (not end))
      (uffi:with-cstring (native$ string)
        (scr2log 
         (round (* *glut-stroke-factor*
                  (glut-stroke-length (font-ffi-glut-id font) native$)
                  (/ (font-size font) (glut-stroke-height (font-ffi-glut-id font))))))))))

;------------------------------------------------------------
(defun font-string-width
     (target-resolution
      font string
      &optional (start 0 start-supplied) (end (length string)))
  
  (floor (* (cs-logical-dpi) (if start-supplied
                                 (font-string-length font string start end)
                               (font-string-length font string)))
    target-resolution))


