;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

(defparameter *ftgl-dll* nil)

(defparameter *ftgl-path* "c:/dev/lib/ftgl.dll"
  #+unstable "c:/dev2/ftgl/win32_vcpp/build/ftgl_dynamic_MTD_d.dll")

(defun-ffx :unsigned-byte "ftgl" "fgcSetFaceSize" (:void *f :int size :int res))
(defun-ffx :float "ftgl" "fgcAscender" (:void *f))
(defun-ffx :float "ftgl" "fgcDescender" (:void *f))
(defun-ffx :float "ftgl" "fgcStringAdvance" (:void *f :cstring text))
(defun-ffx :float "ftgl" "fgcStringX" (:void *f :cstring text))
(defun-ffx :void "ftgl" "fgcRender" (:void *f :cstring text))
(defun-ffx :void "ftgl" "fgcFree" (:void *f))


(defun-ffx (* :void) "ftgl" "fgcBitmapMake" (:cstring typeface))
(defun-ffx (* :void) "ftgl" "fgcPixmapMake" (:cstring typeface))
(defun-ffx (* :void) "ftgl" "fgcTextureMake" (:cstring typeface))
(defun-ffx (* :void) "ftgl" "fgcPolygonMake" (:cstring typeface))
(defun-ffx (* :void) "ftgl" "fgcOutlineMake" (:cstring typeface))
(defun-ffx (* :void) "ftgl" "fgcExtrudedMake" (:cstring typeface))
(defun-ffx :unsigned-byte "ftgl" "fgcSetFaceDepth" (:void * :float depth))

(defun ftgl-init ()
  (unless *ftgl-dll*
    (assert (setq *ftgl-dll*
                  (eko ("Loading FTGL")
                    (uffi:load-foreign-library *ftgl-path*
                    :force-load #+lispworks t #-lispworks t
                    :module "ftgl")))
      () "Unable to load FTGL from: ~a" *ftgl-path* )))


;; --------- ftgl structure -----------------

(defstruct ftgl
  face size depth
  descender ascender bboxes
  ifont)

(defstruct (ftgl-disp (:include ftgl))
  ready-p)
(defstruct (ftgl-pixmap (:include ftgl-disp)))
(defstruct (ftgl-texture (:include ftgl-disp)))
(defstruct (ftgl-bitmap (:include ftgl)))
(defstruct (ftgl-polygon (:include ftgl)))
(defstruct (ftgl-extruded (:include ftgl)))
(defstruct (ftgl-outline (:include ftgl)))

(defmethod ftgl-ready (font)
  (declare (ignorable font))
  t)

(defmethod (setf ftgl-ready) (new-value (font ftgl-disp))
  (setf (ftgl-disp-ready-p font) new-value))

(defmethod (setf ftgl-ready) (new-value font)
  (declare (ignore new-value font)))

(defmethod ftgl-ready ((font ftgl-disp))
  (ftgl-disp-ready-p font))

(defmethod font-height ((font ftgl))
  (+ (abs (font-ascent font))
    (abs (font-descent font))))

(defmethod font-ascent ((font ftgl))
  (trc nil "ascender" (ftgl-get-ascender font))
  (scr2log (ftgl-get-ascender font)))

(defmethod font-descent ((font ftgl))
  "Returns a negative measure of display below font origin"
  (trc nil "descender" (ftgl-get-descender font))
  (scr2log (ftgl-get-descender font)))

(defmethod font-string-length ((font ftgl) string &optional start end)
  (when start
    (unless end
      (setf end (length string))))

  (uffi:with-cstring (cs (if (or start end)
                             (subseq string start end)
                           string))
      (fgc-string-advance (ftgl-get-metrics-font font) cs)))

#+allegro
(defun xftgl ()
  (dolist (dll (ff:list-all-foreign-libraries))
    (when (search "ftgl" (pathname-name dll))
      (print `(Unloading foreign library ,dll))
      (setf *ftgl-dll* nil)
      (ff:unload-foreign-library dll))))

(defun ftgl-font-ensure (w type face size &optional (depth 0))
  (or (cdr (assoc (list type face size) (w-fonts w) :test 'equal))
    (let ((f (funcall (ecase type
                        (bitmap 'make-ftgl-bitmap)
                        (pixmap 'make-ftgl-pixmap)
                        (texture 'make-ftgl-texture)
                        (outline 'make-ftgl-outline)
                        (polygon 'make-ftgl-polygon)
                        (extruded 'make-ftgl-extruded))
               :face face :size size :depth depth)))
      (push (cons (list type face size) f) (w-fonts w))
      f)))

(defun ftgl-get-ascender (font)
  (or (ftgl-ascender font)
    (setf (ftgl-ascender font)
        (fgc-ascender (ftgl-get-metrics-font font)))))

(defun ftgl-get-descender (font)
  (or (ftgl-descender font)
    (setf (ftgl-descender font)
        (fgc-descender (ftgl-get-metrics-font font)))))

(defun ftgl-get-display-font (font)
  (bif (cf (ftgl-ifont font))
    (if (ftgl-ready font)
        cf
      (progn
        (setf (ftgl-ready font) t)
        ;;(gl-color4fv (rgba-fo +white+))
        (fgc-set-face-size cf (ftgl-size font) (cs-target-res))
        cf))
    (setf (ftgl-ifont font)
      (let ((fgc (ftgl-font-make font)))
        (fgc-set-face-size fgc (ftgl-size font) (cs-target-res))
        fgc))))

(defun ftgl-get-metrics-font (font)
  (or (ftgl-ifont font)
    (setf (ftgl-ifont font) (ftgl-font-make font))))

(defun ftgl-font-make (font)
  (let ((path (format nil "C:\\WINDOWS\\Fonts\\~a.ttf" (ftgl-face font))))
    (if (probe-file path)
        (uffi:with-cstring (fpath path)
          (bIf (f (fgc-font-make font fpath))
            (progn
              (fgc-set-face-size f (ftgl-size font) (cs-target-res))
              f)
            (c-break "cannot load ~a font ~a" (type-of font) fpath)))
      (c-break "Font not found: ~a" path))))

(defmethod fgc-font-make ((font ftgl-pixmap) fpath)
  (fgc-pixmap-make fpath))
  
(defmethod fgc-font-make ((font ftgl-bitmap) fpath)
  (trc nil "making new ftgl bitmap font" (ftgl-face font) (ftgl-size font))
  (fgc-bitmap-make fpath))
  
(defmethod fgc-font-make ((font ftgl-texture) fpath)
  (fgc-texture-make fpath))

(defmethod fgc-font-make ((font ftgl-extruded) fpath)
  (let ((fgc (fgc-extruded-make fpath)))
    (fgc-set-face-depth fgc (ftgl-depth font))
    fgc))

(defmethod fgc-font-make ((font ftgl-outline) fpath)
  (fgc-outline-make fpath))

(defmethod fgc-font-make ((font ftgl-polygon) fpath)
  (fgc-polygon-make fpath))

(defmethod font-bearing-x ((font ftgl) &optional (text "m"))
  (uffi:with-cstring (cs text)
    (fgc-string-x (ftgl-get-metrics-font font) cs)))

(defmethod font-bearing-x (font &optional text)
  (declare (ignorable font text))
  0)

