(in-package :cello)

(defun make-rgba-mixer (md-name)
  (make-part md-name 'IXStack
    :justify :right
    :md-value (c? (eko (nil "rgba mixer color" md-name)
                    (make-opengl-rgba (v2-h (^md-value (fm-other :red)))
                      (v2-h (^md-value (fm-other :green)))
                      (v2-h (^md-value (fm-other :blue)))
                      (v2-h (^md-value (fm-other :alpha))))))
    :kids (c? (thekids
               (arow (:spacing (u16ths 1) :justify :center)
                 (alabel "r:" :layers +white+) (make-slider :red))
               (arow (:spacing (u16ths 1) :justify :center)
                 (alabel "g:" :layers +white+) (make-slider :green))
               (arow (:spacing (u16ths 1) :justify :center)
                 (alabel "b:" :layers +white+) (make-slider :blue))
               (arow (:spacing (u16ths 1) :justify :center)
                 (alabel "a:" :layers +white+) (make-slider :alpha))))))

(defun starter-hedron ()
  (aStack (); (:spacing (uin .5))
    (aRow (:spacing (uin .5))
      ;;col 1
      (aStack (:spacing (uPts 8)
                ;;:layers (list +black+ :fill-lr)
                :lighting :off
                :justify :right
                :clipped nil)
        
        (aStack ()
          (aLabel "Color" :layers +white+)
          (make-rgba-mixer :hedro-color))
        
        (mkPart :solidity (CTRadioStack)
          ;;:spacing (uPts 4)
          :md-value (cv (list 'Wireframe))
          :clipped nil
          :kids (c? (theKids
                     (mapcar (lambda (s)
                               (mkPart :rb (CTRadioText)
                                 :associatedvalue s
                                 :title$ (string-capitalize
                                          (format nil "~d" s))))
                       '(Solid Wireframe))))))
      ;; col 2
      (aStack (:spacing (upts 4))
        (aLabel "Sides:"
          :layers +white+)
        (mkPart :shape (CTRadioStack)
          :spacing (uPts 4)
          :md-value (cv (list 'teapot))
          :clipped nil
          :kids (c? (theKids
                     (mapcar (lambda (s)
                               (mkPart :rb (CTRadioText)
                                 :associatedvalue s
                                 :title$ (string-capitalize
                                          (format nil "~d" s))))
                       '(4 8 12 20 cube sphere cone torus teapot))))))
      ;; col 3
      (mkPart :options (ixstack)
        :spacing (upts 4)
        :md-value (c? (fm-value-dictionary self 'md-value))
        :kids (c? (thekids
                   (aStack ()
                     (aLabel "Rotation" :layers +white+)
                     (make-slider :rotx)
                     (make-slider :roty)
                     (make-slider :rotz))
                   (aStack ()
                     (aLabel "Line width" :layers +white+)
                     (make-slider :line-width))
                   (aLabel "Options:"
                                           :layers +white+)
                   (shape-options)))))
  
  (mkPart :hedron (Hedron)
    :lL (u96ths -400) :lT (ups (u96ths 300))
    :lr (u96ths 400) :lb (downs (u96ths 300))
    :clipped t
    :lighting :on
    :layers (c? (list (ins 300) (^md-value (fm-other :hedro-color))))
    :md-value (c? (list* (car (^md-value (fm^ :shape)))
                    (car (^md-value (fm^ :solidity)))
                    (mapcan (lambda (de)
                              (list (car de)(cdr de)))
                      (md-value (fm^ :options))))))))

(defun shape-options ()
  (declare (ignore other))
  (mapcar (lambda (id-max)
            (destructuring-bind (id max) id-max
              (aRow (:spacing (upts 5)
                      :collapsed (c? (not (^visible)))
                      :visible (c? (find id (shape-ids (car (md-value (fm^ :shape)))))))
                (aLabel (string-capitalize id) :layers +white+)
                (make-slider id
                  :md-value-fn (lambda (dragPct)
                                 (* (expt (v2-h dragpct) 2) max))))))
    '((:SIZE  5) (:BASE  5) (:HEIGHT  5)
      
      (:INNER-R  2)(:OUTER-R  5)
      (:SLICES  60) (:STACKS  60) 
      (:SIDES  60) (:RINGS  60))))

(defmethod shape-ids ((shape (eql 'cone)))
  '(:base :height :slices :stacks))

(defmethod shape-ids ((shape (eql 'torus)))
  '(:inner-r :outer-r :sides :rings))

(defmethod shape-ids ((shape (eql 'sphere)))
  '(:size :stacks :slices))

(defmethod shape-ids (other)
  (declare (ignore other))
  '(:size))

(defmodel Hedron (Image)
  ()
  (:default-initargs
      :lighting :on
    :md-value 8))

(def-c-echo rotation ()
  (image-set-dirty self 0 0))

(let ((rx 0)(ry 0)(rz 0))
  (defmethod ix-render ((self Hedron))
    (let ((w .w.))
      (gl-pushm)
      (gl-matrix-mode GL_PROJECTION)
      (gl-Load-Identity)
      (trc nil "tetra frame" (lL self) (lR self) (lB self) (lT self))
      (gl-ortho (lL w) (lR w) (lB w) (lT w) 100 -10000)
      
      (gl-matrix-mode gl_modelview)
      ;(gl-pushm)
      (gl-disable GL_TEXTURE_2D)
      (gl-translatef 0 0 -100) ;;-1440)
      ;;(trc "evaluating md-value" self)
      
      (let ((sc (floor (lWidth self) 4)))
        (gl-scalef sc sc sc))
      
      (destructuring-bind (shape solidity &key
                            (size 1)(height 1)(base 1)
                            (rotx (mkv2 0 0))(roty (mkv2 0 0))(rotz (mkv2 0 0))
                            (inner-r .5)(outer-r 1)
                            (sides 8)(rings 8)(slices 10)(stacks 10)
                            (line-width (mkv2 4 0))
                            &allow-other-keys)
          (md-value self)
        (gl-line-width (* (v2-h line-width) 32))
        (gl-rotatef (incf rx (* 10 (v2-h rotx))) 1 0 0)
        (gl-rotatef (incf ry (* 10 (v2-h roty))) 0 1 0)
        (gl-rotatef (incf rz (* 10 (v2-h rotz))) 0 0 1)
        (apply (intern (string-upcase
                        (conc$ "glut-"
                          (if (eql solidity 'solid) "solid-" "wire-")
                          (case shape
                            (4 (gl-rotatef 10 1 0 0) "tetrahedron")
                            (8 (gl-rotatef 10 1 0 0) "octahedron")
                            (12 "dodecahedron")
                            (20 "icosahedron")
                            (otherwise (string shape))))) :cello)
          (case shape
            (cone 
             (list base height (round slices) (round stacks)))
            ((cube teapot);; (gl-rotatef 10 1 0 0)
             (list size))
            (torus
             ;(gl-rotatef 10 1 0 0)
             (list inner-r outer-r (round sides) (round rings)))
            (sphere
             (list size (round slices) (round stacks))))))
      (gl-popm))))