;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

(defmodel Image (Model)
   (
    (glName :cell nil :initarg :glName :initform nil :accessor glName)
    (clipped :cell nil :initarg :clipped :initform nil :reader clipped)
    (inset :cell nil :initarg :inset :initform (mkv2 0 0) :reader inset)
    (dirty :cell nil :initarg :dirty :initform nil :accessor dirty)
    ;
    ;    visibility
    ;
    (visible :initarg :visible :initform t :accessor visible)
    (collapsed :initarg :collapsed
                     :initform nil :accessor collapsed)
    ;
    ;    geometry
    ;
    (px :initarg :px :initform nil :accessor px)
    (py :initarg :py :initform nil :accessor py)
    (ll :initarg :ll :initform nil :accessor ll)
    (lt :initarg :lt :initform nil :accessor lt)
    (lr :initarg :lr :initform nil :accessor lr)
    (lb :initarg :lb :initform nil :accessor lb)
    (zoom :initarg :zoom :initform nil :accessor zoom)
    (rotation :initarg :rotation :initform nil :accessor rotation)
    ;
    ; appearance
    ;
    (lighting :initarg :lighting :initform nil :accessor lighting)
    (skin :initarg :skin :initform nil :accessor skin)
    (layers :initarg :layers :reader layers :initform nil)
    (mouse-over-p :initarg :mouse-over-p :accessor mouse-over-p :initform (cv nil))

    (menu-right-items-factory :initform nil
                              :initarg :menu-right-items-factory
                              :reader menu-right-items-factory)

    (menu-select-handler :initform nil
                             :initarg :menu-select-handler
                             :reader menu-select-handler)

    ;
    ; cached calculations
    ;
    (.windowCache :cell nil :initarg :windowCache :initform nil :accessor windowCache)
    ))

(def-c-echo skin ()
  (ix-invalidate self))

(defun image-set-dirty (self x y)
  (declare (ignore x y))
  (ix-invalidate self))

(defmacro dbggeo (slot)
  `(defmethod ,slot :around (self)
     (bIf (rv (call-next-method))
       rv
       (c-break "null ~a for ~a" ',slot self))))

;;;(dbggeo px)
;;;(dbggeo py)
;;;(dbggeo ll)
;;;(dbggeo lt)
;;;(dbggeo lr)
;;;(dbggeo lb)

(defmethod ix-invalidate (self)
  (format t "~&ix-invalidate what? ~a" self))

(defmethod ix-invalidate ((self image))
  (trc nil "imageex> invalidating" self)
  (ix-invalidate .w.))

;-------------------------------

(defun find-layer-typed (layers type)
  (if (listp layers)
     (find-if (lambda (layer)
                  (typep layer type))
              layers)
     (when (typep layers type)
       layers)))

;------------------------------
(def-c-echo mouse-over-p ()
  (bwhen (p .parent)
    (when (typep p 'Image)
      (setf (mouse-over-p p) new-value))))

(defmethod ix-render (self)
  (declare (ignorable self))
  #+shhh(trc "ix-render fell through" self (class-of self)))

(defmethod md-awaken :after ((self Image))
  #+not (unless (typep self 'Window) ;; set during ii
    (setf (glName self) (eko (nil "setting glname" self)
                          (1+ (fm-max .w. (lambda (node)
                                            (if (eql self node)
                                                -1 (or (glName node) -1)))))))))

(let ((ixr-box (mkR 0 0 0 0)))
  (defmethod ix-render :around ((self Image))
;;;    (when (and *render-clip-l*
;;;            (< (pR self) *render-clip-l*))
;;;      (return-from ix-render nil))
;;;
;;;    (when (and *render-clip-r*
;;;            (> (pL self) *render-clip-r*))
;;;      (return-from ix-render nil))
;;;
;;;    (when (and *render-clip-t*
;;;            (< (pB self) *render-clip-t*))
;;;      (return-from ix-render nil))
;;;
;;;    (when (and *render-clip-b*
;;;            (> (pT self) *render-clip-b*))
;;;      (trc nil "clipping self too low" self (pT self) *render-clip-b*)
;;;      (return-from ix-render nil))

    (count-it :ix-render)
    (let ((latePrimary t))
      ;; (trc "ix-render around sees" self )
      
      (unless (c-stopped)
        (when (and (or (not *selecting*)
                     (ix-selectable self))
                (visible self)
                (not (collapsed self)))
          (let ((oldlighting (and (lighting self)
                               (cello-get-boolean GL_LIGHTING))))
            (gl-pushm)
            (case (lighting self) ;; default is "same as parent"
              (:on (gl-enable GL_LIGHTING))
              (:off (gl-disable GL_LIGHTING)))
            
            (bWhen (n (glName self))
              (gl-load-name n))
            
            (cgl-normal 1 0 0 0 0 0 0 0 1) ;; all messed up
            
            (with-clipping (self)
              (flet ((dispatch (lyr)
                       (when lyr
                         (case lyr
                           (:primary (setf latePrimary nil)
                             (call-next-method))
                           ;;(:invert (setf lateInvert t)) ;; do last always
                           (otherwise
                            (when (find (md-name self) '(:xy-pos :xyz-pos))
                              (trc "render layers od" self))
                            (ix-render-layer lyr ixr-box))))))
                
                ;-------
                ; layers
                ;
                (bwhen (layers (layers self))
                  (nr-make ixr-box (ll self) (lt self) (lr self) (lb self))
                  (if (listp layers)
                      (dolist (layer layers)
                        (dispatch layer))
                    (dispatch layers))))
              ;-------------------------------------------------
              ; primary (unless explicitly triggered in 'layers)
              ;
              (when latePrimary
                (call-next-method)))
            
            (when (lighting self)
              (if oldlighting
                  (gl-enable GL_LIGHTING)
                (gl-disable GL_LIGHTING)))
            
            (gl-popm)))))))
  
(defmethod ix-selectable ((self image)) nil)

(defmethod ix-click-transparent ((self Image))
  nil)


(defun inset-h (v)
  (etypecase v
    (number v)
    (v2 (v2-h v))
    (image (inset-h (inset v)))))

(defun inset-v (v)
  (etypecase v
    (number v)
    (v2 (v2-v v))
    (image (inset-h (inset v)))))

(defmethod call-^fillRight (self filled padding)
  (- (inset-lR filled)
     (or padding 0)
     (v2-h (^offset-within self filled))))

(defmethod (setf pXY) (newOffset self)
  (setf (px self) (v2-h newOffset))
  (setf (py self) (v2-v newOffset)))

(defmethod dirty (other)
  (declare (ignore other))
  nil)

(defmethod gOffset ((self Image) &optional (accumH 0) (accumV 0))
  (trc nil "goffset self" self 'px (px self) 'py (py self) 'fm-parent (fm-parent self))
  (let (
        (oH (+ accumH (or (px self) 0)))
        (oV (+ accumV (or (py self) 0)))
        )                               
     (if (null (fm-parent self))
        (mkv2 oH oV)
        (gOffset (fm-parent self) oH oV))))

(defmethod winOffset ((self Image) &optional (accumH 0) (accumV 0))
  ;(trc "goffset self" self 'px (px self) 'py (py self) 'fm-parent (fm-parent self))
  (let (
        (oH (+ accumH (or (px self) 0)))
        (oV (+ accumV (or (py self) 0)))
        )                               
     (if (null (fm-parent self))
        (mkv2 oH oV)
        (winOffset (fm-parent self) oH oV))))

(defmethod gUnScaled ((self null) value)
  value)

(defmethod gUnScaled (self value)
  (gUnScaled (fm-parent self) value))

(defmethod visible-fully ((self Image)) ;; this used to be an :around on visible, but then focus-first
   (and (visible self)                  ;; could not find focus on page it was /going to/ (not yet visi)
        (or (null (fm-parent self))      ;; ...not sure who need visible to go up all the way
            (visible (fm-parent self)))))

(defmethod visible (other)
   (declare (ignore other))
   t)

(defmethod visible ((other null))
   (c-break "visible called on NIL"))

(defmethod not-to-be :before ((self Image)) ;kt; was :after until 01-03-03
  ;;(ix-invalidate self)
  )

(defmethod dbgAwake ((ap Image))
   (and (dbgAwakeNum ap 'px)
        (dbgAwakeNum ap 'py)
        (dbgAwakeNum ap 'lL)
        (dbgAwakeNum ap 'lR)
        (dbgAwakeNum ap 'lT)
        (dbgAwakeNum ap 'lB)
        )
   #+nope (unless (>= (lB ap) (lT ap)) ;; this happens normally as structures get "collapsed" etc
            (inspect ap)
            (error 'xSystemfatal :appFunc 'dbgAwake :errortext "Bottom less than top: self, lT, height, lB"
              :otherdata (list ap  (lT ap) (lHeight ap) (lB ap))))
   (call-next-method))

(defun dbgAwakeNum (model slotName)
   (when (slot-value model slotname)
      (c-assert (slot-value model slotname) nil
        "Error upon awakening: slot ~s of model ~s not specified; must be" slotname model)
      (c-assert (numberp (slot-value model slotName)) nil
        "Error upon awakening: slot ~s of model ~s not numeric" slotname model)))


; ------------------- right-click -------------------------

(defmethod make-menu-right-items ((self Image))
  (bwhen (f (menu-right-items-factory self))
     (funcall f self)))

(defmethod menu-right-select ((self Image) item)
  (when item
    (bwhen (h (menu-select-handler self))
       (funcall h self item))))

(defstruct gl-optioner option enable)

(defmacro glon (option)
  `(make-gl-optioner :option ,option :enable t))

(defmacro gloff (option)
  `(make-gl-optioner :option ,option :enable nil))

(defmethod ix-render-layer ((self gl-optioner) lbox)
  (declare (ignore lbox))
  (funcall (if (gl-optioner-enable self) 'gl-enable 'gl-disable)
    (gl-optioner-option self)))