;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

(defmodel IXCanvas (IXFamily)
  (
   (targetRes :cWhen (:DocPaint)
              :initarg :targetRes
              :initform (cs-logical-screen-resolution)
              :accessor targetRes)
   
   (enclosingRes :initarg :enclosingRes
                 :initform nil
                 :accessor enclosingRes)
   
   (content :initarg :content
            :initform nil
            :accessor content))
  
  (:default-initargs
    :kids (c? (theKids (^content)))))

(defmethod ncanvas-to-screen-point ((self IXCanvas) point)
  (if (enclosingRes self)
      (nres-to-res point (targetRes self) (enclosingRes self))
    ;
    ; if no encloser we are at the top, time to convert
    ; to screen res...
    (logical-to-screen-point point (targetRes self))))

(defmodel IXCanvasNested (IXCanvas)
  (
   (targetRes :cWhen (:DocPaint)
              :initarg :targetRes
              :initform (cs-logical-screen-resolution)
              :accessor targetRes)
   
   (enclosingRes :initarg :enclosingRes
                 :initform (c? (bWhen (es (upper self IXCanvas))
                                   (^targetRes es)))
                 :accessor enclosingRes)
   
   (content :initarg :content
            :initform nil
            :accessor content)))
  
(defun at-enclosing-res (self n)
  (if (or (null (enclosingRes self))
          (eql (enclosingRes self) (targetres self)))
      n
    (progn
      (trc nil "targetres adj" self (targetres self) (enclosingRes self) n (round (* n (targetres self))
             (enclosingRes self)))
      (res-to-res n (targetres self) (enclosingRes self)))))

(defmodel IXCanvasKidSized (IXCanvasNested)
  ()
  (:default-initargs
      :lL (c? (bIf (root (car (^kids)))
                (at-enclosing-res self (lL root))
                0))
    :lT (c? (bIf (root (car (^kids)))
                (at-enclosing-res self (lT root))
                0))
    :lR (c? (bIf (root (car (^kids)))
                (at-enclosing-res self (lR root))
                0))
    :lB (c? (bIf (root (car (^kids)))
                (at-enclosing-res self (lB root))
                0))
    :kids (c? (theKids (mkPart :root (IXRootKidSized)
                           :layers nil ;; (c? (layers .parent))
                           :kids (c? (theKids (content .parent))))
                        ))))

(defmodel IXCanvasParentSized (IXCanvasNested)
  ()
  (:default-initargs
      :lL 0
    :lT 0
    :lR  (c? (inset-Width .parent))
    :lB  (c? (downs (inset-Height .parent)))
    :kids (c? (theKids (mkPart :root (IXRoot)
                           :lL 0
                           :lT 0
                           :lR  (c? (round (* (^enclosingres .parent)
                                               (inset-Width .parent))
                                            (^targetRes .parent)))
                           :lB  (c? (downs (round (* (^enclosingres .parent)
                                                   (inset-Height .parent))
                                            (^targetRes .parent))))
                           :layers :white
                           :kids (c? (theKids (^content .parent))))
                        ))))

;-------------------------------------------

(defmacro canvasRes (&optional (self 'self))
  `(targetRes (canvas ,self)))

(defmacro sCanvas ()
  `(nearest self IXCanvas))

(defun canvas (self)
  (upper self IXCanvas))

(defun at-canvas-res (self screenAmount)
  (scr2log screenAmount (targetRes (nearest self IXCanvas))))

(defun in-canvas (self point)
  (mkv2 (in-canvas-vector self (v2-h point))
           (in-canvas-vector self (v2-v point))))

(defmethod in-canvas-vector ((self IXCanvas) amount)
  (if (enclosingRes self)
      (res-to-res amount (enclosingRes self) (targetRes self))
    amount))

(defmethod in-canvas-vector (self amount)
  (declare (ignorable self))
  amount)

;-------------------------------------------

(defmethod gOffset ((self IXCanvas) &optional (accumH 0) (accumV 0))
  ;(trc "goffset self" self 'px (px self) 'py (py self) 'fm-parent (fm-parent self))
  (if (fm-parent self)
      (gOffset (fm-parent self)
               (+ (res-to-res accumH (targetRes self) (enclosingRes self))
                  (or (px self) 0))
               (+ (res-to-res accumV (targetRes self) (enclosingRes self))
                  (or (py self) 0)))
    (mkv2 accumH accumV)))

(defmodel IXRoot (IXFamily)
  ()
  (:default-initargs
      :px (c? (^px-maintain-pl 0))
      :py (c? (^py-maintain-pT 0))))

(defmodel IXRootKidSized (IXKidSized IXRoot)())


                                
                                  
(defmethod winOffset :around ((self IXCanvas) &optional (accumH 0) (accumV 0))
  ;; (trc "winoffset scaling from to" (targetres self)(cs-logical-screen-resolution))
  (call-next-method self
                    (at-enclosing-res self accumH)
                    (at-enclosing-res self accumV)))

(defmethod gUnScaled ((self IXCanvas) value) ;; fixcanvas
  (gUnScaled (fm-parent self) (res-to-res value
                                         (targetres self)
                                         (cs-logical-screen-resolution))))

(def-c-echo targetRes ((self IXCanvas))
  ;(talkdirty)
  (ix-invalidate self))

(defpackage :Cello
  (:export IXCanvas targetRes))

;;;(defmethod ix-paint :around ((self IXCanvas) gR w)
;;;  (if (or (not (enclosingRes self))
;;;          *compiling-emf*
;;;          )
;;;      (Call-next-method)
;;;    (let ((dc (device-context w))
;;;          (oldext (ff:allocate-fobject 'win:point :c))
;;;          )
;;;      ;;(trc "ipt setting viewport" self (targetres self))
;;;      (os-SetViewportExtEx dc (targetRes self) (targetRes self) oldext)
;;;      
;;;      (with-target-resolution ((targetRes self))
;;;        (call-next-method self (res-to-res gR (enclosingres self) (targetRes self)) w)
;;;        
;;;        (os-SetViewportExtEx dc
;;;                             (point-slot oldext x)
;;;                             (point-slot oldext y) ct:hnull)
;;;        (ff:free-fobject oldext)
;;;        ))))
  

#+test
(archos::pv-reset-adhoc)


