;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

; /// looks like BYKIDP has some clients that want twice the outset (IXStack.lR)
; and some that do not. I remember changing this from twice the outset to once
; to fix something. sort out someday. 210807kt
;
(defmacro BYKIDP (outsetOptr minMax attribute)
  `(progn
    (c-assert (numberp (outset self)))
     (,outsetOptr
      (if (^kids self)
         (with-dynamic-fn (roomy (kid) 
                            (not (^collapsed kid)))
           (,minMax ,attribute
            :test roomy))
         0)
      (outset self))))

(defmacro ByLastKid (attribute)
  (let ((lastKid (copy-symbol 'lastkid))
        (temp (gensym)))
    `(progn
       (c-assert (numberp (outset self))() "Outset ~a of ~a not numberp" (outset self) self)
       (,(ecase attribute (pR '+)(pB 'downs))
         (bIf (,lastKid (find-if (lambda (k)
                                   (not (^collapsed k)))
                          (^kids self)
                          :from-end t))
           (let ((,temp (,attribute ,lastKid)))
             (c-assert ,temp () "attribute ~a of ~a is nil" ',attribute ,lastKid)
             ,temp)
           0)
         (outset self)))))

(defun MinKid-lL () (c? (BYKIDP - ^fm-min-kid pL)))
(defun MaxKId-lT () (c? (BYKIDP - ^fm-max-kid pT)))
(defun MaxKid-lR () (c? (BYKIDP + ^fm-max-kid pR)))
(defun MinKid-lB () (c? (BYKIDP + ^fm-min-kid pB)))

(defun MaxKid-lWidth () (c? (+ (BYKIDP + ^fm-max-kid lWidth) (outset self))))
(defun MaxKid-lHeight () (c? (+ (outset self) (BYKIDP + ^fm-max-kid lHeight))))

;;------- IXFamily -----------------------------
;;
(defmodel IXFamily (Image Family)
   (
    (styles :initform nil :reader styles :initarg :styles)
    
    (effectiveStyles :cWhen (:DocPaint)
                                :reader effectiveStyles :initarg :effectiveStyles
                                :initform nil #+not (IXFamily-effectiveStyles))
    
    (outset :cell nil :initarg :outset :initform 0 :accessor outset)
    (showkids :initarg :showkids :initform nil :accessor showkids)
    (kidsevershown
      :initarg :kidsEverShown
      :cWhen (:DocEdit)
      :initform (c? (or .cache (^showkids)))
      :reader kidsevershown)
    )
  (:default-initargs
      :clipped nil))

(defmethod focus-starting ((self IXFamily))
  (some #'focus-find-first (kids self)))

;;-------- ZeroTL ----------------------------
;;
(defmodel IGZeroTL (IXFamily) 
   ()
   (:default-initargs
    :ll (c? (- (outset self))) 
    :lt (c? (ups (outset self))) 
    :lr (maxKid-lR) 
    :lb (minKid-lB)
    :kid-slots (def-kid-slots
                   (mk-kid-slot (px :ifMissing t)
                     (c? (^px-maintain-pL 0)))
                   (mk-kid-slot (py :ifMissing t)
                     (c? (^py-maintain-pT 0))))))

(defmodel IXKidSized (IXFamily) 
    ()
    (:default-initargs
     :ll (minkid-ll) :lt (maxkid-lt)
     :lr (maxkid-lr) :lb (minkid-lb)))

;----------- OfKids -----------------------
;
(defmacro SMKIDP (outsetOptr minMax attribute)
  `(c? (,outsetOptr
         (if (^kids self)
           (with-dynamic-fn (roomy (kid) (not (^collapsed kid)))
             (,minMax ,attribute
                      :test roomy))
           0)
         (outset self))))



(defun v2-in-subframe (super h v sub)
  (if (eql super sub) ;; bingo
      (values h v)
    (dolist (kid (kids super))
      (multiple-value-bind (subH subV)
          (v2-in-subframe kid h v sub)
        (when subH
          (return-from v2-in-subframe (values (- subH (px kid))
                                              (- subV (py kid)))))))))

;-------------------------- IMMatrix ------------------------------------------

(defmodel IMMatrix (IGZeroTL)
  ((columns :cell nil :initarg :columns :initform nil :accessor columns)
   (indenthz :cell nil :initarg :indenthz :initform 0 :accessor indenthz)
   (spacinghz :cell nil :initarg :spacinghz :initform 0 :accessor spacinghz)
   (spacingvt :cell nil :initarg :spacingvt :initform 0 :accessor spacingvt))
  (:default-initargs
      :kid-slots (lambda (self)
                   (declare (ignore self))
                   (list
                    (mk-kid-slot (px)
                                 (c? (let ((parent (fm-parent self)))
                                       (+ (indenthz parent)
                                          (if (zerop (mod (fm-pos self)
                                                          (or (columns parent)
                                                              (length (kids parent)))))
                                              0
                                            (+ (spacingHz parent)
                                               (pR (find-prior self (kids parent)))))))))
                    (mk-kid-slot (py)
                                 (c? (let* ((parent (fm-parent self))
                                            (pSib (find-prior self (kids parent))))
                                       (if (and pSib (columns parent))
                                           (if (zerop (mod (fm-pos self) (columns parent)))
                                               (+ (- (abs (spacingVt parent))) (pB psib))
                                             (pT psib))
                                         0))))))))

