;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

;;; (defclass Helper ()
;;;   ((aa-host :initform nil :initarg :host :accessor host)))

(eval-when (:compile-toplevel :load-toplevel :execute)

 (defstruct (ACStyle
             (:conc-name nil))
   styleName
   textFace
   textStyle
   textsizes
   textsize
   fgcolor
   bkcolor)

 (defun make-style (stylename &rest styleArgs)
   (c-assert (not (listp stylename)))
   (apply #'make-ACStyle
          :styleName stylename ;; ///settle which
          styleargs)))

(let (stylesDefault)

  (defun styles-default-clear ()
    (setf stylesdefault nil))

  (defun styles-default ()
    (or stylesDefault
        (setf stylesDefault
                (list (make-style :tn-button 
                                  :textFace '|Arial|
                                  :textSizes (uCvt '(8 10 12 18 45) :points)
                                  :textStyle '(:bold)
                                  :fgColor :white
                                  :bkColor :transparent
                                  )
                      
                      (make-style :normal 
                                :textFace '|Times New Roman|
                                :textSizes (uCvt '(10 12 14 16 18) :points)
                                :textStyle '(:plain)
                                :fgColor :black
                                :bkColor :transparent
                                )
;;;                       (make-style :label 
;;;                                   :textFace '|Arial|
;;;                                   :textSizes (uCvt '(10 12 16 24 45) :points)
;;;                                   :textStyle '(:plain)
;;;                                   :fgColor :black
;;;                                   :bkColor :transparent
;;;                                   )
;;;                       (make-style :header ;; not a bad place to segue off another style ala :derivedFrom :label
;;;                                   :textFace '|Arial|
;;;                                   :textSizes (uCvt '(10 12 14 16 18) :points)
;;;                                   :textStyle '(:plain)
;;;                                   :fgColor :black
;;;                                   :bkColor :transparent
;;;                                   )
                      (make-style :label 
                                  :textFace '|Times New Roman|
                                  :textSizes (uCvt '(8 10 12 14 16) :points)
                                  :textStyle '(:plain)
                                  :fgColor :black
                                  :bkColor :transparent
                                  )
                      (make-style :header ;; not a bad place to segue off another style ala :derivedFrom :label
                                  :textFace '|Times New Roman|
                                  :textSizes (uCvt '(10 12 14 16 18) :points)
                                  :textStyle '(:plain)
                                  :fgColor :black
                                  :bkColor :transparent
                                  )

                      (make-style :panelTitle
                                  :textFace '|Lucida Sans Unicode| 
                                  :textSizes (uCvt '(8 10 12 14 16) :points)
                                  :textStyle '(:plain)
                                  :fgColor :black
                                  :bkColor :transparent
                                  )
                      (make-style :button
                                  :textFace '|MS Sans Serif|
                                  :textSizes (uCvt '(10 10 10 10 10) :points)
                                  :textStyle '(:plain)
                                  :fgColor :black
                                  :bkColor :transparent
                                  )
                      (make-style :data
                                  :textFace '|Courier New|
                                  :textSizes (uCvt '(12 16 20 24 28) :points)
                                  :textStyle '(:plain)
                                  :fgColor :black
                                  :bkColor :transparent)
                      (make-style :datawritten
                                  :textFace '|Courier New|
                                  :textSizes (uCvt '(10 12 18 20 24) :points)
                                  :textStyle '(:plain)
                                  :fgColor :black
                                  :bkColor :transparent)
                      (make-style :dataDirectEntry
                                  :textFace '|Lucida Console|
                                  :textSizes (uCvt '(9 10 11 12 14) :points)
                                  :textStyle '(:plain)
                                  :fgColor :black
                                  :bkColor :transparent)
                      )))))

(defmodel IXStyled ()
  (
   (style :cWhen (:DocPaint) :initarg :style
     :initform :default
     :reader Style)
   
   (textFont :cWhen (:DocPaint) 
             :reader textFont :initarg :textFont
             :initform (c? (ix-style-font (^style))))))

(DEFMETHOD ix-style-font (other)
  (declare (ignore other))
  #+celloglut (ftgl-font-ensure .w. 'texture 'sylfaen 12)
  (make-font-glut-stroke :glutID GLUT_STROKE_ROMAN))

(DEFMETHOD ix-style-font ((id (eql :button)))
  #+celloglut
  (ftgl-font-ensure .w. 'texture 'arialn 14)
  (make-font-glut-stroke
   :glutID GLUT_STROKE_ROMAN
   :font-size 24))

(DEFMETHOD ix-style-font ((id (eql :label)))
  #+celloglut
  (ftgl-font-ensure .w. 'texture 'arialn 18)
  (make-font-glut-stroke
   :glutID GLUT_STROKE_ROMAN
   :font-size 24))

(defun ix-string-width (self string)
  (c-assert (sCanvas) () "~a not contained by any canvas" self)
  (font-string-width (targetRes (sCanvas)) (textFont self) string))


;--------------------------------------------------------

(defun styles-or (styles superStyles)
  (cond
   ((null styles) superstyles)
   ((null superstyles) styles)
   (t
    (let (effectives)
      (dolist (style styles)
        (c-assert style)
        (push (bIf (superStyle (find (styleName style) superStyles :key #'styleName))
                 (make-style-customizing superstyle
                                         :textface (textface style)
                                         :textStyle (textStyle style)
                                         :textSize (textSize style)
                                         :textSizes (textSizes style)
                                         :fgColor (fgColor style)
                                         :bkColor (bkColor style))
                 
                 style)
              effectives))
      
      (dolist (superStyle superStyles)
        (pushnew superStyle effectives :key #'styleName))
      
      effectives))))

(defun make-style-customizing (superstyle &key textFace textStyle textsizes textsize fgcolor bkcolor)
  (c-assert superstyle)
  (make-Style (stylename superstyle)
              :textface (or textface (textface superStyle))
              :textStyle (or textStyle (textStyle superStyle))
              :textSize (or textSize (textSize superStyle))
              :textSizes (or textSizes (textSizes superStyle))
              :fgColor (or fgColor (fgColor superStyle))
              :bkColor (or bkColor (bkColor superStyle))))




