;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

(defun lRect (ap)
  (count-it :lrect)
   (mkR (lL ap) (lT ap) (lR ap) (lB ap)))

;---------- gOffset -------------------
;
(defun offset-within (inner outer &optional dbg)
  (declare (ignorable dbg))
  (trc nil "offset-within inner outer" inner outer)
  (do (
       (offsetH 0 (progn
                    (trc nil "offset-within delta-h, from" from (px from))
                    (incf offsetH (px from))))
       (offsetV 0 (incf offsetV (py from)))
       (from inner (fm-parent from)))
      ((or (null from)
           (null outer)
           (eql from outer)) (eko (nil "offset-within returns")
                                  (mkv2 offsetH offsetV)))))

(defun offset-within2 (inner outer)
  (do (
       (offsetH 0 (incf offsetH (px from)))
       (offsetV 0 (incf offsetV (py from)))
       (from inner (fm-parent from)))
      ((or (null from)
           (null outer)
           (eql from outer)) (mkv2 offsetH offsetV))
    ;(trc "inner outer" inner outer)
    ))

(defmacro ^offset-within (inner outer)
  (let ((offsetH (gensym)) (offsetV (gensym)) (from (gensym)))
     `(let ((,offsetH 0)
            (,offsetV 0))
         (do ((,from ,inner (fm-parent ,from)))
             ((or (null ,from)
                  (eql ,from ,outer))
              ;
              (mkv2 ,offsetH ,offsetV))
           
           (incf ,offsetH (^px ,from))
           (incf ,offsetV (^py ,from))))))

(defun mk-gR (ap)
   (c-assert ap)
  (count-it :mk-gr)
  (let ((gOffset (gOffset ap))) ;; /// wastes a v2
    (nr-offset (mkR (lL ap) (lT ap) (lR ap) (lB ap)) (v2-h gOffset) (v2-v gOffset))))

;sum pXYs up the family tree    ;gave an odd result for cursor display....

(defun v2-xlate (outer inner outerv2)
  (if (eql outer inner)
     outerv2
     (v2-xlate outer (fm-parent inner)
               (v2-subtract outerv2
                            (mkv2 (px inner) (py inner))))))

(defun h-xlate (outer inner outerH)
  (if (eql outer inner)
     outerh
     (h-xlate outer (fm-parent inner)
               (- outerH (px inner)))))

(defun v-xlate (outer inner outerV)
  (if (eql outer inner)
     outerV
     (v-xlate outer (fm-parent inner)
               (- outerV (py inner)))))

(defmethod gOffset (self &optional (accumH 0) (accumV 0))
  (declare (ignorable self))
  (mkv2 accumH accumV))

(defmethod winOffset (self &optional (accumH 0) (accumV 0))
  (declare (ignorable self))
  (mkv2 accumH accumV))

(defun gOffsetH (ap)
   (v2-H (gOffset ap)))

(defun gOffsetV (ap)
     (v2-V (gOffset ap)))

(defun gBox (ap)
  (count-it :gBox)
  (if (c-stopped)
      (trc "gbox sees stop" ap)
    (progn
      (c-assert ap)
      (let* ((gOffset (gOffset ap))
             (oh (v2-h gOffset)))
        (c-assert (typep goffset 'v2))
        (c-assert (numberp oh))
        (c-assert (numberp (lR ap)))
        (let ((r (nr-offset (mkR (lL ap) (lT ap) (lR ap) (lB ap)) oh (v2-v gOffset))))
          (c-assert (numberp (r-left r)))
          (c-assert (numberp (r-top r)))
          (c-assert (numberp (r-right r)))
          (c-assert (numberp (r-bottom r)))
          r)))))

;____________________________________________

(defun pL (self) (+ (px self) (lL self)))
(defun pR (self) (+ (px self) (lR self)))
(defun pT (self) (+ (py self) (lT self)))
(defun pB (self)
  (c-assert (lB self))
  (c-assert (py self))
  (+ (py self) (lB self)))

(defun pXY (self)
  (mkv2 (px self) (py self)))

;--------------------------------------------------------


(defun lWidth (i)
  (c-assert (lr i))
  (c-assert (ll i))
  (- (lR i) (lL i)))

(defun lHeight (i)
   (abs (- (lB i) (lT i))))

;;-----------------------------------------------

(defun inset-Width (image)
   (- (lWidth image) (outset image) (outset image)))

(defun inset-LR (image)
   (- (lR image) (outset image)))

(defun inset-LB (image)
   (- (lB image) (outset image)))

(defun inset-Height (image)
   (- (lHeight image) (outset image) (outset image)))

;---------------------------------

(defmacro ^MinKid-pL () `(BYKIDP - ^fm-min-kid pL))
(defmacro ^MinKId-pT () `(BYKIDP - ^fm-min-kid pT))
(defmacro ^MaxKid-pR () `(BYKIDP + ^fm-max-kid pR))
(defmacro ^MaxKid-pB () `(BYKIDP + ^fm-max-kid pB))
(defmacro ^MaxKid-lWidth () `(BYKIDP + ^fm-max-kid lWidth))
(defmacro ^MaxKid-lHeight () `(BYKIDP + ^fm-max-kid lHeight))

;---------------------------------

(defmacro ^lL-Width (width)
     `(- (^lR self) ,width))

(defmacro ^lR-Width (width)
     `(+ (^lL self) ,width))

(defmacro ^lT-Height (height)
     `(- (^lB self) ,height))

(defmacro ^lB-Height (height)
     `(+ (^lT self) ,height))

;----------------------------------

(defmacro ^ll-maintain-pL (pl)
     `(- ,pL (^px self)))

(defmacro ^lr-maintain-pr (pr)
     `(- ,pr (^px self)))

(defmacro ^fill-right (upperType &optional (padding 0))
  `(call-^fillRight self (upper self ,upperType) ,padding))

;recalc local top based on pT and offset
(defmacro ^lt-maintain-pT (pT)
     `(- ,pT (^py self)))

;recalc local bottom based on pB and offset
(defmacro ^lb-maintain-pB (pB)
     `(- ,pB (^py self)))

;--------------
;recalc offset based on p and local 
(defmacro ^px-maintain-pL (pL)
  (let ((lL (gensym)))
     `(- ,pL (let ((,lL (^lL self)))
               (c-assert ,lL () "^px-maintain-pL sees nil lL for ~a" self)
               ,lL))))

(defmacro ^px-maintain-pR (pR)
  `(- ,pR (^lR self)))

(defmacro ^py-maintain-pT (pT)
  `(- ,pT (^lT self)))

(defmacro ^py-maintain-pB (pB)
  `(- ,pB (^lB self)))

(defmacro ^centeredH? ()
  `(c? (^px-maintain-pL (round (- (lWidth .parent) (lWidth self)) 2))))

(defmacro ^centeredV? ()
  `(c? (^py-maintain-pT (round (- (lHeight .parent) (lHeight self)) 2))))

(defmacro ^fillDown (upperType &optional (padding 0))
  (let ((filled (gensym)))
    `(let ((,filled (upper self ,upperType)))
       #+qt (trc "^fillDown sees filledLR less offH"
                 (^lB ,filled)
                 ,padding
                 (v2-v (^offset-within self ,filled)))
       (- (^lB ,filled)
          ,padding
          (v2-v (^offset-within self ,filled))))))

(defmacro ^lBMax? (&optional (padding 0))
  `(c? (^lB-maintain-pB (- (inset-lB .parent)
                            ,padding))))

(defmacro ^lRMax? (&optional (padding 0))
  `(c? (^lR-maintain-pR (- (inset-lR .parent)
                            ,padding))))


(defmacro ^priorSib (self)
   (let ((kid (gensym)))
      `(let* ((,kid ,self))
          (find-prior ,kid (^kids (fm-parent ,kid))))))

(defmacro ^priorSib-pB (self &optional (spacing 0))
   (let ((kid (gensym))
         (pSib (gensym)))
      `(let* ((,kid ,self)
               (,pSib (find-prior ,kid (^kids (fm-parent ,kid))
                        :test (lambda (sib)
                                (not (^collapsed sib)))))
               )
          ;(trc "^priorSib-pb > kid, sib" ,kid ,pSib)
          (if ,pSib
              (+ (- (abs ,spacing)) ;; force spacing to minus(= down for OpenGL)
                (pB ,psib))
            0))))

(defmacro ^priorSib-pT (self &optional (spacing 0))
   (let ((kid (gensym))
         (pSib (gensym)))
      `(let* ((,kid ,self)
              (,pSib (find-prior ,kid (^kids (fm-parent ,kid)))))
          ;(trc "^priorSib-pb > kid, sib" ,kid ,pSib)
          (if ,pSib
             (+ (- (abs ,spacing)) (pT ,psib))
             0))))

; "...return the sib's pL [if ,alignment is :left] or pR, plus optional spacing"

(defmacro ^priorSib-pR (self &optional (spacing 0) alignment)
   (let ((kid (gensym))
         (pSib (gensym)))
      `(let* ((,kid ,self)
              (,pSib (find-prior ,kid (^kids (fm-parent ,kid)) :test (lambda (k) (not (^collapsed k))))))
          (if ,pSib
             (case ,alignment
               (:left (+ ,spacing (pL ,psib)))
               (otherwise (+ ,spacing (pR ,psib))))
             0))))

(defmacro ^px-stay-right-of (other &key (by '0))
   `(^px-maintain-pL (+ (pR (fm-other ,other)) ,by)))

; in use; adjust offset to maintain pL based on ,justify
(defmacro ^px-self-centered (justify)
   `(^px-maintain-pL
     (ecase ,justify
       (:left 0)
       (:center (floor (- (inset-Width .parent) (lWidth self)) 2))
       (:right (- (inset-LR .parent) (lWidth self))))))

; in use; same idea for pT
(defmacro ^py-self-centered (justify)
   `(^py-maintain-pT
     (ecase ,justify
       (:top  0)
       (:center (floor (- (inset-Height .parent) (lHeight self)) -2))
       (:bottom (- (inset-Height .parent) (lHeight self))))))

(defmacro ^py-stay-below (other &key by)
     (if by
        `(+ (pB (fm-other ,other)) ,by)
        `(pB (fm-other ,other))))

(defmacro ^bottom-aligned (other &key (off-by 0))
  `(+ (pB (fm-other ,other)) ,off-by))

(defmacro ^left-aligned (other &key (off-by 0))
  `(+ (pL (fm-other ,other)) ,off-by))

(defmacro ^as-left-as-kids ()
     `(^fm-min-kid pL))

(defmacro ^as-high-as-kids ()
     `(^fm-min-kid pT))

(defmacro ^as-low-as-kids ()
     `(^fm-max-kid pB))

(defmacro ^as-low-as-parent ()
     `(inset-lB (fm-parent self)))

(defmacro ^as-right-as-kids ()
     `(^fm-max-kid pR))

(defmacro ^fillParentRight (&optional (inset 0))
  `(^lR-maintain-pR (- (inset-LR .parent) ,inset)))

(defmacro ^fillParentDown ()
  `(^lB-maintain-pB (inset-LB .parent)))

