;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

(defmacro uGrid () `(fm-parent self))

(defun IXGrid-lR ()
  (c? (col-pR self (1- (^colCt)))))

(defun IXGrid-lB ()
  (c? (row-pB self (cell-row (last1 (^kids))))))

(defmodel IXGrid (IGZeroTL)
  ((colct :initarg :colct :initform nil :accessor colct)
   (rowct :initarg :rowct :initform nil :accessor rowct)
   ;
   (colPadding :cell nil :initarg :colPadding :initform 0 :reader colPadding)
   (rowPadding :cell nil :initarg :rowPadding :initform 0 :reader rowPadding)
   ;
   (allcellwidth :initarg :allcellwidth :initform nil :reader allcellwidth)
   (allcellheight :initarg :allcellheight :initform nil :reader allcellheight)
   (allCellFrame :cell nil :initarg :allCellFrame :initform nil :reader allCellFrame)
   ;
   (rowOffsets :initarg :rowOffsets :initform (IXGrid-rowOffsets) :reader rowOffsets)
   (colOffsets :initarg :colOffsets :initform (IXGrid-colOffsets) :reader colOffsets)
   ;
   (rowJustifys :initarg :rowJustifys :initform :center :reader rowJustifys)
   (colJustifys :initarg :colJustifys :initform :left :accessor colJustifys)
   )
  (:default-initargs
      :lL (c? (- (outset self)))
      :lT (c? (- (outset self)))
      :lB (IXGrid-lb)
      :lR (IXGrid-lr)
      :clipped t
    :colCt (c? (ceiling (length (^kids)) (^rowCt)))
    :rowCt (c? (ceiling (length (^kids)) (^colCt)))
    :kid-slots (lambda (self)
                  (declare (ignore self))
                  (list
                   (mk-kid-slot (px)
                     (make-IXGrid-px))

                   (mk-kid-slot (py)
                     (c? (let ((justify (rowJustifys (uGrid))))
                           (case (if (listp justify)
                                     (elt justify (cell-row self))
                                   justify)
                             (:top (^py-maintain-pT (cell-pT self)))
                             (:center (^py-maintain-pT
                                       (+ (cell-pT self)
                                         (floor (- (cell-height self) (lheight self)) 2))))
                             (:bottom (^py-maintain-pB (cell-pB self)))
                             (otherwise 0)))))))))

(defun row-pB (grid rowN)
  (- (elt (rowOffsets grid) (1+ rowN))
      (rowPadding grid)))

(defun col-pR (grid colN)
  (- (elt (colOffsets grid) (1+ colN))
           (colPadding grid)))

(defun cell-pL (self)
  (elt (colOffsets (fm-parent self)) (cell-col self)))

(defun cell-pR (self)
  (col-pR (uGrid) (cell-col self)))

(defun cell-width (self)
  (- (cell-pR self) (cell-pL self)))

(defun cell-pT (self)
  (elt (rowOffsets (fm-parent self)) (cell-row self)))

(defun cell-pB (self)
  (row-pB (uGrid) (cell-row self)))

(defun cell-height (self)
  (- (cell-pB self) (cell-pT self)))

(defmodel IXGridUniform (IXGrid)
  ()
  (:default-initargs
      :allCellWidth (c? (round (lWidth self) (^colct self)))
    :allCellHeight (c? (round (lHeight self) (^rowct self)))
    :colJustifys :left
    :rowJustifys :top))
                   
(defmodel IXGridFV (family-values IXGrid)())
(defmodel IXGridUniformFV (IXGridUniform IXGridFV)())

;
;---------- painting --------------------------

;;;(defmethod ix-paint :after ((self IXGrid) gR w)
;;;  (bWhen (frame (allCellFrame self))
;;;     (let ((cellR #.(mkR 0 0 0 0)))
;;;       (dotimes (rowNo (1- (length (rowOffsets self))))
;;;         (doTimes (colNo (1- (length (colOffsets self))))
;;;           (let ((left (elt (colOffsets self) colNo))
;;;                 (top (elt (rowOffsets self) rowNo))
;;;                 (right (elt (colOffsets self) (1+ colNo)))
;;;                 (bottom (elt (rowOffsets self) (1+ rowNo)))
;;;                 )
;;;             (nr-make cellR left top right bottom)
;;;             (nr-move cellR (r-top-left gR))
;;;             (trc nil "grid R" self rowno colno cellR)
;;;             (ix-paint frame cellR w)))))))

;
;---------- geometry ----------------------------

(defun IXGrid-colOffsets ()
  (c? (let* ((colCt (^colCt self))
              (offsets (make-array (1+ colCt)
                                 :element-type 'fixnum
                                 :initial-element (or (^allCellWidth) 0))))

         (unless (^allCellWidth)
           (dolist (kid (^kids self))
             (maxf (elt offsets (mod (kid-no kid) colCt)) (lWidth kid))))

         (let ((offset 0) thisSize)
            (dotimes (colNo (1+ colCt))
              (setf thisSize (elt offsets colNo))
              (setf (elt offsets colNo) offset)
              (incf offset (+ (colPadding self) thisSize))))

         offsets)))

(defun IXGrid-rowOffsets ()
  (c? (let* ((colCt (^colCt self))
              (rowCt (ceiling (length (^kids self)) colCt))
              (offsets (make-array (1+ rowCt)
                                 :element-type 'fixnum
                                 :initial-element (or (^allCellHeight) 0))))

         (unless (^allCellHeight)
           (dolist (kid (^kids self))
             (maxf (elt offsets (floor (kid-no kid) colCt)) (lHeight kid))))

         (let ((offset 0) thisSize)
            (dotimes (rowNo (1+ rowCt))
              (setf thisSize (elt offsets rowNo))
              (setf (elt offsets rowNo) offset)
              (incf offset (+ (rowPadding self) thisSize))))
         offsets)))

(defun cell-row (self)
  (floor (kid-no self) (colct (uGrid))))

(defun cell-col (self)
  (mod (kid-no self) (colct (uGrid))))

(defun cell-sib (self rowNo colNo)
  (elt (kids .parent) (+ (* rowNo (colct (uGrid))) colno)))

(defun col-head (self)
  (cell-sib self 0 (cell-col self)))

(defun row-head (self)
  (cell-sib self (cell-row self) 0))


(defun make-IXGrid-px ()
  (c? (let ((justify (colJustifys (uGrid))))
        (trc nil "coljustifys ugrid" (uGrid) (colJustifys (uGrid))(cell-col self))
         (case (if (listp justify)
                   (elt justify (cell-col self))
                 justify)
           (:left (^px-maintain-pL (cell-pl self)))
           (:center (^px-maintain-pL
                          (+ (cell-pl self)
                             (floor (- (cell-width self) (lWidth self)) 2))))
           (:right (^px-maintain-pR (cell-pR self)))
           (otherwise (trc "woohoo bad justify" (if (listp justify)
                                                    (elt justify (cell-col self))
                                                  justify)(cell-col self)justify)
                      0)))))



(defun elt1 (list-or-atom n)
   (trc "elt1 > la, latyp" list-or-atom (type-of list-or-atom))
   (if (listp list-or-atom)
      (elt list-or-atom n) list-or-atom))

(defun grid-cell-ref (grid rowNo colNo)
  (elt (kids grid) (+ (* rowNo (colCt grid)) colNo)))


