;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

;--------------- IGInline -----------------------------
;

(defmodel IGInline (IGZeroTL)
  ((justify :cell nil :initarg :justify :initform 0 :accessor justify)
   (spacing :cell nil :initarg :spacing :initform 0 :accessor spacing)
   (inlineCount :cell nil :initarg :inlineCount :reader inlineCount)))

;--------------- Stacks ------------------------------
;


(defmodel IXStack (IGInline)
  ()
  (:default-initargs
      :lR  (c? (^lr-width (+ (BYKIDP + ^fm-max-kid lWidth)
                         (outset self)))) ;; 210807kt uh-oh, see BYKIDP commnet
    :lB  (c? (ByLastKid pB))
    :justify :left
    :kid-slots (lambda (self)
                 (declare (ignore self))
                 (list
                  #+WHATTHE? (mk-kid-slot (lB :ifmissing t)
                     (c? (+ (r-left .parent)
                            (ecase (justify .parent)
                              (:left 0)
                              (:center (floor (- (lwidth .parent) (lWidth self)) 2))
                              (:right (- (lwidth .parent) (lWidth self)))))))
                  (mk-kid-slot (px :ifMissing t)
                    (c? (^px-self-centered (justify .parent))))
                  (mk-kid-slot (py)
                    (c? (^py-maintain-pT
                         (^priorSib-pB self (spacing .parent)))))))))

(defmodel IXStackOfKids (IXStack)
  ()
  (:default-initargs
   :lL (c? (BYKIDP - ^fm-min-kid pL))
   :lR (c? (BYKIDP + ^fm-max-kid pR))
   :lB (c? (ByLastKid pB))
    :justify :left))

(defmacro aStack ((&rest stackArgs) &body ddkids)
  `(mkPart ,(copy-symbol 'aStack) (IXStack)
      ,@stackArgs
      :kids (c? (packed-flat! ,@ddKids))))

(defmacro aStackOfKids ((&rest stackArgs) &body ddkids)
  `(mkPart ,(copy-symbol 'aStack) (IXStackOfKids)
      ,@stackArgs
      :kids (c? (packed-flat! ,@ddKids))))



;----------------------- IXRow ------------------------------
;

(defun ups (&rest values)
  (apply '+ values))

(defun ups-most (&rest values)
  (apply 'max values))

(defun downs (&rest values)
  (apply '- values))

(defun downs-most (&rest values)
  (apply 'min values))

(defun ins (&rest values)
  (apply '+ values))

(defun outs (&rest values)
  (apply '- values))


(defmodel IXRow (IGZeroTL)
   ((justifyVT :cell nil :initarg :justify :initform :top :accessor justifyVT)
    (spacingHz :cell nil :initarg :spacing :initform 0 :accessor spacingHz))
   (:default-initargs
    :lL  (c? (- (outSet self)))
    :lT  (c? (ups (outSet self)))
    :lB  (c? (downs (outset self) (^lB-height (^fm-max-kid lHeight))))
    :lR  (c? (ByLastKid pR))
    :kid-slots (lambda (self)
                 (declare (ignore self))
                 (list
                  (mk-kid-slot (py :ifMissing t)
                    (c? (^py-self-centered (justifyVt .parent))))
                  (mk-kid-slot (px)
                    (c? (^px-maintain-pL
                         (^priorSib-pR self (spacingHz .parent)))))))))

(defmodel IXRowFV (family-values IXRow)())
(defmodel IXStackFV (family-values IXStack)())
;--------------- IGRowFlow ----------------------------

(defmodel IGRowFlow (IXRow)
  ((spacingHz :cell nil :initarg :spacingHz :initform 0 :reader spacingHz)
   (spacingVT :cell nil :initarg :spacingVt :initform 0 :reader spacingVt)
   (aligned :cell nil :initarg :aligned :initform nil :reader aligned))
  (:default-initargs
   :lB  (minKid-lB)
    :kid-slots (lambda (self)
                 (declare (ignore self))
                 (list
                  (mk-kid-slot (py)
                    (c? (^py-maintain-pT
                         (let ((pH (^priorSib-pR self (spacingHz .parent) (aligned .parent))))
                           (if (> (+ pH (lWidth self)) (lWidth .parent))
                               (^priorSib-pB self (spacingVT .parent))
                             (^priorSib-pT self))))))
                  (mk-kid-slot (px)
                    (c? (^px-maintain-pL
                         (let ((pH (^priorSib-pR self (spacingHz .parent) (aligned .parent))))
                           (if (> (+ pH (lWidth self)) (lWidth .parent))
                               0
                             pH)))))))))

(defmacro aRow ((&rest rowArgs) &body ddkids)
  `(mkPart ,(copy-symbol 'aRow) (IXRow)
      ,@rowargs
      :spacing 0
      :kids (c? (packed-flat! ,@ddKids))))

