;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

(defun remove-keys (kargs keys)
  (c-assert (evenp (length kargs)) () "Argument list not even length: ~a" kargs)
  (when kargs
    (if (member (car kargs) keys)
       (remove-keys (cddr kargs) keys)
       (list* (car kargs) (cadr kargs) (remove-keys (cddr kargs) keys)))))

(defmacro mkLabel (text &rest keyArgPairs)
  `(cells::make-part (gensym "MKLABEL") 'IXText
      ,@keyArgPairs
      :style :label
      :text$ ,text))
        
(defMacro aLabel (label$  &rest keyArgPairs)
  `(mkLabel ,label$ ,@keyArgPairs))
            
(defmacro aNumLabel (label$  &rest keyArgPairs)
  `(mkLabel ,label$ ,@keyArgPairs :justifyHZ :right))
            
(defmacro mkGUILabel (text &rest keyArgPairs)
  `(mkPart :md-value (IXText)
      ,@keyArgPairs
      :style :label
      :text$ ,text))
                    
(defmacro aLabelTall (text &body keyArgPairs)
  `(mkPart :labelTall (IXTextTall)
      ,@keyArgPairs
      :text$ ,text))
                    
(defmacro mkRow (fmName (&rest keyArgPairs) &body parts)
  `(mkPart ,fmName (IXRow)
     ,@keyArgPairs
     :kids (c? (list ,@parts))))

;;; (acRow (rowField :px 0 :py 10) (mp 1) (mp 2))

(defmacro mkStack (fmName (&rest keyArgPairs) &body parts)
  `(mkPart ,fmName (IXStack)
     ,@keyArgPairs
     :kids (c? (packed-flat! ,@parts))))

