;; -*- mode: Lisp; Syntax: Common-Lisp; Package: ogb; -*-
;;;
;;; Copyright  1995,2003 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(defmethod ix-align-text (self (font ftgl-pixmap))
  (ecase (justifyhz self)
    (:left)
    (:center
     (gl-raster-pos3i (round (- (lwidth self) (^text-width)) 2) 0 0))
    (:right
     (gl-raster-pos3i (- (lwidth self) (v2-h (inset self)) (^text-width)) 0 0))))

(defmethod ix-align-text (self (font ftgl-bitmap))
  (ecase (justifyhz self)
    (:left)
    (:center
     (gl-raster-pos3i (round (- (lwidth self) (^text-width)) 2) 0 0))
    (:right
     (gl-raster-pos3i (- (lwidth self) (v2-h (inset self)) (^text-width)) 0 0))))

(defmethod ix-render-in-font :around ((font ftgl) self)
  (bwhen (t$ (display-text$ self))
    (ix-align-text self font)
    
    (if (cello-get-boolean GL_CURRENT_RASTER_POSITION_VALID)
        (call-next-method)
      (trc nil "rasterpos offscreen" self :goffset (gOffset self)))))

(defmethod ix-render-in-font ((font ftgl) self)
  (count-it :render-in-font)
  (uffi:with-cstring (cs (display-text$ self))
    (fgc-render (ftgl-get-display-font font) cs)))

(defmethod ix-render-in-font ((font ftgl-outline) self)
  (let* ((t$ (display-text$ self)))
    (gl-Push-Attrib (+ GL_ENABLE_BIT GL_HINT_BIT GL_LINE_BIT GL_COLOR_BUFFER_BIT))
    
    (gl-Disable GL_TEXTURE_2D)
    
    (gl-Enable GL_LINE_SMOOTH)
    (gl-Hint GL_LINE_SMOOTH_HINT GL_DONT_CARE)
    (gl-Enable GL_BLEND)
    (gl-Blend-Func GL_SRC_ALPHA GL_ONE_MINUS_SRC_ALPHA)

    (let ((df (ftgl-get-display-font font)))
      (uffi:with-cstring (cs t$)
        (trc nil "texture render" t$ df (ftgl-face font) (ftgl-size font))
        (fgc-render df cs)))

    (gl-Pop-Attrib)
    ))

(defmethod ix-render-in-font ((font ftgl-texture) self)
  (let* ((t$ (display-text$ self)))
    (trc nil "ix-render-in-font ftgl-texture" :pxy (pxy self) (lrect self))
    
    (gl-enable GL_TEXTURE_2D)
    (gl-enable GL_BLEND)
    (gl-blend-func GL_SRC_ALPHA GL_ONE_MINUS_SRC_ALPHA)
    (gl-polygon-mode GL_FRONT_AND_BACK GL_FILL)

    (when (zoom self)
      (apply 'gl-Scalef (zoom self)))

    (when (rotation self)
      (apply 'gl-rotatef (rotation self)))
    
    (let ((df (ftgl-get-display-font font)))
      (uffi:with-cstring (cs t$)
        (trc nil "texture render" t$ df (ftgl-face font) (ftgl-size font))
        (fgc-render df cs)))

;;;    (gl-disable GL_TEXTURE_2D)
;;;    (gl-disable GL_BLEND)
    ))

(defmethod not-to-be :after ((w window))
  (loop for font-entry in (w-fonts w)
        for ff = (cdr font-entry)
        do (bwhen (cfont (ftgl-ifont ff))
             (trc nil "freeing ff" ff cfont)
             (fgc-free cfont))))