;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

;===========================================================

(defmodel IXText (IXStyled Image)
  (
   (text$ :initform nil :initarg :text$ :accessor text$)

   (displaytext$ :initform (c? (remove-if-not #'graphic-char-p (^text$)))
     :initarg :displaytext$ :accessor displaytext$)
   
   (text-width :initarg :text-width :accessor text-width
     :initform (c? (bwhen (t$ (^displaytext$))
                     (ix-string-width self t$))))

   (charMask :cell nil :initarg :charmask
     :initform nil
     :reader charMask)
   
   (maxcharwidth :initarg :maxcharwidth
     :initform (c? (ix-string-width self "M"))
     :accessor maxcharwidth)
   
   (wrapp :initform nil :accessor wrapp :cell nil :initarg :wrapp)
   
   (justifyhz :initform :left
     :accessor justifyhz
     :initarg :justifyhz)
   
   (justifyvt :initform :top
     :accessor justifyvt :initarg :justifyvt)
   
   (inset :cell nil :initarg :inset :initform (mkv2 0 0))
   (lL :initform (c? (- (inset-h self))))
   (lT :initform (c? (+ (font-ascent (textFont self))
                       (inset-v self))))
   (lR :initform (c? (^lr-width (+ (cond
                                    ((charmask self) (ix-string-width self (charmask self)))
                                    ((^text-width))
                                    ((ix-string-width self (charmask self))))
                                  (* 2 (inset-h self))))))
   (lB :initform (c? (+ (font-descent (textFont self))
                       (- (inset-v self)))))
   )
  (:default-initargs
      :lighting :off
      :layers (list +white+ :fill-lr +black+)))

(defmethod display-text$ ((self IXText))
  (text$ self))

(defmethod ix-render ((self IXText))
  (when (display-text$ self)
    (ix-render-in-font (textfont self) self)))


(defmethod ix-align-text (self font)
  (declare (ignorable font))
  (flet ((hxs ()
           (- (lwidth self) (* 2 (v2-h (inset self))) (^text-width)))
         (vxs ()
           (- (lheight self) (* 2 (v2-h (inset self))) (font-height font))))
    (gl-translatef
     (ecase (justifyhz self)
       (:left 0)
       (:center (/ (hxs) 2.0))
       (:right (hxs)))
     (ecase (justifyvt self)
       (:top 0)
       (:center (downs (/ (vxs) 2.0)))
       (:bottom (downs (vxs))))
     0)))

(defmethod ix-render-in-font ((font font-glut-bitmapped) self)
  (bwhen (t$ (^displaytext$))
    (let* ((w (text-width self))
           (extra-x (- (lR self) (inset-h (inset self)) w))
           (tx (ecase (justifyhz self)
                 (:left 0)
                 (:center (floor extra-x 2))
                 (:right extra-x))))

      (gl-raster-pos2i
       tx
       (+ (lb self) (v2-v (inset self)) (round (glut-bitmap-y-orig (font-ffi-glut-id font)))))
  
      (unless (cello-get-boolean GL_CURRENT_RASTER_POSITION_VALID)
        (trc nil "rasterpos offscreen" self :goffset (gOffset self)))
      (trc nil "raster pos valid?" (cello-get-boolean GL_CURRENT_RASTER_POSITION_VALID))
      (glut-bitmap-string (font-ffi-glut-id font) t$))))

(defmethod ix-render-in-font ((font font-glut-stroke) self)
  (bwhen (t$ (^displaytext$))
    (let* ((w (text-width self))
           (extra-x (- (lR self) (inset-h (inset self)) w))
           (tx (ecase (justifyhz self)
                 (:left 0)
                 (:center (floor extra-x 2))
                 (:right extra-x)))

           (fid (font-ffi-glut-id font))
           (raw-height (glut-stroke-height fid))
           (scale (* *glut-stroke-factor* (/ (scr2log (font-size font)) raw-height)))
           (raw-descent (glut-stroke-descent fid))
           #+dbg (log-descent (round (scr2log (* (font-size font)
                                                (/ raw-descent raw-height))))))
    
      (trc nil "stroke l-box, font-size" (lrect self) (font-size font))
      (trc nil "stroke raw height, descent" raw-height raw-descent)
      (trc nil "stroke scale, logdescent" scale log-descent)
    
      (gl-push-matrix)
      (gl-Translatef tx (* -1 scale (+ raw-height raw-descent)) 0)
      (gl-Scalef scale  scale 1.0)
      (gl-enable gl_blend)
      (gl-line-width 1)
      (ogb::glut-stroke-string (make-ff-pointer fid) t$)
      (gl-pop-matrix))))

(defmodel frame-rate-text (Ixtext)
  ()
  (:default-initargs
      :md-value (c? (cons (get-internal-real-time)
                      (frame-ct .w.)))
    :style :button
    :inset (mkv2 (uPts 2)(uPts 0))
    ;;:lt 15 :lb -5
    :charmask "999"
    :text$ (let (last)
             (c? (let ((this (^md-value)))
                   (prog1
                       (cond
                        ((null last)
                         (setf last this)
                         "not yet")
                        ((> 500 (- (car this)(car last)))
                         .cache)
                        (t
                         (prog1
                             (remove #\.
                               (Format nil "~3,0f"
                                 (/ (- (cdr this) (cdr last))
                                   (/ (- (car this) (car last))
                                     1000.0))))
                           (setf last this)))
                     )))))
    :lighting :off
    :layers (list :off +red+ :on)))

(defmodel IXTextTall (IXText)
  ((textheight :reader textHeight
               :initarg :textheight)
   (formatted$ :reader formatted$
               :initarg :formatted$)
   (wrapp :initform t))
  (:default-initargs
      :lB (c? (downs (^textHeight self)
                  (inset-v (inset self))))
    :formatted$ (c? (bIf (text$ (^text$))
                        (wrap$ text$
                               (- (^lR) (inset-h (inset self)))
                               (^targetRes (sCanvas)) (^textFont))
                        ""))
    :textheight (c? (bIf (text$ (^formatted$))
                        (* (font-height (textfont self))
                          (1+ (count #\newline text$)))
                        0))
    
    ))

(defun wrap$ (s w res font)
  (declare (ignorable w res font))
  s)