;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

(defun make-menu-bar (menus)
  (mkPart :menuBar (IXKidSized)
     :px (c? (^px-maintain-pl 0))
     :py 0
     :lL (u8ths -2)
     :lR (c? (^lr-maintain-pr (lr .parent)))
     :lB (c? (downs (lHeight (fm-other :title))))
     :clipped nil
     :lighting :off
     :layers (list +lt-gray+ :fill-lr #+not (mkframe3d :edge-raised))
     :kids (c? (mapcar #'make-menu menus))))

(defmodel CTMenu (Control IXStyled IXStack)
  ((items :initarg :items :reader items :initform nil))
  (:default-initargs
      :style :button
      :kids (c? (theKids
                    (mkPart :title-items (IXStack)
                      :kids (c? (theKids
                                 (mkPart :title (IXText)
                                   :lighting :off
                                   :layers (c? (let ((m (upper self CTMenu)))
                                                 (if (and (mouse-over-p m)
                                                       (eql :down (leftb (mouse *sys*))))
                                                     (list +black+ :fill-lr +white+)
                                                   (list +lt-gray+ :fill-lr +black+))))
                                   :textfont (c? (^textfont (upper self CTMenu)))
                                   :text$ (c? (title$ (upper self CTMenu))))
                                 (mkPart :menu-socket (WindowSocket)
                                   :collapsed t
                                   :window-factory (lambda (socket glut-xy)
                                                     (declare (ignorable socket glut-xy))
                                                     (trc "menu offset" glut-xy)
                                                     (mkPart :menu-window (CTMenuItems)
                                                       ;;:visible t
                                                       :glut-xy glut-xy))
                                   :gen-window-p t)
                                 )))))))



(defmodel CTMenuItems (IXStack Window)
  ()
  (:default-initargs
      :self-sizing t
      :lighting :off
    :outset (u96ths 4)
    :layers (list +blue+ :fill-lr)                  
    :lr (c? (+ (outset self) (loop for k in (^kids) maximize (lwidth k))))
    :menus nil
    :clear-rgba (list 1 0 0 1)
    :visible (c? (eko (nil "menuwindow visi (leftb mover)"
                       (leftb (mouse *sys*))
                       (mouse-over-p (upper self CTMenu)))
                   (and (eql :down (leftb (mouse *sys*)))
                     (mouse-over-p (upper self CTMenu)))))
    :kids (c? (theKids
               (mapcar #'make-menu-item (items (upper self CTMenu)))))))

(defmethod ix-render :around ((self CTMenuItems))
  ;; (trace ix-render ix-render-layer)
  (call-next-method)
  ;;(untrace ix-render ix-render-layer)
  )

(defun make-menu (menu)
  (destructuring-bind (title$ &rest items) menu
      (mkPart :menu (CTMenu)
         :title$ title$
         :items items
         :pY 0
         :px (c? (^px-maintain-pl
                  (bIf (p (psib))
                    (let ((pr (+ (u8ths 2)
                                (pR (cells::fm-find-one p :title
                                      :must-find t
                                      :skiptree nil
                                      :global-search nil)))))
                      (eko (nil "pr xlated from prior menu title" pr)
                        (h-xlate .parent p pr)))
                    0))))))


(def-c-echo visible ((self Window))
  (glut-set-window (glutw self))
  (if new-value
      (glut-show-window)
    (glut-hide-window)))

(defmethod do-menu-command (w cmd)
  (trc "unhandled menu command" w cmd))

(defmodel CTMenuItem (Control IXText)
  ()
  (:default-initargs
      ;;:lR (c? (^lr-maintain-pr (inset-lr .parent)))
      :mouse-up-handler  (c? (controlAction self))
    :inset (mkv2 (uPts 2)(uPts 2))
    :layers (c? (if (and (mouse-over-p self)
                      (eql :down (leftb (mouse *sys*))))
                    (list +dk-gray+ :fill-lr (make-3f :x 0 :y 0 :z 1500) +yellow+)
                  (list +white+ :fill-lr (make-3f :x 0 :y 0 :z 1500) +black+)))
    :style :button
    
    :controlAction (lambda (self event)
                     (declare (ignorable event))
                     (trc nil "bingo do menu" .w. (^md-value))
                     (do-menu-command (upper .w. Window) ;; fails when we get to nested menus <g>
                       (^md-value)))))

(defmethod make-menu-item (other)
  (mkPart :menu-item (CTMenuItem)
    :text$ (format nil "~a" other)
    :md-value other))

(defmethod make-menu-item ((item-def cons))
  (destructuring-bind (title$ . value) item-def
      (mkPart :menu-item (CTMenuItem)
         :text$ title$
         :md-value value)))

