;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

;-----------------------------

(defstruct v2 
  (h 0 )
  (v 0 )
  )

#+test
(instance-slots (mkv2 1 2))

(defmethod print-object ((self v2) s)
  (format s "(~a ~a)" (v2-h self)(v2-v self)))

(defun mkv2 (h v) (make-v2 :h h :v v))

(defun v2-add (p1 p2)
  (make-v2 :h (+ (v2-h p1) (v2-h p2))
    :v (+ (v2-v p1) (v2-v p2))))

(defun v2-subtract (p1 p2)
  (make-v2 :h (- (v2-h p1) (v2-h p2))
    :v (- (v2-v p1) (v2-v p2))))

(defun v2-in-rect (v2 r)
  (mkv2 (min (r-right r) (max (r-left r) (v2-h v2)))
    (min (r-top r) (max (r-bottom r) (v2-v v2)))))

(defun v2-in-rect-ratio (v2 r)
  (assert (<= (r-left r) (v2-h v2) (r-right r)))
  (assert (<= (r-bottom r) (v2-v v2) (r-top r)))
  (mkv2 (div-safe (- (v2-h v2) (r-left r)) (r-width r))
    (div-safe (- (v2-v v2) (r-bottom r)) (r-height r))))


(defmethod c-value-incf (c (base v2) (delta number))
  (declare (ignore c))
  (mkv2 (+ (v2-h base) delta)
    (+ (v2-v base) delta)))

(defmethod c-value-incf (c (base v2) (delta v2))
  (declare (ignore c))
  (v2-add base delta))

; synapse support
;
(defmethod delta-diff ((new v2) (old v2) (subTypeName (eql 'v2)))
  (v2-subtract new old))

(defmethod delta-identity ((dispatcher number) (subTypeName (eql 'v2)))
  (mkv2 0 0))

(defun long-v2 (longhv)
    (c-assert (numberp longhv))
    (multiple-value-bind (fv fh)
           (floor longhv 65536)
          (mkv2 fh fv)))

(defun long-x (longhv)
    (c-assert (numberp longhv))
    (mod longhv 65536))

(defun long-y (longhv)
    (c-assert (numberp longhv))
    (floor longhv 65536))

(defun v2-long (v2)
  (c-assert (typep v2 'v2))
  (xy-long (v2-h v2) (v2-v v2)))

(defun xy-long (x y)
  (+  (* 65536 y) x))

(defun v2-to-vector (v2)
  (vector (v2-h v2) (v2-v v2)))

(defun v2-negative (v2)
   (c-assert (typep v2 'v2))
   (mkv2 (- (v2-h v2)) (- (v2-v v2))))

(defun vector-v2 (vc) (mkv2 (elt vc 0) (elt vc 1)))

(defmethod delta-exceeds ((d1 v2) (d2 v2) (subTypeName (eql 'v2)))
     (c-assert (and (typep d1 'v2) (typep d2 'v2)))
     (> (v2-distance-to d1) (v2-distance-to d2)))

(defun v2-distance (from to)
     (sqrt (+ (expt (v2-dv from to) 2)
                 (expt (v2-dh from to) 2))))

(defun v2-dh (p1 p2)
     (- (v2-h p2) (v2-h p1)))

(defun v2-dv (p1 p2)
     (- (v2-v p2) (v2-v p1)))

(defun v2-angle-between (from to)
  (atan (v2-dv from to) (v2-dh from to)))

(defun v2-distance-to (to)
  (sqrt (+ (expt (v2-H to) 2)
           (expt (v2-V to) 2))))
;-------------------------------------------------

(defstruct (rect (:conc-name r-))
  (left 0 )
  (top 0 )
  (right 0 )
  (bottom 0 ))

(defmethod print-object ((self rect) s)
  (format s "(rect (~a,~a) (~a,~a))" (r-left self)(r-top self)(r-right self)(r-bottom self)))

(defun r-top-left (r)
  (mkv2 (r-left r) (r-top r)))

(defun r-bottom-right (r)
  (mkv2 (r-bottom r) (r-right r)))

(defun mkR (left top right bottom)
   (count-it :mkrect)
   (make-rect :left left :top top :right right :bottom bottom))

(defun nr-make (R left top right bottom)
   (setf (r-left R) left (r-top R) top (r-right R) right (r-bottom R) bottom)
  R)

(defun nr-make-from-corners (r tl br)
   (nr-make r (v2-h tl)(v2-v tl)(v2-h br)(v2-v br)))

(defun nr-copy (r copied-r)
   (setf (r-left R) (r-left copied-r)
     (r-top R) (r-top copied-r)
     (r-right R) (r-right copied-r)
     (r-bottom R) (r-bottom copied-r))
  R)

(defun r-contains (r v2)
  (and (<= (r-left r)(v2-h v2)(r-right r))
    (<= (r-top r)(v2-v v2)(r-bottom r))))

(defun nr-intersect (r sr)
  (let ((r-min-v (min (r-top r) (r-bottom r)))
        (r-max-v (max (r-top r) (r-bottom r)))
        (r-min-h (min (r-left r) (r-right r)))
        (r-max-h (max (r-left r) (r-right r)))
        ;
        (sr-min-v (min (r-top sr) (r-bottom sr)))
        (sr-max-v (max (r-top sr) (r-bottom sr)))
        (sr-min-h (min (r-left sr) (r-right sr)))
        (sr-max-h (max (r-left sr) (r-right sr)))
        )
   (let ((min-v (max r-min-v sr-min-v))
         (max-v (min r-max-v sr-max-v))
         (min-h (max r-min-h sr-min-h))
         (max-h (min r-max-h sr-max-h)))
     (when (or (>= min-v max-v)(>= min-h max-h))
       (setf min-h 0 min-v 0 max-h 0 max-v 0))
     (nr-make r min-h min-v max-h max-v))))



(defun nr-move-to (R h v)
   (setf (r-left R) h
     (r-top R) (+ v (r-width r))
     (r-right R) (+ h (r-width r))
     (r-bottom R) (+ v (r-height r))))


(defun nr-scale (R factor)
   (nr-make R
     (round (* (r-left R) factor))
     (round (* (r-top R) factor))
     (round (* (r-right R) factor))
     (round (* (r-bottom R) factor))))

(defun r-empty (r)
  (or (zerop (r-width r))
    (zerop (r-height r))))

(defun r-width (r) (abs (- (r-right r)(r-left r))))
(defun r-height (r) (abs (- (r-top r)(r-bottom r))))


(defun nr-offset (r dh dv)
;;;   (declare (optimize (speed 3) (safety 0) (debug 0)))
  (declare (type fixnum dh dv))
  (incf (r-left r) dh)
  (incf (r-right r) dh)
  (incf (r-top r) dv)
  (incf (r-bottom r) dv)
  r)

(defun nr-outset (box dh &optional (dv dh))
;;;   (declare (optimize (speed 3) (safety 0) (debug 0)))
  (declare (type fixnum dh dv))
  (decf (r-left box) dh)
  (incf (r-right box) dh)
  (decf (r-top box) dv)
  (incf (r-bottom box) dv)
  box)

(defun pt-in-bounds (point boundsleft boundstop boundsright boundsbottom)
;;;   (declare (optimize (speed 3) (safety 0) (debug 0)))
  (declare (type fixnum boundsleft boundstop boundsright boundsbottom))
  (and (<= boundsleft (progn (v2-h point)) boundsright)
       (<= boundstop (progn (v2-v point)) boundsbottom)))


(defun r-in-bounds (box boundsleft boundstop boundsright boundsbottom)
;;;   (declare (optimize (speed 3) (safety 0) (debug 0)))
  (declare (type fixnum boundsleft boundstop boundsright boundsbottom))
  (and (<= boundsleft (progn (r-left box)) (progn (r-right box)) boundsright)
       (<= boundstop (progn (r-top box)) (progn (r-bottom box)) boundsbottom)))

(defun r-unitize (object-r unit-r &aux (ww (r-width unit-r))(wh (r-height unit-r)))
  (flet ((cf (i) (coerce i 'float)))
    (mkr (cf (/ (- (r-left object-r)(r-left unit-r)) ww))
      (cf (/ (- (r-top unit-r)(r-top object-r)) wh))
      (cf (/ (- (r-right object-r)(r-left unit-r)) ww))
      (cf (/ (- (r-top unit-r)(r-bottom object-r)) wh)))))

(defun r-scale (r x y)
  (mkr (* (r-left r) x)
    (* (r-top r) y)
    (* (r-right r) x)
    (* (r-bottom r) x)))

(defun r-analog (this1 that1 this2)
  (mkr (* (r-left this2) (/ (r-left that1)(r-left this1)))
    (* (r-top this2) (/ (r-top that1)(r-top this1)))
    (* (r-right this2) (/ (r-right that1)(r-right this1)))
    (* (r-bottom this2) (/ (r-bottom that1)(r-bottom this1)))))
