;; -*- mode: Lisp; Syntax: Common-Lisp; Package: Cello; -*-
;;________________________________________________________
;;
;;           M G W i n d o w
;;
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

(defun ncalc-normalfv(vout v0 v1 v2)
  (with-ff-array-elements (v0 :float v0x v0y v0z)
    (with-ff-array-elements (v1 :float v1x v1y v1z)
      (with-ff-array-elements (v2 :float v2x v2y v2z)
        (ncalc-normalf vout v0x v0y v0z v1x v1y v1z v2x v2y v2z)))))

(defun ncalc-normalf(vout v0x v0y v0z v1x v1y v1z v2x v2y v2z
                      &aux d0x d0y d0z d1x d1y d1z)
  (setf d0x (- v1x v0x)
    d0y (- v1y v0y)
    d0z (- v1z v0z))
  (setf d1x (- v2x v1x)
    d1y (- v2y v1y)
    d1z (- v2z v1z))

  (setf (eltf vout 0) (- (* d0y d1z) (* d0z d1y)))
  (setf (eltf vout 1) (- (* d0z d1x) (* d0x d1z)))
  (setf (eltf vout 2) (- (* d0x d1y) (* d0y d1x)))

  (xgl-normalize-3f vout)
  vout)

(defstruct 3f
  (x 0)(y 0)(z 0))

(defmethod ix-render-layer ((self 3f) lBox)
  (declare (ignore lbox))
  (trc nil "ixr-layer translatef xl-z" (3f-z self))
  (gl-translatef (3f-x self)(3f-y self)(3f-z self)))

#+test
(with-ff-array-elements ((ncalc-normalf (make-ff-array :float 0 0 0)
                               0 0 60 -15 0 30 0 0 30)
                             x y z)
  (list x y z))

(defun xgl-normalize-3f (v)
  (trc nil "raw normal" (eltf v 0)(eltf v 1)(eltf v 2))
  (with-ff-array-elements (v :float x y z)
    (let ((m (sqrt (+ (* x x) (* y y) (* z z)))))
      (unless (zerop m)
        ;(trc "normalizing div" m)
        (setf (eltf v 0) (- (/ (eltf v 0) m)))
        (setf (eltf v 1) (- (/ (eltf v 1) m)))
        (setf (eltf v 2) (- (/ (eltf v 2) m)))
        ;(trc "normalized normal" (eltf v 0)(eltf v 1)(eltf v 2))
        v))))

#+test
(with-ff-array-elements ((ncalc-normalfv (make-ff-array :float 0 0 0)
                               (make-ff-array :float 0 0 60)
                               (make-ff-array :float -15 0 30)
                               (make-ff-array :float 0 0 30))
                         :float x y z)
  (list x y z))

(uffi:def-foreign-type bool* (* glboolean))

#-lispworks
(declaim (type bool* *cello-boolean*))

(defparameter *cello-boolean*
  (uffi:allocate-foreign-object 'glboolean 1))

(defun cello-get-boolean (gl-code)
  (gl-get-booleanv gl-code *cello-boolean*)
  (not (zerop (uffi:deref-array *cello-boolean* '(:array glboolean) 0))))

(uffi:def-foreign-type glint* (* glint))

#-lispworks
(declaim (type glint* *cello-int*))

(defparameter *cello-int*
  (uffi:allocate-foreign-object 'glint 1))

(defun cello-get-int (gl-code)
  (gl-get-integerv gl-code *cello-int*)
  (uffi:deref-array *cello-int* '(:array glint) 0))

(defparameter *dbg-viewport-r*
  (uffi:allocate-foreign-object 'GLint 4))

(defun dump-viewport (key)
  (gl-get-integerv gl_viewport *dbg-viewport-r*)
  (format t "~&dump-viewport> ~a: ~a" key
    (loop for n from 0 to 3
        collecting (uffi:deref-array *dbg-viewport-r* '(:array GLint) n))))  

(defparameter *gl-ints-4*
  (uffi:allocate-foreign-object 'GLint 4))

(defun gl-get-ints-4 (gl-param)
  (gl-get-integerv gl-param *gl-ints-4*)
  (loop for n below 4
      collecting (uffi:deref-array *gl-ints-4* '(:array GLint) n)))