;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

;; (pushnew :keying *features*)

;;; Do-virtual-key stuff

(defun buttons-shifted (buttons)
  (logtest buttons glut_active_shift))

(defun shift-key-down (buttons)
  (logtest buttons glut_active_shift))

(defun control-key-down (buttons)
  (logtest buttons glut_active_ctrl))

(defun alt-key-down (buttons)
  (logtest buttons glut_active_alt))

(defun control-shift-key-down (buttons)
  (and (shift-key-down buttons)
       (control-key-down buttons)))

(defun shift-key-only? (buttons)
  (eql glut_active_shift buttons))

;------------------------------------------

(def-c-echo mouseCursor (self newValue)
  (when newValue
     (glut-set-cursor newValue)))

;----------virtual key up -----------------
;

(defmethod virtual-key-up ((w Window) buttons data)
   (declare (ignorable w buttons data))
   ;;(format t "~&virtual-key-up  > Window ~s, data ~d, buttons ~d" w data buttons)
   (setf (kbdModifiers w) buttons)
   (call-next-method))

;---------- virtual-key-down --------------------
;

(defmethod virtual-key-down :after ((w Window) buttons charAscii)
  (declare (ignorable w buttons charAscii))
  (unless (closed-stream-p w)
    ;(format t "~&virtual-key-down :after > Window ~s, data ~d, buttons ~d" w charAscii buttons)
    
    (setf (kbdModifiers w) buttons)
    
    (with-metrics (nil nil "virtual-key-down :after" w charAscii)
      (let ((res (catch :userCancel
                    (do-virtual-key 
                        (or (focus w) w)    ;ML 000830 this is a kludge so that Alt keys fall through.
                      buttons charAscii))))
         res))))

(defgeneric do-virtual-key (self buttons charAscii) (:method-combination or))

(defmethod do-virtual-key or (self buttons charAscii)
  (trc nil "dvk t" buttons charascii (code-char charascii))
  (when self
    (if (fm-parent self)
        (do-virtual-key (fm-parent self) buttons charAscii)
      #+chya (progn
               (trc "dvk addressing debug" self)
               (cellbrk)
               (let ((debug-pane (find-debug-pane))) ;not guaranteed robust, but seems OK
                 (set-focus debug-pane)
                 (unless (virtual-key-down debug-pane buttons charAscii)
                   (set-focus (windowCache self))))))))



;--------- character-message -----------------------
;


;;; (defmethod character-message ((w Window) buttons aChar)
;;;   (declare (ignorable buttons aChar))
;;;   ;;(cellbrk)
;;;   (trc nil "character-message  > Window" w :achar achar :focus (focus w))
;;;   (with-metrics (nil nil () "character message" w)
;;;     (if (focus w)
;;;      (with-invalidation (w)
;;;         (case aChar
;;;           ;
;;;           (#\tab) ;; do in do-virtual-key
;;;           ;
;;;           (otherwise
;;;            (character-message (focus w) buttons aChar))))
;;;      (beep))))

;;; (defmethod win:WM_SYSKEYDOWN :around ((window Window) wparam lparam)
;;;   (declare (ignorable wparam lparam))(cellbrk)
;;;   (call-next-method))
;;; 
;;;(defmethod win:WM_KILLFOCUS :around ((window Window) wparam lparam)
;;;  (declare (ignorable wparam lparam))
;;;  ;;(trc "KILL FOCUS")
;;;  ;;(print (list "FOCUS" (focus window)))
;;;  ;;(cellbrk)
;;;  (prog1
;;;      (call-next-method)
;;;    (when (and nil (typep (focus window) 'Multitext))
;;;      (funcall 'MT-kill-focus (focus window))
;;;      ))
;;;  )
;;;
;;;(defmethod win:WM_SETFOCUS :around ((window Window) wparam lparam)
;;;  (declare (ignorable wparam lparam))
;;;  ;;  (trc "SET FOCUS")
;;;  (prog1
;;;      (call-next-method)
;;;    (when (and nil (typep (focus window) 'Multitext))
;;;      (funcall 'MT-set-focus (focus window))
;;;      ))
;;;  )

;;; (defmethod event :around ((self Window) eno bst dta)
;;;   (unless (eql eno 512)
;;;     (trc "event" self eno bst dta))
;;;   (prog1
;;;       (call-next-method)
;;;     (unless (eql eno 512)
;;;       (trc "event exit" self eno bst dta))))
;;; 
;;; (defmethod win:WM_CHAR :around ((self Window) wParam lParam)
;;;   (trc "WM_CHAR" self wParam lParam)
;;;   (prog1
;;;       (call-next-method)
;;;     (trc "WM_CHAR exit" self wParam lParam)))
;;; 
;;; (defmethod cg::window-procedure :around ((self Window) msg wParam lParam)
;;;   (when (< msg 512)
;;;     (unless t ;(member msg '(15 132 32))
;;;       (trc "window-procedure" self msg wParam lParam)))
;;;   (prog1
;;;       (call-next-method)
;;;     (when (< msg 512)
;;;       (unless t ; (member msg '(15 132 32))
;;;         (trc "window-procedure exit" self msg wParam lParam)))))
;;; 
;;; (defmethod cg::control-procedure :around ((self Window) msg wparam lparam)
;;;   ;; By default, just pass the event back to windows so
;;;   ;; the control can do its usual thing
;;;   (when (< msg 512)
;;;     (trc "control-procedure" self msg wParam lParam))
;;;   (call-next-method))

