;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

;-------------------- double click -----------------------------------

(defmethod do-double-click :around (self osEvent &rest iargs &key  &allow-other-keys)
  (when self
    (or (call-next-method)
        (apply #'do-double-click (fm-parent self) osEvent iargs))))

(defmethod do-double-click (self osEvent &key)
  (declare (ignorable self osEvent))  
  ;;(trc "*** No special do-double-click for image, event:" self osEvent)
  nil)

; ------------------- right button --------------------------------------

(defun geo-dump (i)
  (when (typep i 'Image)
    (print (list :pxy (cons (px i)(py i)) :lt (lt i) :lb (lb i)))
    (geo-dump (fm-parent i))))

(defmethod WM-RBUTTONDOWN ((w Window) buttons mousePos)
  (declare (ignorable buttons mousePos))
  (trc "rbuttondown" w buttons mousepos GLUT_ACTIVE_CTRL (logtest GLUT_ACTIVE_CTRL buttons))
  (bwhen (i (find-ix-under w mousePos))
    (trc "mpos ix=" i)
    (unless (do-right-button i buttons mousePos)
      (cond
       ((logtest GLUT_ACTIVE_CTRL buttons) (geo-dump i))
       (t (print `(inspecting ,i))
         (c-stop :inspecting)
         (inspect i)))))
  (setf (kbdModifiers w) 0) ;; ugh, no WM_KEYUP after ^right-click, so kbdMods gotta be cleared here
  )

(defmethod do-right-button :around (image buttons wxwy)
  (declare (ignorable buttons wxwy))
  (when image
    (or (call-next-method)
        (do-right-button (fm-parent image) buttons wxwy))))

(defmethod do-right-button (image buttons wxwy)
  (declare (ignorable image buttons wxwy)))

(defmethod do-menu-right (self buttons wxwy)
  (declare (ignorable buttons self wxwy)))

(defmethod make-menu-right-items (self)
  (declare (ignorable self)))

(defmethod menu-right-select (self item)
  (declare (ignorable self item)))

(defmethod menu-shortc (self)
  (shortc self))

; --------------- geometry -------------------------------


(defmethod gOffset ((ap Window) &optional (accumH 0) (accumV 0))
   (mkv2 accumH accumV))

(defmethod winOffset ((ap Window) &optional (accumH 0) (accumV 0))
   (mkv2 accumH accumV))

(defun point-in-box (pt box)
  (and (<= (r-left box) (v2-h pt) (r-right box))
       (>= (r-top box) (v2-v pt) (r-bottom box))))


(defun screenbox (self)(gbox self)) ;; good luck

; ---------------------- finding parts ------------------------------

(defun mouseImage-control (w)
  (fm-ascendant-if (mouseImage w)
                   (lambda (node)
                       (and (typep node 'Control)
                            (fully-enabled node)))))

(defmacro dolistreversed ((itemVar listform &optional resultform) &body body)
  (let ((list (gensym))
        (len (gensym))
        (timesvar (gensym)))
    `(let* ((,list ,listForm)
            (,len (length ,list)))
       (dotimes (,timesVar ,len ,resultform)
         (let ((,itemVar (elt ,list (- ,len ,timesvar 1))))
           ,@body)))))

(defun find-ix-under (self osPos &key (test #'true))
  (when (and (visible self)
             (not (collapsed self)))
    (let ((inself (point-in-box osPos (screenBox self))))
      (or (when (or inself (not (clipped self)))
            (trc nil "inside self sbox" self ospos (screenbox self))
            (dolistreversed (k (kids self)) ;; overlap goes to last kid displayed
              (unless (typep k 'Window)
                (trc nil "fixunder kid" k)
                (bWhen (ix (find-ix-under k osPos :test test))
                  (return-from find-ix-under ix)))))
          
          (when (and inself
                     (funcall test self)
                     (not (ix-click-transparent self)))
            (trc nil self osPos (screenbox self))
            self)))))


(defmethod ix-click-transparent (other)
  (declare (ignore other))
  t)

