;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

(defun cello-gl-init (&aux (ct 0))
  (trc "clearing gl errors....")
  (do ((e (glGetError)(glGetError)))
      ((zerop e) t)
    (when (> (incf ct) 10)
      #-lispworks (cellbrk :gl-init)
      #+lispworks (return-from cello-gl-init))
    (trc "clearing gl error" e)))


;------------- Window ---------------
;
(defmodel Window (Focuser IXFamily Control)
  (
   (glutw :initarg :glutw  :accessor glutw
     :initform (c? (without-c-dependency (glutw-create self))))
   (w-fonts :cell nil :initform nil :accessor w-fonts
     :documentation "Widgets share these where attributes match")
   (redisplayp :cell nil :initarg :redisplayp :initform nil :accessor redisplayp)
   (clear-rgba :cell nil :initarg :clear-rgba :initform nil :accessor clear-rgba)
   (lights :initarg :lights :accessor lights
     :initform (list (make-part :light0 'Light
                       :id GL_LIGHT0
                       :enabled t
                       :pos *LightPosTL*
                       :ambient *Average*
                       :diffuse *Bright*
                       :specular *Bright*)
                 (make-part :light1 'Light
                   :id GL_LIGHT1
                   :enabled t
                   :pos *LightPosTR*
                   :ambient *Dim*
                   :diffuse *Bright*
                   :specular *Bright*)))
   (glut-xy :initarg :glut-xy :initform (mkv2 96 96) :accessor glut-xy)
   (self-sizing :cell nil :initarg :self-sizing :initform nil :accessor self-sizing)
   (glut-lbox :reader glut-lbox
     :initform (c? (count-it :glutlbox)
                 (mkR (^lL)(^lT)(^lR)(^lB))))
   (display-continuous :initarg :display-continuous :initform nil :accessor display-continuous)
   (frame-ct :initarg :frame-ct :initform (cv 0) :accessor frame-ct)

   (background-color :cell nil :initarg :background-color :initform nil :accessor background-color)
   (title$ :initarg :title$ :initform "Untitled" :accessor title$)
   (openState :cell nil :initarg :openState :initform nil :accessor openState)
   (closeRequested :cell nil :initform nil :accessor closeRequested)
      
   (kbdmodifiers :initarg :kbdmodifiers :initform nil :accessor kbdmodifiers)
      
   (activep :initarg :activep :initform nil :accessor activep)

   (windowcmd :cell :ephemeral :initarg :windowcmd :initform nil :accessor windowcmd)

   (menus :initarg :menus :initform NIL :accessor menus)

   (mouseGrabber :initarg :mouseGrabber
                 :initform (cv nil)
                 :accessor mouseGrabber)
   (mousepos :initarg :mousepos :initform (cv nil) :accessor mousepos)   ;logical coords.  Try to maintain for now.
   
   (mouseimage :initarg :mouseimage :accessor mouseimage
               :initform (c? (let ((mp (^mousePos)))
                               (trc nil "mouseimage sees pos" .w. mp)
                               (when mp
                                 (eko (nil "mouseimage >" self)
                                   (without-c-dependency
                                    (find-ix-under self mp)))))))
   
   (mouseControl :initarg :mouseControl :accessor mouseControl
                 :initform (c? (bWhen (node (^mouseImage))
                                 (eko (nil "possible mousecontrol" node)
                                   (fm-ascendant-if node #'fully-enabled)))))
   
   (mouseCursor :initarg :mouseCursor :initform nil :accessor mousecursor)
   
   (mouseupevt :cell :ephemeral :initarg :mouseupevt :initform (cv nil) :accessor mouseupevt)
   (mouseDownevt :cell :ephemeral :initarg :mouseDownevt :initform (cv nil) :accessor mouseDownevt)
   (doubleClick? :initform (cv nil) :accessor doubleClick?)
   
   #+lets-find-out (kbSelector :cell nil :initarg :kbSelector ;; what the hell is this for?
     :initform nil
     :reader kbSelector)
   
   (tickcount :initarg :tickcount :initform (cv nil) :accessor tickcount)
   (tickfine :initarg :tickfine :initform (cv nil) :accessor tickfine)
   )
  (:default-initargs
      :kids (c? (theKids (^content)) #+not (theKids
                 (mkPart :wstuff (ixkidsized)
                    :px 0 :pY (c? (bIf (n (nsib))
                                      (pB n) 0))
                   :kids (c? (theKids (content .w.))))
                 (bWhen (ms (menus self))
                   (make-menu-bar ms))
                 ))
      :activeP (cv nil)
    :clear-rgba (list 0 0 0 1)

    :buttons (cv 0) ;; 990728kt was nil
    :focus (cv nil)
    ;;:layers (list +gray+ :fill-lr)
    :focusSource nil
    :kbdModifiers (cv 0) ;; 990728kt was nil
    
    :self-sizing nil
    :lL 0 :lT 0
    :lR (cv (scr2log 800))
    :lB (cv (scr2log -800))
    
    :mouseCursor (c? (context-Cursor (^mouseControl) (^kbdModifiers))) 
    
    :px 0 :py 0
    :windowCmd (cv nil)
    :tickCount (cv (os-TickCount))
    :lighting :off
    :clipped t
    ))

(def-c-echo lights ()
  (dolist (light new-value)
    (to-be light)))

(defun window-menus-basic ()
  (list
   (list "File"
     (cons "New" :menu-file-new)
     (cons "Open" :menu-file-open)
     (cons "Close" :menu-file-close)
     (cons "Quit" :menu-file-quit))
   (list "Edit"
     (cons "Undo" :menu-edit-undo)
     (cons "Cut" :menu-edit-cut)
     (cons "Copy" :menu-edit-copy)
     (cons "Paste" :menu-edit-paste)
     (cons "Delete" :menu-edit-delete))))
                      
(defmethod set-doubleclick? ((self Window) click)
  (setf (doubleclick? self) click))

(defmethod context-cursor (other kbdModifiers)
   (if (and other (fm-parent other))
       (context-cursor (fm-parent other) kbdModifiers)
     GLUT_CURSOR_LEFT_ARROW))

(def-c-echo glut-lbox ()
  (when (self-sizing self) ;; we drive os window
    (with-glutw (self)
      (let ((w (log2scr (lwidth self)))
            (h (log2scr (lheight self))))
        (gl-viewport 0 0 w h)
        (trc "reshaping window #" self (glut-get-window) w h)
        (glut-reshape-window w h)))))

(defmodel WindowSocket (IXRow)
  ((gen-window-p :initarg :gen-window-p :initform #'true :accessor gen-window-p)
   (window-factory :initarg :window-factory :initform nil :accessor window-factory))
  (:default-initargs
      :md-name :socket
    :window-factory nil
    #+likethis (lambda (socket offset)
                 (declare (ignorable socket offset))
                 (make-instance 'starter-w
                   :md-name :starter-w
                   :glut-xy offset))
    :kids (c? (theKids
               (when (^gen-window-p)
                 (assert (window-factory self))
                 (let ((offset (without-c-dependency
                                (eko ("window socket offset" self)
                                  (v2-log2scr (goffset self))))))
                   (setf (v2-v offset) (- (v2-v offset)))
                   (funcall (window-factory self) self offset)))))))

(defun v2-log2scr (xy)
  (mkv2 (log2scr (v2-h xy)) (log2scr (v2-v xy))))

(defmethod do-menu-command ((w Window) (cmd (eql :menu-file-close)))
  (trc "destroying window" w (glutw w))
  (glut-destroy-window (glutw w)))

(defmethod do-keydown :around (self k event)
  (when self
    (unless (call-next-method)
      (do-keydown .parent k event))))

(defmethod do-keydown ((w Window) k event)
    (case k
      (#\Escape (if (shift-key-down (evtbuttons event))
                    (break "user break on window ~a" (mg-window-current))
                  (progn
                    (trc "destroying window" (glutgetwindow) :out-of
                      (mapcar #'glutw (kids *sys*)))
                    (glut-Destroy-Window (glutgetwindow))
                    (setf (kids *sys*) (remove w (kids *sys*))))))
      ))

(defmethod do-keydown (self k event)
  (declare (ignorable self k event)))

(defmethod do-specialkeydown :around (self k event)
  (when self
    (unless (call-next-method)
      (do-specialkeydown .parent k event))))

(defmethod do-specialkeydown (self k event)
  (declare (ignorable self k event)))

(def-c-echo mouseImage ()
  (when old-value
    (trc nil "mouseover lost by" old-value (windowcache old-value))
    (setf (mouse-over-p old-value) nil))
  (when new-value
    (trc nil "mouseover gained by" new-value (windowcache new-value))
    (setf (mouse-over-p new-value) t)))

(def-c-echo mouseDownEvt (self mDown)
  (when mDown
    (trc "mousedown" mdown (mouseControl self))
    (bWhen (clickee (mouseControl self))
      (trc nil "mousedown clickee, clickw" clickee self)
      (mkPart :click (MouseClick)
        :clickwindow self
        :clickee clickee
        :osEvent mdown
        :clickeePXY (mkv2 (px clickee) (py clickee))))))

(def-c-echo mouseUpEvt (self up)
  (when up ;; should be since this is ephemeral, but still..
    (trc nil "mouseup" self up (mouseControl self))
    (bWhen (clickee (mouseControl self))
      (bWhen (upper (mouse-up-handler clickee))
        (trc nil "mouseup clickee, clickw" clickee self)
        (funcall upper clickee up)))))

(defparameter *gw* nil)


(defmethod glutw-create ((self Window))
  (when *gw* (cellbrk :gwcre-renetered))
  (let ((*gw* t))
    (glut-set-option GLUT_ACTION_ON_WINDOW_CLOSE GLUT_ACTION_GLUTMAINLOOP_RETURNS)
    (glut-Init-Display-Mode (+ GLUT_RGB GLUT_DOUBLE))
    
    (let ((glutw (bif (w (upper self Window))
                   (apply #'glut-create-sub-window (glutw w)
                     (v2-h (glut-xy self)) (v2-v (glut-xy self))
                     (if (self-sizing self)
                         (list 100 100)
                       (list (log2scr (lwidth self))
                         (log2scr (lheight self)))))
                   (progn
                     (if (self-sizing self)
                         (glut-Init-Window-Size 100 100)
                       (glut-Init-Window-Size (log2scr (lwidth self))
                         (log2scr (lheight self))))
                     
                     (let ((key (or (title$ self) "Untitled")))
                       (uffi:with-cstring (key-native key)
                         (glut-Create-Window key-native)))))))

      (setf (glName self) glutw)

      (trc nil "glutw-create setting glname" self :to (glName self) :glutw glutw
        :glut-get-w (glut-get-window))

      (cello-gl-init) ;; clear errors
      
      (trc "glutw-create'd window XY" (glut-get-window) self :from (glut-xy self) :to 
        (list (glut-get glut_window_x)(glut-get glut_window_y)
          (glut-get glut_window_width)(glut-get glut_window_height)))

      (gl-Enable GL_TEXTURE_2D)
      (gl-Shade-Model GL_SMOOTH)					;; Enable Smooth Shading
      (apply #'gl-Clear-Color (clear-rgba self))
      (gl-Clear-Depth 1.0f0)					;; Depth Buffer Setup
      (gl-Enable GL_DEPTH_TEST)					;; Enables Depth Testing
      (gl-Depth-Func GL_LEQUAL)					;; The Type Of Depth Testing To Do
      (gl-Hint GL_PERSPECTIVE_CORRECTION_HINT GL_NICEST)

      (glut-callback-set
       ;;:idle 'mg-glut-idle
       :keyboard 'mgwkey
       :close 'mg-glut-close
       :display 'mg-glut-display
       :mouse 'mg-mouse-callback
       :passive-motion 'mg-passive-motion-callback
       :motion 'mg-motion-callback
       :reshape 'mg-glut-reshape)

      glutW)))

(defmethod ix-selectable ((self Window)) t)

(defmethod ix-invalidate ((self Window))
  ;; /// looks slow
  (when (slot-value self 'glutw) ;; not until ready, and use backdoor else reenter creation
    (setf (redisplayp self) t)))

(defun w-post-redisplay (self)
  (let ((w (glut-get-window))
        (gw (glutw self)))
    (c-assert gw)
    (glut-set-window gw)
    (count-it :post-redisplay)
    (trc nil "posting redisplay" self (glutw self) :currentw w)
    (glut-post-redisplay)
    (c-assert w)
    (glut-set-window w)))

(defparameter *w* nil)

(defmacro def-Window-callback (fn-name args &body body)
  `(ff-defun-callable :cdecl :void ,fn-name ,args
     (unless (c-stopped)
       ;;
       ;; this next bit makes sense because no cell rule evaluation could
       ;; depend on something touched during a callback, but then no cell
       ;; rule should dynamically encompass a callback, so...why reset 
       ;; the calculators (dependents) global? it is necessary
       ;; because, when an error occurs, error-handling can cause 
       ;; re-entrance and, if a cell rule was being evaluated, suddenly
       ;; the programmer is looking at an error about "too many dependencies"
       ;; instead of the original error. there is probably a better way to handle
       ;; all this, but for now... 2003-04-05kwt
       ;;
       (let (cells::*c-calculators*
             (*w* (mg-window-current)))
         (if *w*
             (prog2
               (setf (redisplayp *w*) nil)
                 (progn ,@body)
               (when (redisplayp *w*)
                 (w-post-redisplay *w*)))
           (progn ,@body))))))

(def-Window-callback mgwkey (k x y)
  (trc nil "mgwkey" k x y (glutgetwindow))
  (bWhen (w *w*)
    (trc nil "mgwkey" k x y w)
    (let ((mods (glut-get-modifiers)))
      ;;(print  (list  :keyboard k mods x  y (code-char (logand k #xff)) (focus w)))
      (do-keydown (or (focus w) w)
        (code-char (logand k #xff))
        (mk-osevent mods (mkv2 x y))))))

(def-Window-callback mgwspecial (k x y)
  (trc nil "mgwspecial" k x y (glutgetwindow))
  (bWhen (w *w*)
    (trc nil "mgwspecial" k x y w)
    (let ((mods (glut-get-modifiers)))
      (do-specialkeydown (or (focus w) w)
        k
        (mk-osevent mods (mkv2 x y))))))


(defmethod do-specialkeydown ((w Window) k event)
  (declare (ignorable k event)))

(defun mg-window-current (&optional must-find-p)
  (let ((gw (glutgetwindow)))
    (if (zerop gw)
        (when must-find-p
          (cellbrk 'cannotfindcurrentwindow))
      (or (catch 'mg-window-current
            (fm-traverse *sys* (lambda (node)
                                 (when (and (typep node 'Window)
                                         (eql gw (glutw node)))
                                   (throw 'mg-window-current node)))
              :skiptree nil
              :global-search nil))
        (when must-find-p
          (trc "no mgw matches glutw ~d" gw)
          (cellbrk))))))

(defmethod ix-idle ((w Window))
  (setf (tickCount w) (os-TickCount)))

(def-Window-callback mg-glut-idle ()
  (trc nil "mg-glut-idle entry" (mg-window-current))
  (when *w*
    (ix-idle *w*)))

(def-Window-callback mg-glut-display ()
  (trc nil "mg-glut-display > entry" (glutgetwindow))
  (let ((*selecting* nil)
        )
    (progn ;; handler-case
        (progn ;; with-metrics (t t "mg-glut-display")
          (trc nil "mg-glut-display > about to render w " *w* (glutgetwindow))
          ;;(trace ix-render)
          (let (*render-clip-l* *render-clip-r* *render-clip-t* *render-clip-b*)
            (ix-render *w*))
            (glut-swap-buffers)
            (incf (frame-ct *w*))
            (trc nil "mg-glut-display > rendered w " *w* (glutgetwindow))
            (when (display-continuous *w*)
              (trc nil "mg-glut-display > posting redisplay w " *w* (glutgetwindow))
              (glut-post-redisplay))
            (trc nil "mg-glut-display finis frame" (frame-ct *w*))))
      #+nah (t (c)
        (DECLARE (ignorable c))
        (trc "display callback error" c)
        (c-stop :mg-glut-display)
        (invoke-debugger c)
        )))

(def-Window-callback mg-glut-close ()
  (trc "bingo close ID" (glut-get-window))
  (when *w*
    ;; knowing about a window CLO has forgotten
    (c-assert (fm-includes *sys* *w*))
    (trc "closing ~a" *w*)
    (setf (kids *sys*) (remove *w* (kids *sys*)))
    (trc nil "closed ~a" *w*)))

(defmethod not-to-be :before ((self Window))
  (when (upper self Window) ;; better way to detect appropriateness?
    (when (glutw self)
      (glut-Destroy-Window (glutw self)))))

(defconstant *mgw-znear* 1500)
(defconstant *mgw-zfar* -1500)

(def-Window-callback mg-glut-reshape (x y)
  (unless (or (null *w*)(zerop x) (zerop y)(self-sizing *w*))
    (trc "mg-glut-reshape entry" (mg-window-current t) x y)
    (mg-window-reshape *w* x y)))

(defmethod mg-window-reshape (self width height)
  ;;(trc "mg-window-reshape" self width height)
  (trc "mg-window-reshape" self width height)
  (gl-viewport 0 0 width height)
  (gl-matrix-Mode GL_PROJECTION)
  (gl-Load-Identity)
  
  (trc "mg-window-reshape ortho"  0 width (- height) 0 *mgw-znear* *mgw-zfar*)
  (gl-ortho 0 width (- height) 0 *mgw-znear* *mgw-zfar*)
  (gl-Load-Identity)
  (trc nil "mg-window-reshape > new window wid,hei:" self width height)
  (setf (lr self) (+ (ll self) (scr2log width)))
  (setf (lb self) (- (lt self) (scr2log height))))

(defun run-window (new-window)
  (let ((ogb::*gl-stop* nil)
        #+cellogut (*wands-loaded* nil)
        (ogb::*glut-lib-path*
         "c:/dev/lib/freeglut.dll"
         #+not "c:/windows/system32/glut32.dll"
         #+unstable "c:/dev2/cvs/freeglut/freeglut/debug/freeglut.dll")
        )
    (setf cello::*sys* nil)
    (cello-reset 'MGSystem)
    (with-metrics (nil nil "make starter window")
      (push new-window
        (kids *sys*)))
    (glut-post-window-redisplay (glutw new-window))
    (handler-bind ((error #'(lambda (c) (print `(bingo glut leave ,c))
                              (c-stop :top-handler)
                              (glut-leave-main-loop))))
      (glutMainLoop)
      #+nah (do ()
          ((zerop (glut-get-window)))
        ;;(format t "before main loop ~a | ~&" (glut-get-window))
        (progn ;; with-render-lock ((glut-get-window))
          (glutMainLoopEvent))
        (sleep 0.1)
        ))))
;;; ;;
;;;  ;; In Lisp we "run" again and again in what is really one process
;;;  ;; invocation, so we need to reset the universe on each iteration
;;;  ;;
;;;  (cello-reset 'MGSystem)
;;;  ;;
;;;  ;; At this point the global parameter *sys* is bound to a fresh instance
;;;  ;; of MGSystem, which models the workstation on which the
;;;  ;; app is running.
;;;  ;;
;;;  ;; So let's make a blank window and push it onto the kids cell of *sys*
;;;  ;; (all the kids of an MGSystem are windows). An echo function on the
;;;  ;; kids cell brings the window model to life, including triggering
;;;  ;; its initial display.
;;;  ;;
;;;  (push new-window (kids *sys*))
;;;
;;;  ;;
;;;  ;; now we pump the main event loop
;;;  ;;
;;;
;;;  (do ((ok nil nil))
;;;      ((or (c-stopped)
;;;         (null (kids *sys*))) new-window) ;; when the last window closes, we're outtahere
;;;    (unwind-protect
;;;        (progn
;;;
;;;          ;;-----------------------------------
;;;          ;; process one event (via callbacks)
;;;          ;;
;;;          (glutMainLoopEvent)
;;;
;;;          ;;------------------------------------------------------
;;;          ;; now sleep to give IDE backtrace a chance to get opened 
;;;          ;; on an error. 
;;;          ;;
;;;          (when t ;; (zerop (mod (get-internal-real-time) 4))
;;;            (sleep 0.1))
;;;
;;;          (setf ok t))
;;;      (unless ok
;;;        ;;-----------------------------------------
;;;        ;; event-driven Lisp apps are hard to kill. a runtime error
;;;        ;; may cause the IDE to try to put up a backtrace, which will
;;;        ;; lead to the dying app to be given a fresh event to handle,
;;;        ;; such as deactivate window. The app may well fail at that,
;;;        ;; and we're in loop city. Here we set a global cease-and-desist
;;;        ;; flag, and throughout the app we have sprinkled:
;;;        ;;
;;;        ;;     (unless (c-stopped) <code which does anything interesting>)
;;;        ;;
;;;        (c-stop :run-window-event-loop))))
;;;  (trc "seeya!!!"))
