;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)


;-------------------- resize window ---------------------------
;


;;;(defparameter *resizers* nil)


(defmethod WM-LBUTTONDOWN ((w Window) buttons mousePos)
  (trc nil "WM_LBUTTONDOWN " buttons mousePos)
  (setf (mousePos w) mousePos) ; trigger mouseImage recalc
  (setf (mouseDownEvt w) (make-osEvent
                                :modifiers buttons
                                :where mousePos
                                :realTime (get-internal-real-time))))

(defmethod do-click :around (self osEvent)
  (declare (ignorable osEvent))
  (when self
    (or (call-next-method)
        (do-click (fm-parent self) osEvent))))

(defmethod do-click (self osEvent)
  (declare (ignorable self osEvent))
  nil)

;
; ------------ double click ---------------------------------------
;


;;;(defmethod win::WM_LBUTTONDBLCLK ((w Window) buttons wxwy)
;;;  (declare (ignorable buttons wxwy))
;;;  ;(trc "win::WM_LBUTTONDBLCLK "   buttons wxwy)
;;;  
;;;  (with-window-message (w)
;;;    (let ((osWhere (long-v2 wxwy)))
;;;      (setf (mousePos w) osWhere) ; trigger mouseImage recalc
;;;      (bif (click (doubleClick? w)) ;; won't be if clickee not enabled
;;;          (with-invalidation (w)
;;;            (do-double-click (clickee click) click))
;;;         (bWhen (unclickee (mouseImage w))
;;;            (with-invalidation (w)
;;;              (do-double-click unclickee (make-osEvent
;;;                                          :evtWParam buttons
;;;                                          :evtLParam wxwy
;;;                                          :evtTime (get-internal-real-time)))))))))


(defstruct (osEvent 
           (:conc-name nil))
  modifiers
  where
  realTime)

(defun mk-OSEvent (modifiers where)
  (make-osevent :modifiers modifiers
                :where where
                :realTime (get-internal-real-time)))

(defun evtButtons (osEvent)
  (modifiers osEvent))

(defun evtWhere (osEvent)
  (where osEvent))

(defun evtWhereX (osEvent)
  (declare (optimize (speed 3) (safety 0) (debug 0)))
  ;; (logand (the fixnum (evtLParam osEvent)) (1- 65536))
  (v2-h (evtWhere osEvent)))

(defun evtWhereY (osEvent)
  (declare (optimize (speed 3) (safety 0) (debug 0)))
  (v2-v (evtWhere osEvent)))

(defmethod WM-LBUTTONUP ((w Window) modifiers mousePos)
  (with-metrics (nil nil "win:WM_LBUTTONUP " w modifiers mousePos)
    (setf (mouseUpEvt w) (make-osEvent
                          :modifiers modifiers
                          :where mousePos
                          :realtime (get-internal-real-time)))))

(defparameter *mouseMoveOccupado* nil
  "Vestigial? Under CG/Win32 mouse move could be received during mouse move")

(defparameter *mouseWhere* nil)

(def-Window-callback mg-motion-callback (x y)
  (let ((w (mg-window-current t))
        (where (mkv2 (scr2log x)
                 (scr2log (- y)))))
    (setf *mouseWhere* where)
    (trc nil "motion callback" w x y where *mouseMoveOccupado*)
    (unless (and *mouseMoveOccupado*
              (mousePos w))
      (let ((*mouseMoveOccupado* t)
            #+not (mtr (zerop (mod (get-internal-real-time) 10))))
        (c-assert where)                  
        (with-metrics (nil nil () "Setf mousepos")
          (trc nil "setting mouse pos" where (mod (get-internal-real-time)
                                           (* 10 internal-time-units-per-second)))
          (setf (mousePos w) where)
          (glutpostredisplay)
          )))))

(def-Window-callback mg-passive-motion-callback (x y)
  (let ((w (mg-window-current t)))
      (let ((where (mkv2 (scr2log x)
                     (scr2log (- y)))))
        (setf *mouseWhere* where)
        (trc nil "passive motion callback" w x y where *mouseMoveOccupado*)
        (unless (and *mouseMoveOccupado*
                  (mousePos w))
          (let ((*mouseMoveOccupado* t)
                (mtr nil #+not (zerop (mod (get-internal-real-time) 10))))
            (c-assert where)                  
            (with-metrics (mtr mtr () "Setf mousepos")
              (setf (mousePos w) where)))))))


(def-Window-callback mg-mouse-callback (button up-or-down x y)
  (trc nil "mouse callback entry" button up-or-down x y)
  (let ((w (mg-window-current t))
        (mp (mkv2 (scr2log x)
              (scr2log (- y))))
        (modifiers (glut-get-modifiers)))
    (trc nil "mg-mouse-callback" w button x y)
    (cond
     ((eql button GLUT_LEFT_BUTTON)
      (setf (leftb (mouse *sys*)) (if (eql up-or-down GLUT_DOWN) :down :up))
      (funcall (if (eql up-or-down GLUT_DOWN)
                   #'wm-lbuttondown #'wm-lbuttonup)
        w modifiers mp))

     ((eql button GLUT_MIDDLE_BUTTON)
      (setf (middleb (mouse *sys*)) (if (eql up-or-down GLUT_DOWN) :down :up)))

     ((eql button GLUT_RIGHT_BUTTON)
      (setf (rightb (mouse *sys*)) (if (eql up-or-down GLUT_DOWN) :down :up))
      (when (eql up-or-down GLUT_UP)
        (wm-rbuttondown w modifiers mp)))

     ((eql button GLUT_MOUSE_WHEEL_CLICK)
      (trc "mouse wheel click>" button up-or-down x y))

     ((eql button GLUT_MOUSE_WHEEL_BACK)
      (trc "mouse wheel back>" button up-or-down x y))

     ((eql button GLUT_MOUSE_WHEEL_FWD)
      (trc "mouse wheel>" button up-or-down x y))

     (t (break "unhandled button ~a" (list button up-or-down x y))))))


