;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright (c) 1995,2003 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.


(in-package :cello)

(defmodel CTItemStack (CTSelector IXStackFV)
  ((items :initarg :items :accessor items)
   (max-item-count :initarg :max-item-count :accessor max-item-count))
  (:default-initargs
      :lighting :off
    :layers (list +white+ :fill-lr)
    :selection (c? (when (^visible)
                     (list (elt (^kids) (^selection .inspection.)))))
    :lR (c? (^fill-right window)) ;; ugh, active even when not visible
    :items (c? (when (^visible)
                 (un-dot (target .inspection.))))
    :max-item-count (c? (max (or .cache 0) (length (^items))))
    :kidvalues (c? (loop for n below (^max-item-count) collect n))
    :kvkey #'item-no
    :kidfactory (lambda (item-stack item-no)
                  (declare (ignorable item-stack))
                  (mkPart :iteminfo (item-inspector)
                    :item-no item-no))))

(defmethod un-dot ((c cons))
  (cons (car c) (un-dot (cdr c))))

(defmethod un-dot (other) (when other (list other)))

(def-c-echo selection ((self CTItemStack))
  (when (visible self)
    (setf (focus .w.) (car new-value))))

(defmethod do-specialkeydown ((self CTItemStack) k event)
  (declare (ignorable event))
  (case k
    (#.GLUT_KEY_UP
     (when (plusp (selection .inspection.))
       (decf (selection .inspection.))))
    (#.GLUT_KEY_DOWN		
     (when (< (1+ (selection .inspection.))
             (length (items self)))
       (incf (selection .inspection.))))
    (#.GLUT_KEY_PAGE_UP		(trc "fm-parent!"))
    (#.GLUT_KEY_PAGE_DOWN		(trc "first kid!"))
    (#.GLUT_KEY_HOME		(trc "first slot"))
    (#.GLUT_KEY_END			(trc "last slot"))
    (otherwise (return-from do-specialkeydown nil)))
  t) ;; handled

(defmodel item-inspector (Focus CTSelectable IXText)
      ((item-no :initform nil :reader item-no :initarg :item-no))
      (:documentation "Inspector sub-view of a CLOS slot (the md-value)")
      (:default-initargs
          :spacing (u8ths 2)
        :visible (c? (< (item-no self)
                       (length (items (upper self CTItemStack)))))
        :enabled (c? (^visible))
        :md-value (c? (nth (item-no self) (items (upper self CTItemStack))))
        :text$ (c? (format nil "~a" (^md-value)))
        :kbSelector #\return
        :controlaction (lambda (self event)
                         (declare (ignorable event))
                         (trc nil "return->item inspect" self (md-value self))
                         (setf .inspection. (make-instance 'inspection
                                                  :inspector .inspector.
                                                  :init-target (md-value self))))
        ))