;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright (c) 1995,2003 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.


(in-package :cello)

(defmodel Clos-slots-view (tab-bar-tracker IXStack)
  ()
  (:default-initargs
      :kids (list (aStack (:lR (c? (^fill-right window))
                            :spacing (u8ths 2))
                    (aRow (:spacing (u16ths 1) :justify :center)
                      (aLabel "Class:" :layers +white+)
                      (mkPart :class (cttext)
                        :md-value (c? (class-of (target .inspection.)))
                        :text$ (c? (string-capitalize (symbol-name (class-name (^md-value)))))
                        :layers +green+
                        :controlAction (lambda (self event)
                                         (declare (ignorable event))
                                         (setf .inspection.
                                           (make-instance 'inspection
                                             :init-target (md-value self))))))
                    (make-slider :htabs
                      :bar-color +yellow+
                      :height (upts 8)
                      :width (c? (lwidth .parent))
                      :initial-pcts (list (mkv2 0 0) (mkv2 0.3 0)))
                   
                    (mkPart :slotstack (CTSlotStack))))))

(defmethod md-awaken :after ((self clos-slots-view))
  (assert (tab-bar self)))

(defmodel CTSlotStack (CTSelector IXStackFV)
  ((slots :initarg :slots :accessor slots)
   (max-slot-count :initarg :max-slot-count :accessor max-slot-count))
  (:default-initargs
      :visible (c? (visible (upper self Clos-slots-view)))
    :selection (c? (when (^visible)
                     (list (elt (^kids) (^selection .inspection.)))))
    :pX (u8ths 1)
    :lR (c? (^fill-right window)) ;; ugh, active even when not visible
    :slots (c? (when (^visible)
                 (bwhen (tgt (target .inspection.)) ;; should be if visible
                   (sort (copy-list (class-slots (class-of tgt))) ;; (subseq  0 20)
                     #'string< :key #'slot-definition-name))))
    :max-slot-count (c? (max (or .cache 0) (length (^slots))))
    :kidvalues (c? (loop for n below (^max-slot-count) collect n))
    :kvkey #'md-value
    :KID-SLOTS nil
    ;;:layers (list :primary +red+ :line-lr)
    :kidfactory (lambda (slot-stack slot-no)
                  (declare (ignorable slot-stack))
                  (cells::make-part (gensym "SLOT-INFO") 'slot-inspector
                    :slot-no slot-no
                    :Px (U8THS 1)
                    :pY (* slot-no (upts -18))))))

(def-c-echo selection ((self CTSlotStack))
  (when (visible self)
    (setf (focus .w.) (car new-value))))

(defmethod do-specialkeydown ((self CTSlotStack) k event)
  (declare (ignorable event))
  (case k
    (#.GLUT_KEY_UP
     (when (plusp (selection .inspection.))
       (decf (selection .inspection.))))
    (#.GLUT_KEY_DOWN		
     (when (< (1+ (selection .inspection.))
             (length (slots self)))
       (incf (selection .inspection.))))
    (#.GLUT_KEY_PAGE_UP		(trc "fm-parent!"))
    (#.GLUT_KEY_PAGE_DOWN		(trc "first kid!"))
    (#.GLUT_KEY_HOME		(trc "first slot"))
    (#.GLUT_KEY_END			(trc "last slot"))
    (otherwise (return-from do-specialkeydown nil)))
  t) ;; handled