;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright (c) 1995,2003 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.


(in-package :cello)

(defmodel inspection ()
  ((init-target :initarg :init-target :reader init-target)
   (target :initarg :target :reader target
     :initform (c? (bif (it (^init-target))
                     (if (and (consp it)(null (cdr it)))
                         (car it) it)
                     *sys*)))
   (selection :initform (cv 0) :accessor selection)))

(defmethod print-object ((self inspection) s)
  (format s "<Inspect ~a ~a>" (type-of (target self)) (target self)))

(define-symbol-macro .inspector.
    (nearest self inspector-frame))

(define-symbol-macro .ihistory.
    (history (nearest self inspector-frame)))

(define-symbol-macro .inspection.
    (inspection (nearest self inspector-frame)))

(defmethod do-menu-command ((w Window) (i Inspection))
  "Support History menu in toolbar"
  (trc "destroying window" w (glutw w))
  (setf (inspection w) i))

(defun ensure-inspection-target-view (w target)
  (let ((k (inspection-target-key target)))
    (or (cdr (assoc k (target-views w)))
      (let ((v (inspection-target-view k)))
        (push (cons k v) (target-views w))
        v))))

(defun inspection-target-key (target)
  (typecase target
    (cons :list-items)
    (structure-object :struct-slots)
    (standard-object :clos-slots)
    (otherwise :plain-label)))

(defun inspection-target-view (target-key)
  (ecase target-key
    (:plain-label (mkPart :ivlbl (IXText)
                   :text$ (c? (when (^visible)
                                (string-downcase
                                 (princ-to-string (target .inspection.)))))))
    (:list-items (mkPart :list-items (CTItemStack)))
    (:struct-slots (mkPart :struct-slots (struct-slots-view)))
    (:clos-slots (mkPart :clos-slots (Clos-slots-view)))))