;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright (c) 1995,2003 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.


(in-package :cello)

(defmacro ins-nav (name title vrule  caction)
  `(mkPart ,name (CTButton)
      :md-value ,vrule
      :enabled (c? (^md-value))
      :lb (c? (^lb-height (downs 36)))
      :style :inspector
      :title$ ,title
      :controlAction (lambda (self event)
                       ,caction)))

(defun inspector-toolbar ()
  (aRow (:spacing (u8ths 1) :lighting :on :outset (upts 4))
    (mkPart :genealogy (ins-family-tree))
    

    (ins-nav :back "<"
      (c? (bWhen (i .inspection.)
            (cadr (member i .ihistory.))))
      (do-specialkeydown .inspector. GLUT_KEY_LEFT event))
      
    (ins-nav :forth ">"
      (c? (bWhen (i .inspection.)
            (find-prior i .ihistory.)))
      (do-specialkeydown .inspector. GLUT_KEY_RIGHT event))

    (mkPart :history (Inspector-menu)
      :items (c? (mapcar (lambda (h)
                           (cons (format nil "~a" (target h))
                             h))
                   .ihistory.)))
    ))
(defmodel CTImage (Control Image)())

(defmodel ins-family-tree (IXStack)
  ()
  (:default-initargs
      :justify :center
    :lighting :off
    :outset (upts 4)
    :layers (c? (list :on
                  (make-instance 'Frame3D
                    :3dStyle :EDGE-SUNKEN
                    :of-window .w.
                    :w-box (gbox self)
                    :thickness (uPts 4)
                    )
                  :off))
    :md-value (c? (bwhen (i .inspection.)
                    (target i)))
    :visible (c? (^kids))
    :spacing (upts 2)
    :lR (uPts 74)
    :kids (c? (let ((tgt (^md-value)))
                (when (typep tgt 'cells::model-object)
                  (let* ((par (fm-parent tgt))
                        (gpar (when par (fm-parent par))))
                    (flet ((ft-ctl (it color)
                             (mkPart :ft-ctl (CTImage)
                               :lighting :off
                               :layers (list :off (gloff GL_BLEND)
                                         (gloff gl_texture_2d) color :fill-lr)
                               :ll 0 :lt 0 :lr (u8ths 1) :lb (downs (u8ths 1))
                               :controlAction (lambda (self event)
                                                (declare (ignorable event))
                                                (setf .inspection.
                                                  (make-instance 'inspection
                                                    :inspector .inspector.
                                                    :init-target it))))))
                      (theKids
                       (arow (:spacing (u16ths 1)  :lb (downs (u8ths 1)))
                         (when par
                           (if gpar
                               (mapcar (lambda (psib)
                                         (ft-ctl psib (if (eql psib par)
                                                          +blue+
                                                        +lt-gray+)))
                                 (kids gpar))
                             (ft-ctl par +blue+))))
                       (arow (:spacing (u16ths 1)  :lb (downs (u8ths 1)))
                         (if par
                             (mapcar (lambda (sib)
                                       (ft-ctl sib (if (eql sib tgt)
                                                       +green+
                                                     +lt-gray+)))
                               (kids par))
                           (ft-ctl tgt +green+)))
                       (arow (:spacing (u16ths 1)  :lb (downs (u8ths 1)))
                         (mapcar (lambda (k)
                                   (ft-ctl k +yellow+))
                           (kids tgt)))))))))))

(defmodel Inspector-menu (CTmenu)
  ((items :initarg :items :reader items :initform nil))
  (:default-initargs
      :lighting :off
      :kids (c? (theKids
                    (mkPart :title-items (IXStack)
                      :lighting :off
                      :kids (c? (theKids
                                 (mkPart :title (IXText)
                                   :lighting :off
                                   :layers (list +white+ :fill-lr +black+)
                                   #+not (c? (let ((m (upper self CTMenu)))
                                                 (if (and (mouse-over-p m)
                                                       (eql :down (leftb (mouse *sys*))))
                                                     (list +black+ :fill-lr +white+)
                                                   (list +white+ :fill-lr +black+))))
                                   :textfont (c? (ftgl-font-ensure .w. 'texture 'arialn 14 24))
                                   :text$ (c? (bwhen (i .inspection.)
                                                    (trc "target i" (target i))
                                                    (princ-to-string (target i))
                                                    )))
                                 (mkPart :menu-socket (WindowSocket)
                                   :collapsed t
                                   :window-factory (lambda (socket glut-xy)
                                                     (declare (ignorable socket glut-xy))
                                                     (c-break "menu offset ~a" glut-xy)
                                                     (mkPart :menu-window (CTMenuItems)
                                                       ;;:visible t
                                                       ;;:lR (uin 2)
                                                       :glut-xy glut-xy))
                                   :gen-window-p nil)
                                 )))))))