;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright (c) 1995,2003 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

(defun cloucell ()
  (mkpart :cloucell (inspector-frame)
    :pX 0 :pY 0
    :inspection (cv (make-instance 'inspection
                      :init-target (c? *sys*)))))

(defmodel inspector-frame (ixstack)
    ((inspection :initarg :inspection :accessor inspection)
     (target-views :cell nil :initform nil :initarg :target-views :accessor target-views)
     (history :initarg :history :accessor history
        :initform (cv nil)))
    (:default-initargs
        :title$ "Cello Inspector"
      ;;:lighting :on
      :inspection (cv nil)
      :spacing (u8ths 1)
      ;;:layers (list +white+ :fill-lr)
      :kids (c? (theKids
                   (inspector-toolbar)
                   (mkPart :tgtcontainer
                     (target-view-container))))))

(def-c-echo inspection ((self inspector-frame) current-i prior-i)
    "Maintain navigation history"
    (trc "echo inspection" self current-i (history self))
    (when current-i
      (let ((h (history self)))
        (unless (find current-i h)
          (setf (history self)
              (cons current-i
                (or (when prior-i
                      (member prior-i h)) ;; toss "forward" chain from prior-i
                  h)))))))

(defmodel target-view-container (igzerotl)
  ()
  (:default-initargs
      :kid-slots (def-kid-slots
                     (mk-kid-slot (px :ifMissing t)
                       (c? (^px-maintain-pL 0)))
                     (mk-kid-slot (py :ifMissing t)
                       (c? (^py-maintain-pT 0)))
                   (mk-kid-slot (visible :ifmissing nil)
                     (c? (bwhen (i .inspection.)
                           (eql self (cdr (assoc (inspection-target-key (target i))
                                            (target-views .inspector.))))))))
    :kids (c? (assert .inspection. () "Gotta start somewhere")
            (pushnew (ensure-inspection-target-view
                      .inspector. (target .inspection.)) .cache))))

(defmethod do-specialkeydown ((w inspector-frame) k event)
  (declare (ignorable event))
  (case k
    (#.GLUT_KEY_RIGHT
     (let (fwd)
       (dolist (i (history w))
         (if (eq i (inspection w))
           (progn
             (when fwd
               (setf (inspection w) fwd))
             (return-from do-specialkeydown t))
           (setq fwd i)))))

    (#.GLUT_KEY_LEFT
     (bwhen (back (cadr (member (inspection w) (history w))))
       (setf (inspection w) back)))

    (otherwise (return-from do-specialkeydown nil)))
  t)

(defmethod do-right-button :around ((self Image) buttons wxwy)
  (declare (ignorable buttons wxwy))
  (trc "do-right-button entry" self)
  (if (not (control-key-down buttons))
      (call-next-method)
    (push (make-instance 'window
            :kids (let ((i self))
                    (c? (List (mkpart :ccell (inspector-frame)
                                :pX 0 :pY 0
                                :inspection (cv (make-instance 'inspection
                                                  :init-target i)))))))
      (kids *sys*))))
