;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright (c) 1995,2003 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.


(in-package :cello)

#|
Widgets schmidgets, let's use the keyboard.

First, we allow the user to "select" a slot by using the up/down
cursor keys. <enter> inspects the selected slot's value.
left/right translate to moving back and forth over the history
of inspected targets. <escape> closes the inspector (and any
other Cellow window, it so happens, during this development
phase).

|#

(defun findlist (cons list)
  "Mostly a test to see if the cons is in the list"
  (when cons
    (mapl (lambda (listcons)
               (when (eq cons listcons)
                 (return-from findlist cons)))
      list)
    nil))

#+when-ready
(progn
  (defmodel inspector-window (Window)
    (
     (new-target :initarg :new-target :initform (cv nil) :accessor new-target)
     (targets :initarg :targets :initform nil :accessor targets)
     )
    (:default-initargs
        :title$ "Cello Inspector"
      :md-value (c? (car (^new-target)))
      :menus nil
      :lighting :off
      :layers (list +white+ :fill-lr)
      :content (inspector-window-content)
      ))
  (defmethod initialize-instance :after ((self inspector-window) &key)
    (unless (new-target self)
      (setf (new-target self) (list self))))

  (defmethod do-right-button :around ((self Image) buttons wxwy)
  (declare (ignorable buttons wxwy))
  (if (not (control-key-down buttons))
      (call-next-method)
    (push (make-instance 'inspector-window
            :md-value (cv (list self)))
      (kids *sys*))))

  (def-c-echo new-target ((w inspector-window))
    "Maintain navigation history"
    (trc nil "echo md-value" w new-value (targets w))
    (let ((tgts (targets w)))
      (when new-value
        (if (null tgts)
            (setf (targets w) new-value)
          (unless (findlist new-value tgts)
            (if (findlist old-value (targets w))
                (progn
                  (trc "bingo nconcing targets" new-value)
                  (rplacd old-value new-value) ;; lose any "forward" history
                  )
              (setf (targets w) new-value)))))))
  
  (defmethod do-specialkeydown :around ((w inspector-window) k event)
    (declare (ignorable event))
    (case k
      (#.GLUT_KEY_DOWN		
       (trc "selnext")
       (let ((selector (fm-other :slots :starting w))
             use-next)
         (maplist (lambda (slots)
                    (cond
                     (use-next
                      (setf (selection selector) (list (car slots)))
                      (return-from do-specialkeydown t))

                     ((eql (car (selection selector)) (car slots))
                      (setq use-next t))))
           (kids selector))))
      (otherwise (call-next-method))))
  
  (defmethod do-keydown :around ((w inspector-window) k event)
    (declare (ignorable event))
    (case k
      (#\Return (setf (new-target w)
                  (list (si-value (car (selection (fm-other :slots :starting w)))))))
      (otherwise (call-next-method)))))

#+when-ready
(progn
  (defmethod do-specialkeydown :around ((w inspector-window) k event)
    (declare (ignorable event))
    (case k
      (#.GLUT_KEY_LEFT
       (let (prior)
         (maplist (lambda (tgts)
                    (cond
                     ((eq (md-value w) tgts)
                      (when prior
                        (setf (md-value w) prior))
                      (return-from do-specialkeydown t))
                     (t (setq prior tgts))))
           (targets w))))
      (#.GLUT_KEY_RIGHT
       (when (cdr (md-value w))
         (setf (md-value w)(cdr (md-value w)))))
      (#.GLUT_KEY_UP
       (let ((selector (fm-other :slots :starting w))
             prior)
         (maplist (lambda (slots)
                    (cond
                     ((eql (car (selection selector)) (car slots))
                      (setf (selection selector) (list prior))
                      (return-from do-specialkeydown t))
                     (t (setq prior (car slots)))))
           (kids selector))))
      (#.GLUT_KEY_DOWN		
       (let ((selector (fm-other :slots :starting w))
             use-next)
         (maplist (lambda (slots)
                    (cond
                     (use-next (setf (selection selector) (list (car slots)))
                       (return-from do-specialkeydown t))
                     ((eql (car (selection selector)) (car slots))
                      (setq use-next t))))
           (kids selector))))
      (#.GLUT_KEY_PAGE_UP		(trc "fm-parent!"))
      (#.GLUT_KEY_PAGE_DOWN		(trc "first kid!"))
      (#.GLUT_KEY_HOME		(trc "first slot"))
      (#.GLUT_KEY_END			(trc "last slot"))
      (otherwise (call-next-method)))))