;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright (c) 1995,2003 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

(defmodel slot-inspector (Focus CTSelectable IXtext)
  ((slot-no :cell nil :initform nil :reader slot-no :initarg :slot-no)
   (slot-def :reader slot-def
     :initform (c? (nth (slot-no self) (slots (upper self CTSlotStack)))))
   (slot-name :reader slot-name :initarg :slot-name
     :initform (c? (bwhen (sdef (^slot-def))
                     (slot-definition-name sdef))))
   (si-value :reader si-value :initarg :si-value
     :initform (c? (bwhen (slot-name (^slot-name))
                     (let ((tgt (target .inspection.)))
                       (if (slot-boundp tgt slot-name)
                           (if (typep tgt 'cells::model-object)
                               (cells::md-slot-value tgt slot-name)
                             (slot-value tgt slot-name))
                         "<unbound>")))))
   )
  (:documentation "Inspector sub-view of a CLOS slot (the md-value)")
  (:default-initargs
      :tab-bar (c? (tab-bar (upper self tab-bar-tracker)))
    :enabled (c? (^si-value))
    :lT 14 :lB -7
    :inset (mkv2 (u8ths 1) 0)
    :lR (c? (^fillparentright (u96ths 6)))
    :lighting :off
    :visible (c? (^slot-def)) ;; ie, current target has more slots than slot-no
    :enabled (c? (^slot-def))
    :layers (c? (if (^selectedp)
                    (list +yellow+ :fill-lr +black+)
                  +white+))
    :textfont (c? (ftgl-font-ensure .w. 'texture 'cour 14))
    :text$ (c? (with-billing-outer (:si-text)
                 (bwhen (sn (^slot-name))
                   (nstring-downcase
                    #+not (concatenate 'string (symbol-name sn) " : " (princ-to-string
                                                                       (^si-value)))
                    (format nil "~a~1,16T~a" (symbol-name sn)
                      (with-billing-adj
                          (^si-value)))))))

    :kbSelector #\return
    :controlaction (lambda (self event)
                     (declare (ignorable event))
                     (trc "return->slot inspect" self (si-value self))
                     (setf .inspection. (make-instance 'inspection
                                              :inspector .inspector.
                                              :init-target (si-value self))))
    ))
;;;:layers +black+ #+not (c? (let ((tgt (target .inspection.)))
;;;                            (or (when (typep tgt 'cells::model-object)
;;;                                  (bwhen (c (cells::md-slot-cell-type
;;;                                             (class-name (class-of tgt)) (^slot-name si)))
;;;                                    (case c
;;;                                      (:ephemeral +red+)
;;;                                      (t +blue+))))
;;;                              +black+))))
;;;:layers (c? (let ((tgt (target .inspection.)))
;;;(or (when (and (typep tgt 'cells::model-object)
;;;            (cells::md-slot-cell-type
;;;             (class-name (class-of tgt)) (^slot-name si)))
;;;      (bIf (c (cells::md-slot-cell tgt (^slot-name si)))
;;;        (etypecase c
;;;          (cells::c-variable +red+)
;;;          (cells::c-dependent
;;;           (ecase (cells::c-state c)
;;;             (:valid +green+)
;;;             (:unbound +lt-gray+))))
;;;        (when (cells::md-slot-cell-flushed tgt (^slot-name si))
;;;          +blue+)))
;;;  +black+)))))))
