;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cl-user; -*-

(in-package :cl-user)

;;; ***********************************************************************
;;; Begin configuration section
;;;
;;; Before building Cells, review and customize the following settings:
;;;
(let (
      ;; Step 1
      ;; ------
      ;; The path to ASDF.  This is only necessary if it's not already loaded.
      ;;
      ;; Examples:
      ;;   Unix: (asdf-path (pathname "/usr/local/src/asdf.lisp"))
      ;;   Windows: (asdf-path (pathname "C:\\dev\\asdf.lisp"))
      ;;   Windows: (asdf-path (make-pathname :directory '(:absolute "dev")
      ;;                                      :name "asdf" :type "lisp"))
      (asdf-path nil)

      ;; Step 2
      ;; ------
      ;; The path to the Cells source directory.
      ;;
      ;; Examples:
      ;;   Unix: (cells-path (pathname "/usr/local/src/cells/"))
      ;;   Windows: (cells-path (pathname "C:\\dev\\cells\\"))
      ;;   Windows: (cells-path (make-pathname
      ;;                          :directory '(:absolute "dev" "cells")
      ;;                          #+lispworks :host #-lispworks :device
      ;;                          "C"))
      (cells-path nil)

      ;; Step 3
      ;; ------
      ;; Decide if you want to load and run the regression test suite.
      ;; If you want to validate the system or explore the test suite,
      ;; some of which is heavily annotated, set TESTP to T
      (testp t)
      )

;;; Yer done
;;;
;;; End configuration section
;;; ***********************************************************************
  

  ;; Ensure ASDF is loaded
  #-asdf
  (progn (assert (not (null asdf-path))
		 (asdf-path)
		 "ASDF is not loaded, and ASDF-PATH was not supplied.  Please edit build.lisp")
	 (load asdf-path))

  ;; Build Cells.
  (load (merge-pathnames "build-sys.lisp" cells-path))
  (funcall (intern "BUILD-SYS" "CELLS-BUILD-PACKAGE")
	   :force t :source-directory cells-path)

  ;; Load and run the test suite, if requested.
  (when testp
    (funcall (intern "BUILD-SYS" "CELLS-BUILD-PACKAGE")
	     :force t
	     :source-directory (merge-pathnames
				(make-pathname :directory '(:wild "cells-test"))
				cells-path))
    (funcall (intern "CV-TEST" "CELLS")))

  ;; Remove build package
  (delete-package "CELLS-BUILD-PACKAGE"))
