;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cells; -*-
;;;
;;; Copyright (c) 1995,2003 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cells)

;____________________________ cell calculate and set ___________________
;

(defun c-calculate-and-set (c)
  (when (c-stopped)
    (princ #\.)
    (return-from c-calculate-and-set))
  
  (c-assert (not (cmdead c)))

  (when (find c *c-calculators*) ;; circularity
    (c-stop :c-calculate-and-set-circ-ask)
    (trc "md-slot-value breaking on circularity" c *c-calculators*)
    (break ;; problem when testing cells on some CLs
     "cell ~a midst askers: ~a" c *c-calculators*))
                  
  (count-it :c-calculate-and-set)
  ;;;  (count-it :c-calculate-and-set (type-of (c-model c))) ;; (c-slot-name c))

  ;;(with-metrics (nil nil () "calc n set" (c-slot-name c) (c-model c))
  (progn ;; wtrc (0 200 "calc n set" (c-slot-name c) (c-model c))
    (cd-usage-clear-all c)
    
    (let ((mycalc (incf (cr-rethinking c) 1))
          (newvalue (let ((*c-calculators* (cons c *c-calculators*))
                          *synapse-factory* ;; clear, then if desired each access to potential other cell must estab. *synapse-factory*
                          )
                      (c-assert (c-model c))
                      #+not (when (plusp *trcdepth*) 
                        (format t "ccalcnset> calcing ~a calcers ~a" c *c-calculators*))
                      (funcall (cr-rule c) c))))
      #+nahhh (when (> (cr-rethinking c) 1)
        (c-break "rethink-reentry ~a" c))
      #+notso (c-assert (not (typep newvalue 'cell)) ()
                "new value for cell ~s is itself a cell: ~s. probably nested (c? ... (c? ))"
                c newvalue)
      (when (and *c-debug* (typep newvalue 'cell))
        (trc "new value for cell ~s is itself a cell: ~s. probably nested (c? ... (c? ))"
                c newvalue))
      (when (< mycalc (cr-rethinking c))
        ;;
        ;; means we re-entered rule and managed to compute without re-entering under new circumstances
        ;; of later entry. use later calculation result..
        (trc nil "calc-n-set > breaking off, not lg" c)
        ;;
        (c-assert (c-validp c))
        (return-from c-calculate-and-set (c-value c)))
      
     (c-unlink-unused c)

     (md-slot-value-assume (c-model c)
                            (c-slot-spec c)
                            (c-absorb-value c newvalue)))))

#+test
(loop for useds on '(1 2 3 4 5)
    for used = (car useds)
    for mapn upfrom 5 
    when (oddp used)
    do (print (list useds mapn))(print used))

(defun c-unlink-unused (c &aux (usage (cd-usage c)))
  (loop for useds on (cd-useds c)
        for used = (car useds)
        for mapn upfrom (- *cd-usagect* (length (cd-useds c)))
        when (zerop (sbit usage mapn))
        do
        (c-assert (not (minusp mapn)))
        (c-assert (< mapn *cd-usagect*))
        (if (typep used 'synapse)
            (progn
              (setf (syn-relevant used) nil) ;; 030826synfix
              )
          (progn
            (trc nil "dropping unused" used :mapn-usage mapn usage)
            (c-unlink-user used c)
            (rplaca useds nil))))
  (setf (cd-useds c) (delete-if #'null (cd-useds c))))


