;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cells; -*-
;;;
;;; Copyright (c) 1995,2003 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cells)

(defun slot-spec-name (slot-spec)
  slot-spec)


(defstruct (cell (:conc-name c-))
  waking-state
  model
  mdinfo
  slot-spec
  value
  (users nil :type list))

(defun test ()
  (let (x)
    (makunbound x)
    x))

(defun c-slot-name (c)
  (slot-spec-name (c-slot-spec c)))

(defun c-validate (self c)
  (when (not (and (c-slot-spec c) (c-model c)))
    (format t "~&unadopted cell: ~s md:~s, awake:~s" c self (c-waking-state self))
    (c-break "unadopted cell ~a ~a" self c)
    (error 'c-unadopted :cell c)))

(defmethod c-when (other)
	(declare (ignorable other)) nil) ;; /// needs work

(defstruct (synapse
            (:conc-name syn-))
  user
  used
  (relevant t) ;; not if unused during subsequent eval. but keep to preserve any state
  fire-p
  relay-value)

(defmacro mksynapse ((&rest closeovervars) &key trcp fire-p relay-value)
  (let ((used (copy-symbol 'used)) (user (copy-symbol 'user)))
    `(lambda (,used ,user)
       ,(when trcp
          `(trc "making synapse between user" ',trcp ,user :and :used ,used))
       (let (,@closeovervars)
         (make-synapse
          :used ,used
          :user ,user
          :fire-p ,fire-p
          :relay-value ,relay-value)))))

(defmethod print-object ((syn synapse) stream)
  (format stream "{syn ~s ==> ~s" (syn-used syn) (syn-user syn)))

(defmethod c-true-stalep ((syn synapse))
  (cd-stale-p (syn-user syn)))

(defstruct (c-variable
            (:include cell)
            (:conc-name cv-))
  cyclic-p
  setting-p)

(defstruct (c-ruled
            (:include cell)
            (:conc-name cr-))
  (state :unbound :type symbol)
  (rethinking 0 :type number)
  lazy
  rule)

(defmethod c-lazy-p ((c c-ruled)) (cr-lazy c))
(defmethod c-lazy-p (c) (declare (ignore c)) nil)

(defun c-optimized-away-p (c)
  (eql :optimized-away (c-state c)))

;----------------------------

(defmethod c-true-stalep (c)
  (declare (ignore c)))

(defmethod trcp-slot (self slot-name)
  (declare (ignore self slot-name)))

(defconstant *cd-usagect* 64)

(defstruct (c-dependent
            (:include c-ruled)
            (:conc-name cd-))
  (useds nil :type list)
  (code nil :type list) ;; /// feature this out on production build
  (usage (make-array *cd-usagect* :element-type 'bit
                        :initial-element 0) :type vector)
  stale-p
  )

;;;(defmethod trcp ((c c-dependent))
;;;  (or (trcp-slot (c-model c) (c-slot-name c))
;;;    ;;(c-lazy-p c)
;;;    nil))

(defmethod c-true-stalep ((c c-dependent))
  (cd-stale-p c))

(defstruct (c-stream
            (:include c-ruled)
            (:conc-name cs-))
  values)

;;; (defmacro cell~ (&body body)
;;;   `(make-c-stream
;;;     :rule (lambda ,@*c-lambda*
;;;                 ,@body)))

(defstruct (c-drifter
            (:include c-dependent)))

(defstruct (c-drifter-absolute
            (:include c-drifter)))

;_____________________ accessors __________________________________


(defun (setf c-state) (new-value c)
  (if (typep c 'c-ruled)
      (setf (cr-state c) new-value)
    new-value))

(defun c-state (c)
  (if (typep c 'c-ruled)
      (cr-state c)
    :valid))

(defun c-unboundp (c)
  (eql :unbound (c-state c)))

(defun c-validp (c)
  (find (c-state c) '(:valid :optimized-away)))

;_____________________ print __________________________________

(defmethod print-object :before ((c c-variable) stream)
  (declare (ignorable c))
  (format stream "[var:"))

(defmethod print-object :before ((c c-dependent) stream)
	(declare (ignorable c))
  (format stream "[dep~a:" (cond
                            ((null (c-model c)) #\0)
                            ((eq :eternal-rest (md-state (c-model c))) #\_)
                            ((cd-stale-p c) #\#)
                            ((sw-pending c) #\?)
                            (t #\space))))

(defmethod print-object ((c cell) stream)
  (c-print-value c stream)
  (format stream "=~a/~a]"
    (symbol-name (or (c-slot-name c) :anoncell))
    (or (c-model c) :anonmd)))

;__________________

(defmethod c-print-value ((c c-ruled) stream)
  (format stream "~a" (cond ((c-validp c) "<vld>")
                            ((c-unboundp c) "<unb>")
                            ((cd-stale-p c) "<obs>")
                            (t "<err>"))))

(defmethod c-print-value (c stream)
  (declare (ignore c stream)))

;____________________ constructors _______________________________

(defmacro c-lambda (&body body)
  (let ((c (gensym)))
    `(lambda (,c &aux (self (c-model ,c))
               (.cache (c-value ,c)))
       (declare (ignorable .cache self))
       (c-assert (not (cmdead ,c)) "cell dead entering rule ~a" ,c)
       ,@body)))

(defmacro c? (&body body)
  `(make-c-dependent
    :code ',body
    :rule (c-lambda ,@body)))

(defmacro c?_ (&body body)
  `(make-c-dependent
    :code ',body
    :lazy t
    :rule (c-lambda ,@body)))

(defmacro c?? ((&key (tagp nil) (in nil) (trigger nil) (out t))&body body)
  (let ((result (copy-symbol 'result))
        (thetag (gensym)))
     `(make-c-dependent
       :code ',body
       :rule (c-lambda
              (let ((,thetag (gensym "tag"))
                    (*trcdepth* (1+ *trcdepth*))
                    )
                (declare (ignorable self ,thetag))
                ,(when in
                   `(trc "c??> entry" (c-slot-name c) (c-model c) (when ,tagp ,thetag)))
                ,(when trigger `(trc "c??> trigger" *cause* c))
                (count-it :c?? (c-slot-name c) (md-name (c-model c)))
                (let ((,result (progn ,@body)))
                  ,(when out `(trc "c?? result:" ,result (c-slot-name c) (when ,tagp ,thetag)))
                  ,result))))))

(defmacro cv (defn)
  `(make-c-variable
    :value ,defn))

(defmacro cv8 (defn)
  `(make-c-variable
    :cyclic-p t
    :value ,defn))

(defmacro c... ((value) &body body)
  `(make-c-drifter
    :code ',body
    :value ,value
    :rule (c-lambda ,@body)))

(defmacro c-abs (value &body body)
  `(make-c-drifter-absolute
    :code ',body
    :value ,value
    :rule (c-lambda ,@body)))


(defmacro c-envalue (&body body)
  `(make-c-envaluer
    :envaluerule (c-lambda ,@body)))
