;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cells; -*-
;;;
;;; Copyright (c) 1995,2003 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cells)

(defparameter *df-interference-detection* t)

(eval-when (compile eval load)
  (export '(*df-interference-detection*)))

(defmethod sw-detect-interference (user trigger)
  (declare (ignorable trigger))
  (when #+runtime-system t #-runtime-system *df-interference-detection*
    (trc nil "detect entry" user (cd-useds user))
    (dolist (used (cd-useds user))
      (do ((deep-stale (cd-deep-stale used)(cd-deep-stale used)))
          ((null deep-stale))
        ;;(trc nil "sw-detect-interference trying deep stale" deep-stale)
        (c-rethink deep-stale)
        (cond
         ((c-true-stalep deep-stale)
          (trc "!! true deep stalep: user>" user)
          (trc "!! true deep stalep: used>" used)
          (trc "!! true deep stalep: deepstale>" deep-stale)
          (return-from sw-detect-interference deep-stale #+debugging (list user used deep-stale)))

         ((not (c-true-stalep user))
          (return-from sw-detect-interference nil)))))))

(defmethod sw-detect-interference ((user c-variable) trigger)
  (declare (ignore trigger)))

(defmethod sw-detect-interference ((user synapse) trigger)
  (sw-detect-interference (syn-used user) trigger))

(defmethod cd-deep-stale ((c c-dependent))
  (trc nil "cd-deep-stale entry" c)
  (if (cd-stale-p c)
      c ;; (eko ("deep stalep bingo !!!!!!") c)
    (some #'cd-deep-stale (cd-useds c))))

(defmethod cd-deep-stale ((syn synapse))
  (cd-deep-stale (syn-used syn)))

(defmethod cd-deep-stale (c)
  (declare (ignore c)))

(defparameter *sw-pending* nil)
(defparameter *dataflowing* nil)

(defun dump-pending ()
  (dotimes (x (length *sw-pending*))
    (let ((p (nth x *sw-pending*)))
      (destructuring-bind (heldup . holdup) p
        (declare (ignorable holdup))
        (trc heldup "        pending!!!!!!!!!!" p)
        )))
  )

;; mo better diags: holdup (c-true-stalep holdup) heldup (c-true-stalep heldup))))))

(defun call-with-dataflow-management (c-originating bodyfn)
  (declare (ignorable c-originating))
  (if *dataflowing*
      (progn
        (setf (cv-setting-p c-originating) t)
        (prog1
            (funcall bodyfn)
          (setf (cv-setting-p c-originating) nil)))
    
    (let ((*dataflowing* t)
          *sw-pending*)
      #+dfdbg (trc nil ">>>>> with-dataflow-management: 001" c-originating)
      (setf (cv-setting-p c-originating) t)
      (prog1
          (funcall bodyfn)
        
        (loop while (and *sw-pending*
                      (not *sw-looping*))
              do
              
              #+dfdbg
              (progn
                (trc nil "we have pending!!!!!!!!!!" (length *sw-pending*))
                (dump-pending))
              
              (let ((pct (length *sw-pending*))
                    (oldpending (copy-list *sw-pending*)))
                ;;(trace c-rethink)
                (labels ((do-last (pending)
                           (when pending
                             (do-last (cdr pending))
                             ;;(trace c-rethink cd-deep-stale sw-detect-interference)
                             (destructuring-bind (heldup . holdup) (car pending)
                               (trc heldup "pending sweep sees held up" heldup :holdup holdup)
                               (c-assert (find heldup (cells (c-model heldup)) :key #'cdr))
                               (c-assert (find holdup (cells (c-model holdup)) :key #'cdr))
                               ;; (unless (c-true-stalep holdup)
                               ;;     (trc nil "dataflow sees freed blocker" holdup))
                               (if (c-true-stalep holdup)
                                   (if (eq :eternal-rest (md-state (c-model holdup)))
                                       (progn (trc "holdup is no more!!!!!!!" holdup (c-true-stalep heldup) heldup))
                                     (progn
                                       (trc holdup "dataflow retrying blocker" holdup)
                                       (c-rethink holdup)))
                                 (progn
                                   (trc heldup "holdup not stale!!!" holdup :heldup> heldup)
                                   (c-pending-set heldup nil :holdup-unstale) ))
                               ;;(unless (c-true-stalep heldup)
                               ;;     (trc nil "dataflow sees freed blocked" heldup))
                               (when (c-true-stalep heldup)
                                 (trc heldup "dataflow retrying blocked" heldup)
                                 (c-rethink heldup))))))
                  ;; (trace c-rethink cd-deep-stale sw-detect-interference)
                  (do-last *sw-pending*)
                  ;; (trc "post sweep pending leftovers:" (length *sw-pending*))
                  ;; (untrace c-rethink cd-deep-stale sw-detect-interference)
                  )
                ;;(untrace c-rethink)
                (when (and (equal oldpending *sw-pending*)
                        (eql pct (length *sw-pending*))
                        (not *sw-looping*))
                  (setf *sw-looping* t)
                  #+nah (dolist (p *sw-pending*)
                          (destructuring-bind (heldup . holdup) p
                            (dump-dependency-path holdup heldup)))
                  #+nah (dolist (p *sw-pending*)
                          (destructuring-bind (heldup . holdup) p
                            (declare (ignorable heldup))
                            (when t ;; (trcp holdup)
                              (dump-stale-path holdup))))
                  (break "trigger ~a stuck; cant lose pendings ~a"
                    c-originating
                    *sw-pending*))
                
                ;; (trc "after sweep sw-pending" *sw-pending*)
                ;; (cellbrk)
                ))
        (when c-originating
          (setf (cv-setting-p c-originating) nil))
        (trc nil "<<<< with-dataflow-management:" c-originating)))))

(defun dump-stale-path (used)
  (c-assert used)
  (when (typep used 'c-dependent)
    (loop with any
        for used-used in (cd-useds used)
        when (dump-stale-path used-used)
        do (progn
             (setf any t)
             (trc "stale-path" used :uses... used-used))
        finally
          (when (or any (cd-stale-p used))
            (trc "stale" used)
            (return any)))))

(defun dump-dependency-path (used user)
  (c-assert (and used user))
  (if (eql used user)
      (progn
        (trc "bingo---------------")
        (trc "user" user :uses...)
        t)
    (let (any)
      (dolist (used-user (cd-users used) any)
        (when (dump-dependency-path used-user user)
          (setf any t)
          (trc "user" used-user :uses... used))))))

(defun c-pending-set (c newvalue debug-tag)
  (declare (ignorable debug-tag))
  (c-assert (find c (cells (c-model c)) :key #'cdr))
  (when newvalue (trc nil "still pending!!!!!!!!!!!!!!!!!!" c newvalue debug-tag))
  (if newvalue
        (bif (known (assoc c *sw-pending*))
            (cond
             ((eq newvalue (cdr known))
              (break "hunh? re-pending ~a on same holdup ~a?" c (cdr known)))
             ((c-true-stalep (cdr known))
              (break "hunh? pending ~a on second holdup ~a as well as ~a?" c newvalue (cdr known)))
             (t
              (trc nil "re-pending ~a on new holdup ~a, last ok: ~a" c newvalue (assoc c *sw-pending*))
              (rplacd known newvalue))) ;; risky business, might need whole new assoc entry
          (let ((newpending (cons c newvalue)))
            (progn
              (c-assert (typep c 'c-dependent))
              (c-assert (not (eq :eternal-rest (md-state (c-model c)))))
              ;;(trc nil "pending on, genealogy holdup: held, holder:" debug-tag c newvalue)
              ;;(dump-pending)
              )
            ;;; hunh?> (pushnew newpending *sw-pending* :test #'equal)
            (push newpending *sw-pending*)))
    (bwhen (p (assoc c *sw-pending*))
      (trc nil "clear from sw-pending" debug-tag c (remove-if (lambda (p)
                                                                (not (eql c (car p))))
                                                     *sw-pending*))
      (setf *sw-pending* (delete (assoc c *sw-pending*) *sw-pending*))
      (progn
        (trc nil "pending off, genealogy holdup: held, holder:" debug-tag p
          (count c *sw-pending* :key #'car))
        (dump-pending))
      ))
  newvalue)

(defmethod sw-pending ((c cell))
  (assoc c *sw-pending*))

(defmethod sw-pending ((s synapse))
  (sw-pending (syn-used s)))


