;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cells; -*-
;;;
;;; Copyright (c) 1995,2003 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cells)

(defun cellbrk (&optional (tag :anon))
  (unless (or (c-stopped))
    ;; daring move, hoping having handler at outside stops the game (cellstop)
    (print `(cell break , tag))
    (break)))

;----------- trc -------------------------------------------

(defun trcdepth-reset ()
  (setf *trcdepth* 0))

(defmacro trc (tgtform &rest os)
  (if (eql tgtform 'nil)
      '(progn)
    (if (stringp tgtform)
        `(without-c-dependency
          (call-trc t ,tgtform ,@os))
      (let ((tgt (gensym)))
        `(without-c-dependency
          (bif (,tgt ,tgtform)
            (if (trcp ,tgt)
                (progn
                  (c-assert (stringp ,(car os)))
                  (call-trc t ,@os)) ;;,(car os) ,tgt ,@(cdr os)))
              (progn
                (c-break :trcfailed)
                (count-it :trcfailed)))
            (count-it :tgtnileval)))))))

(defun call-trc (stream s &rest os)
  (if #+cormanlisp nil #-cormanlisp (and (boundp '*trcdepth*)
          *trcdepth*)
        (format stream "~&~v,,,'.<~d~>> " (mod *trcdepth* 100) *trcdepth*)
      (format stream "~&"))
    
    (format stream "~a" s)
    (let (pkwp)
      (dolist (o os)
        (format stream (if pkwp " ~s" " | ~s") o)
        (setf pkwp (keywordp o))))
    (values))

(defun call-trc-to-string (fmt$ &rest fmtargs)
  (let ((o$ (make-array '(0) :element-type 'base-char
              :fill-pointer 0 :adjustable t)))
    (with-output-to-string (ostream o$)
      (apply 'call-trc ostream fmt$ fmtargs))
    o$))

#+findtrcevalnils
(defmethod trcp :around (other)
  (unless (call-next-method other)(break)))

(defmethod trcp (other)
  (eq other t))
  
(defmethod trcp (($ string))
  t)
  
(defun trcdepth-incf ()
  (incf *trcdepth*))
  
(defun trcdepth-decf ()
  (format t "decrementing trc depth" *trcdepth*)
  (decf *trcdepth*))
  
(defmacro wtrc ((&optional (min 1) (max 50) &rest banner) &body body )
  `(let ((*trcdepth* (if *trcdepth*
                         (1+ *trcdepth*)
                       0)))
     ,(when banner `(when (>= *trcdepth* ,min)
                      (if (< *trcdepth* ,max)
                          (trc ,@banner)
                        (progn
                          (break "excess trace notttt!!! ~d" *trcdepth*) ;; ,@banner)
                          nil))))
     (when (< *trcdepth* ,max)
       ,@body)))

(defmacro wnotrc ((&optional (min 1) (max 50) &rest banner) &body body )
  (declare (ignore min max banner))
  `(progn ,@body))
  
;------ eko --------------------------------------


(defmacro eko ((&rest trcargs) &rest body)
  (let ((result (gensym)))
     `(let ((,result ,@body))
         (trc ,(car trcargs) :=> ,result ,@(cdr trcargs))
         ,result)))

(defmacro ek (label &rest body)
  (let ((result (gensym)))
     `(let ((,result (,@body)))
         (when ,label
           (trc ,label ,result))
         ,result)))

;------------- counting ---------------------------
(defvar *count* nil)
(defvar *counting* nil)

(defmacro with-counts ((onp &rest msg) &body body)
  `(if ,onp
       (Let ((*counting* (cons t *counting*)))
         (prog2
           (count-clear ,@msg)
             (progn ,@body)
           (show-count t ,@msg)))
     (progn ,@body)))
  
(defun count-clear (&rest msg)
  (declare (ignorable msg))
  (format t "~&count-clear > ~a" msg)
  (setf *count* nil))

(defmacro count-it (&rest keys)
  `(when *counting*
     (call-count-it ,@keys)))

(defun call-count-it (&rest keys)
    (declare (ignorable keys))
  ;;; (when (eql :TGTNILEVAL (car keys))(break))
  (let ((entry (assoc keys *count* :test #'equal)))
      (if entry
          (setf (cdr entry) (1+ (cdr entry)))
        (push (cons keys 1) *count*))))

(defun show-count (clearp &rest msg)
  (format t "~&Counts after: clearp ~a, length ~d: ~s" clearp (length *count*) msg)
  (let ((res (sort (copy-list *count*) (lambda (v1 v2)
                                           (let ((v1$ (symbol-name (caar v1)))
                                                 (v2$ (symbol-name (caar v2))))
                                             (if (string= v1$ v2$)
                                                 (< (cdr v1) (cdr v2))
                                               (string< v1$ v2$))))))
        )
     (loop for entry in res
         for occs = (cdr entry)
         when (plusp occs)
           sum occs into running
           and do (format t "~&~4d ... ~2d ... ~s" running occs (car entry))))
  (when clearp (count-clear "show-count")))
  

;-------------------- timex ---------------------------------

(eval-when (compile eval load)
  (export '(timex)))

(defmacro timex ((onp &rest trcArgs) &body body)
  `(if ,onp
       (prog1
           (time (PROGN ,@body))
         (trc "timing was of" ,@trcARgs))
     (progn ,@body)))

(defparameter *billing* nil)
(defparameter *bill-to* nil)


(defmacro with-billing-outer ((key) &body body)
  (declare (ignorable key))
    #-allegro
    `(progn ,@body)
    #+allegro
  (let ((result (gensym))
        (e (gensym)))
    `(let ((cpu-gc-user 0) (cpu-gc-sys 0) (cpu-tot-user 0) (cpu-tot-sys 0)
             (real-time 0) (conses 0) (other-bytes 0) (static-bytes 0))
       (if (or (null *bill-to*)
             (eql ,key *bill-to*))
           (progn
             (trc nil "bill-to indicates just run" *bill-to*)
             ,@body)
         (let (
               (,result (EXCL::TIME-A-FUNCALL
                         (lambda (c-gc-user c-gc-sys c-tot-user c-tot-sys r-time 
                                   cons other-b static-b)
                           (incf cpu-gc-user c-gc-user)
                           (incf cpu-gc-sys c-gc-sys)
                           (incf cpu-tot-user c-tot-user)
                           (incf cpu-tot-sys c-tot-sys)
                           (trc nil "outer ups r/t from" real-time :by r-time)
                           (incf real-time r-time)
                           (incf conses cons)
                           (incf other-bytes other-b)
                           (incf static-bytes static-b))
                         (LAMBDA () 
                           (let ((*bill-to* ,key))
                             ,@body)))))
           (trc nil "outer incr bill r/t" ,key *bill-to* real-time)
           (bIf (,e (cdr (assoc ,key *billing*)))
             (progn
               (incf (car ,e))
               (rplacd ,e (mapcar '+ (cdr ,e)
                            (list cpu-gc-user cpu-gc-sys cpu-tot-user cpu-tot-sys 
                              real-time conses other-bytes static-bytes))))
             (push (cons ,key (cons 1 (list cpu-gc-user cpu-gc-sys cpu-tot-user cpu-tot-sys 
                                        real-time conses other-bytes static-bytes)))
               *billing*)
             )
           ,result)))))

(defmacro with-billing ((&rest trcArgs) &body body)
  (declare (ignorable trcArgs))
    #-allegro
    `(progn ,@body)
    #+allegro
  (if (car trcargs)
      `(let (*billing* (*bill-to* :init))
         (prog1
             (with-billing-outer (:top)
               (progn ,@body))
           ;;(sort *billing* '> :key 'cddr)
           (loop for b in *billing*
                 do (destructuring-bind (key . (ct . (c-gc-user c-gc-sys c-tot-user 
                                                       c-tot-sys r-time 
                                                       cons other-b static-b)))
                        b
                      (declare (ignorable c-gc-user c-gc-sys c-tot-user 
                                 c-tot-sys r-time 
                                 cons other-b static-b))
                      (when (> r-time 1)
                        (trc "bill:" key :count ct :cpu c-tot-user 
                          :real (/ r-time 1000.0) :cons cons))))
           (trc ,@trcargs)))
      `(progn ,@body)))

(defmacro with-billing-adj (&body body)
    #-allegro
    `(progn ,@body)
    #+allegro
  `(EXCL::TIME-A-FUNCALL
    (lambda (c-gc-user c-gc-sys c-tot-user c-tot-sys r-time 
              cons other-b static-b)
      (decf cpu-gc-user c-gc-user)
      (decf cpu-gc-sys c-gc-sys)
      (decf cpu-tot-user c-tot-user)
      (decf cpu-tot-sys c-tot-sys)
      (decf real-time r-time)
      (decf conses cons)
      (decf other-bytes other-b)
      (decf static-bytes static-b))
    (LAMBDA () 
      (let ((*bill-to* :subtracted))
        ,@body))))


#+allegro
(eval-when (:compile-toplevel :load-toplevel :execute)
  (export '(with-billing with-billing-adj with-billing-outer)))

#+test
(time
 (with-billing ("test billing")
   (with-billing-adj (a 2))))

#+allegro
(defun a (recur)
  (print `(enter-a ,recur))
  (when (plusp recur) 
      (c (1- recur)))
  (with-billing-outer (:a)
    (sleep 1)
    (assert (eql 17
              (with-billing-adj
                  (if (plusp recur) 
                    (b (1- recur))
                    17))))
    42))

#+allegro
(defun b (recur)
  (declare (ignorable recur))
  (with-billing-outer (:b)
    (print `(enter-b ,recur))
    (with-billing-adj (a recur))
    (sleep 0.5)
    17))

#+allegro
(defun c (recur)
  (declare (ignorable recur))
  (print `(enter-c ,recur))
  (a recur))

#+allegro
(defun cells-time-report (cpu-gc-user cpu-gc-sys cpu-tot-user cpu-tot-sys real-time conses other-bytes static-bytes)
  (trc "cpu-gc-user" cpu-gc-user)
  (trc "cpu-gc-sys" cpu-gc-sys)
  (trc "cpu-tot-user" cpu-tot-user)
  (trc "cpu-tot-sys" cpu-tot-sys)
  (trc "<non-gc user cpu>" (- cpu-tot-user cpu-gc-user))
  (trc "<non-gc sys cpu>" (- cpu-tot-sys cpu-gc-sys))
  (trc "conses" conses)
  (trc "other-bytes" other-bytes)
  (trc "static-bytes" static-bytes)
  (excl::time-report cpu-gc-user cpu-gc-sys cpu-tot-user cpu-tot-sys real-time conses other-bytes static-bytes))

;---------------- Metrics -------------------

(defmacro with-metrics ((countp timep &rest trcargs) &body body)
  `(with-counts (,countp ,@trcargs)
    (timex (,timep ,@trcargs)
       ,@body)))


; -------- cell conditions (not much used) ---------------------------------------------

(define-condition xcell () ;; new 2k0227
  ((cell :initarg :cell :reader cell :initform nil)
   (appfunc :initarg :appfunc :reader appfunc :initform 'badcell)
   (errortext :initarg :errortext :reader errortext :initform "<???>")
   (otherdata :initarg :otherdata :reader otherdata :initform "<nootherdata>"))
  (:report (lambda (c s)
             (format s "~& trouble with cell ~a in function ~s,~s: ~s"
               (cell c) (appfunc c) (errortext c) (otherdata c)))))

(define-condition c-enabling ()
   ((name :initarg :name :reader name)
    (model :initarg :model :reader model)
    (cell :initarg :cell :reader cell))
   (:report (lambda (condition stream)
                 (format stream "~&unhandled <c-enabling>: ~s" condition)
                 (break "~&i say, unhandled <c-enabling>: ~s" condition))))

(define-condition c-fatal (xcell)
   ((name :initarg :name :reader name)
    (model :initarg :model :reader model)
    (cell :initarg :cell :reader cell))
   (:report (lambda (condition stream)
              (format stream "~&fatal cell programming error: ~s" condition)
              (format stream "~&  : ~s" (name condition))
              (format stream "~&  : ~s" (model condition))
              (format stream "~&  : ~s" (cell condition)))))

(define-condition c-unadopted (c-fatal)
   ()
   (:report
    (lambda (condition stream)
      (format stream "~&unadopted cell >: ~s" (cell condition))
      (format stream "~& >: often you mis-edit (c? (c? ...)) nesting is error"))))


;----------------------------- link debugging -----------------------


(defun dump-users (c &optional (depth 0))
     (format t "~&~v,4t~s" depth c)
     (dolist (user (c-users c))
          (dump-users user (+ 1 depth))))

(defun dump-useds (c &optional (depth 0))
     ;(c.trc "dump-useds> entry " c (+ 1 depth))
     (when (zerop depth)
          (format t "x~&"))
     (format t "~&|usd> ~v,8t~s" depth c)
     (when (typep c 'c-ruled)
          ;(c.trc "its ruled" c)
          (dolist (used (cd-useds c))
               (dump-useds used (+ 1 depth)))))


(defun cell-reset ()
  (setf *count* nil
    *stop* nil
    *dbg* nil
    *c-prop-depth* 0
    *sw-looping* nil
    *to-be-awakened* nil
    *trcdepth* 0))