;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cells; -*-
;;;
;;; Copyright (c) 1995,2003 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cells)

(defun md-slot-cell-flushed (self slot-spec)
  (cdr (assoc (slot-spec-name slot-spec) (cells-flushed self))))

(defun md-slot-value (self slot-spec &aux (slot-c (md-slot-cell self slot-spec)))
  ;;(count-it :md-slot-value #+nah (type-of self) slot-spec)
  (progn ;; wtrc (0 200 slot-c "md-slot-value" self slot-spec)
    (when *stop*
      (princ #\.)
      (return-from md-slot-value))
    ;; (count-it :md-slot-value (slot-spec-name slot-spec))
    #+badidea
    (when (mdead self)
      (trc "md-slot-value> model dead" (type-of self) slot-spec)
      (return-from md-slot-value nil))

    ;; next bit recurses insanely, and this work is not here for a good
    ;; reason, more as a "why not?". so find a more targeted fix if it seems
    ;; this following overkill is what is needed
    #+nil
    (when (eql :nascent (md-state self))
      (let (*c-calculators*) 
        (md-awaken self)))
    
    ; this next bit could be coded: (c-relay-value <link> (etypecase slot-c...))
    ; but that would link before accessing possibly an invalid ruled slot
    ; (during md-awaken), and after calculating it would propagate to users and
    ; re-enter this calculation. Switching the order of the parameters would
    ; also work, but we need to document this very specific order of operations
    ; anyway, can't just leave that to the left-right thing.
    ;
    (let ((slot-value (etypecase slot-c
                        (null (bd-slot-value self slot-spec))
                        (c-variable (c-value slot-c))
                        (c-ruled (c-ruled-slot-value slot-c)))))
      (c-relay-value
       (when (car *c-calculators*)
         (c-link-ex slot-c))
       slot-value))))

(defun c-ruled-slot-value (slot-c)
  (progn ;; wtrc (0 200 nil "c-ruled-slot-value"  slot-c)
    
    (c-assert (not (cmdead slot-c)))
    
    (cond
     ((c-validp slot-c)
      (if (c-true-stalep slot-c)
          (c-calculate-and-set slot-c)
        (bif (deep (cd-deep-stale slot-c))
          (progn
            (trc nil "valid ~a :but-deepstale ~a, cause: ~a. calcing"
              slot-c deep *cause*)
            (c-calculate-and-set slot-c))
          #+worked (progn
                     (trc "valid ~a :but-deepstale ~a, cause: ~a. calcing"
                       slot-c deep *cause*)
                     (c-calculate-and-set deep)
                     (bIf (deep2 (cd-deep-stale slot-c))
                       (break "deep, deep trouble ~a :deep2 ~a :deep1 ~a, cause: ~a."
                         slot-c deep2 deep *cause*)
                       (progn
                         (trc "cleared valid with deep stale" slot-c)
                         (c-calculate-and-set slot-c))))
          (c-value slot-c)))) ;; good to go
     
     (t (let ((*cause* :on-demand)) ; normal path first time asked
          (progn ;; wtrc (0 200 slot-c "c-ruled-slot-value :on-demand"  slot-c)
            (c-calculate-and-set slot-c)))))))

;-------------------------------------------------------------

(defun (setf md-slot-value) (newvalue self slot-spec)
  (when (mdead self)
    (return-from md-slot-value))
  (let ((c (md-slot-cell self slot-spec)))

    (when *c-debug*
      (c-setting-debug self slot-spec c newvalue))

    (unless c
      (c-stop :setf-md-slot-value)
      (c-break "(setf md-slot-value)> cellular slot ~a of ~a cannot be setf unless initialized to c-variable"
        slot-spec self)
      )

    (if (cv-setting-p c)
      (if (cv-cyclic-p c)
          newvalue
          (c-break "setf of ~a propagated back; declare as cyclic (cv8...)" c))
      (let ((absorbedvalue (c-absorb-value c newvalue)))
        ;;(c-assert (not (mdead self)))
        ;;(count-it :md-slot-value-setf slot-spec)
        (with-dataflow-management (c)
          (md-slot-value-assume self slot-spec absorbedvalue)) ;; /// uh-oh. calc-n-set uses this return value
        absorbedvalue))))

;;;(defmethod trcp ((c c-ruled))
;;;  ;;(trc "trcp ruled" (c-slot-name c) (md-name (c-model c)))
;;;  (and (not (eql '.kids (c-slot-name c)))
;;;    (typep (c-model c) 'cello::slot-inspector)
;;;    #+not (eql :mklabel (md-name (c-model c)))))


(defun md-slot-value-assume (self slot-spec absorbedvalue
                              &aux
                              (c (md-slot-cell self slot-spec))
                              (priorstate (when c (c-state c)))
                              (priorvalue (when c (c-value c)))
                              )
  (without-c-dependency
    (when (mdead self)
      (return-from md-slot-value-assume nil))
    (md-slot-value-store self (slot-spec-name slot-spec)
      (if c
          (setf (c-value c) absorbedvalue)
        absorbedvalue))
    
    (when (typep c 'c-ruled)
      (trc nil " setting cellstate :valid" c)
      (setf (c-state c) :valid)
      (setf (cd-stale-p c) nil)
      (setf (c-waking-state c) :awake)
      (c-pending-set c nil :sv-assume)
      (c-optimize-away?! c)) ;;; put optimize as early as possible
    
    ;--- propagation -----------
    ;
    (if (and (eql priorstate :valid) ;; ie, priorvalue meaningful (nil is ambiguous)
          (c-no-news c absorbedvalue priorvalue))
        (progn
          (trc nil "(setf md-slot-value) >no-news" priorstate (c-no-news c absorbedvalue priorvalue))
          #+not (count-it :no-news))
      (progn
        (when (eql '.kids (slot-spec-name slot-spec))
          #+dfdbg (dolist (K absorbedvalue) (trc k "md-slot-value-assume -> kids change" k self))
          (md-kids-change self absorbedvalue priorvalue :md-slot-value-assume))
        (md-propagate self slot-spec absorbedvalue priorvalue (not (eql :unbound priorstate)))))
    absorbedvalue))

