;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cells; -*-
;;;
;;; Copyright (c) 1995,2003 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cells)

(defparameter *echodone* nil)

(defun c-echo-defined (slot-name)
  (getf (symbol-plist slot-name) :echo-defined))

(defmethod (setf c-true-stalep) (newvalue (user c-ruled))
  #+dfdbg (trc user "setting c-true-stalep" user newvalue)
  (c-assert (find user (cells (c-model user)) :key #'cdr))
  (setf (cd-stale-p user) newvalue))

(defmethod (setf c-true-stalep) (newvalue (usersyn synapse))
  #+dfdbg (trc (syn-user usersyn) "synapse setting c-true-stalep" (syn-user usersyn) newvalue usersyn)
  (setf (cd-stale-p (syn-user usersyn)) newvalue))

(defmethod (setf c-true-stalep) (newvalue other)
  (declare (ignore other))
  newvalue)

(defun c-echo-initially (self slot-spec)
  "call during instance init. if echo is defined for slot, and value is non-nil (controversial) force initial echo."
  (trc nil "c-echo-initially" self slot-spec
    (c-echo-defined (slot-spec-name slot-spec))
    (md-slot-cell self slot-spec))
  (if (c-echo-defined (slot-spec-name slot-spec))
    (bif (c (md-slot-cell self slot-spec))
         (etypecase c
           (c-variable (md-propagate self slot-spec (c-value c) nil nil))
           (c-ruled (md-slot-value self slot-spec))) ;; this will echo after calculating if not nil
         ;
         ; new for 22-03-07: echo even if slot value is nil...
         (c-echo-slot-name (slot-spec-name slot-spec)
                           self
                           (bd-slot-value self slot-spec)
                           nil nil))
    (bwhen (c (md-slot-cell self slot-spec))
      (c-ephemeral-reset c))))

#-(or cormanlisp clisp)
(defgeneric c-echo-slot-name (slotname self new old old-boundp) (:method-combination progn))

(defun c-echo-slot (slot-spec self newvalue priorvalue priorvalue-supplied)
  (with-billing-outer ((slot-spec-name slot-spec))
    (c-echo-slot-name (slot-spec-name slot-spec)
      self
      newvalue
      priorvalue
      priorvalue-supplied)))

(defmethod c-echo-slot-name
    #-(or cormanlisp clisp) progn
  #+(or cormanlisp clisp) :before
  (slot-name self new old old-boundp)
  (declare (ignorable slot-name self new old old-boundp)))

#+(or cormanlisp clisp)
(defmethod c-echo-slot-name (slot-name self new old old-boundp)
  (declare (ignorable slot-name self new old old-boundp)))

;--------------- propagate  ----------------------------
;
; n.b. 990414kt the cell argument may have been optimized away,
; though it is still receiving final processing here.
;


(defun md-propagate (self slot-spec newvalue priorvalue priorvalue-supplied)
  (with-billing-outer (:md-prop)
    (when (mdead self)
      (trc nil "md-propagate n-opping dead" self)
      (return-from md-propagate nil))
    
    (count-it :md-propagate)
    
    (let (*c-calculators* 
          (*c-prop-depth*  (1+ *c-prop-depth*))
          (c (md-slot-cell self slot-spec)))
      ;
      ;------ debug stuff ---------
      ;
      (when *stop*
        (princ #\.)(princ #\!)
        (return-from md-propagate))
      
      (when c
        (trc nil "md-propagate> propping" self slot-spec (length (c-users c)) c)
        )
      
      (when *c-debug*
        (when (> *c-prop-depth* 250)
          (trc "md-propagate deep" *c-prop-depth* self (slot-spec-name slot-spec) #+nah c))
        (when (> *c-prop-depth* 300)
          (c-break "md-propagate looping ~c" c)
          ))
      
      (when c
        ; ------ flag dependents as stale ------------
        ; do before echo in case echo gets back to some user
        ;
        (dolist (user (c-users c))
          #+dfdbg (trc user  "md-prop now setting stale (changer, stale):" c user)
          (when (c-user-cares user)
            (setf (c-true-stalep user) c))))
      
      ; --- manifest new value as needed -----------
      (when (c-echo-defined (slot-spec-name slot-spec)) ;; /// faster than just dispatching?
        (when c (trc nil "md-prop now echoing" c))
        (count-it :echo #+nah (slot-spec-name slot-spec))
        (with-billing-adj
            (c-echo-slot slot-spec self newvalue priorvalue priorvalue-supplied)
          )
        
        (when (mdead self) ;; hopefully expiration on perishable class
          (return-from md-propagate)))
      
      (with-billing-adj
          (when c ; --- now propagate to dependents ------------
            (trc nil "md-prop checking dependents" c (c-users c))
            (let ((*cause* c))
              (dolist (user (c-users c))
                (unless (cmdead user)
                  (when (c-user-cares user)
                    (if (c-user-lazy user)
                        (progn
                          (trc nil "lazy user not being propagated to" user :by c)
                          (dolist (u (c-users user))
                            (c-propagate-staleness u)))
                      (progn
                        (c-rethink user)
                        (when (mdead self)
                          (trc nil "md-propagate> self now dead after rethink user: ~a" self user)
                          (return-from md-propagate nil))
                        )))))
              (c-ephemeral-reset c)))))))

(defmethod c-propagate-staleness ((c c-ruled))
  (trc nil "inheriting staleness" c)
  (dolist (u (cr-users c))
    (c-propagate-staleness u)))

(defmethod c-propagate-staleness ((s synapse))
  (trc "I hope this synapse isn't for efficiency" s)
  (break)
  (c-propagate-staleness (syn-user s)))

(defmethod c-propagate-staleness (c)
  (declare (ignorable c))
  (trc "not inheriting or proagating staleness" c)
  )

(defmethod c-user-cares (c) c) ;; ie, t
(defmethod c-user-cares ((s synapse))
  (syn-relevant s))

(defmethod c-user-lazy (c) (declare (ignore c)) nil)
(defmethod c-user-lazy ((c c-ruled))
  (cr-lazy c))


(defun c-ephemeral-reset (c)
    (when c
      (when (c-ephemeral-p c)
        (trc nil "c-ephemeral-reset resetting:" c)
        (setf (c-value c) nil)))) ;; good q: what does (setf <ephem> 'x) return? historically nil, but...?

;----------------- change detection ---------------------------------

(defun c-no-news (c newvalue oldvalue)
  ;;; (trc nil "c-no-news > checking news between" newvalue oldvalue)
          
  (bIf (test (c-unchanged-test (c-model c) (c-slot-name c)))
      (funcall test newvalue oldvalue)
      (eql newvalue oldvalue)))

(defmacro def-c-unchanged-test ((class slotname) &body test)
  `(defmethod c-unchanged-test ((self ,class) (slotname (eql ',slotname)))
     ,@test))
     
(defmethod c-unchanged-test (self slotname)
  (declare (ignore self slotname))
  nil)

(defmethod c-identity-p ((value null)) t)
(defmethod c-identity-p ((value number)) (zerop value))
(defmethod c-identity-p ((value cons))
  ;; this def a little suspect?
  (and (c-identity-p (car value))
       (c-identity-p (cdr value))))


;------------------- re think ---------------------------------

(defmethod cmdead ((c synapse))
  (cmdead (syn-user c)))

(defmethod cmdead (c)
;  (count-it :cmdead)
  (if (or (null (c-model c)) (consp (c-model c)))
      t ;; 240128kt: was bizarre: (not (c-optimized-away-p c))
    (mdead (c-model c))))

(defun mdead (m) (eq :eternal-rest (md-state m)))

(defun c-rethink (c)
  (with-billing-outer (:rethink)
    (when *stop*
      (princ #\.)
      (return-from c-rethink))
    ;;(trc "rethink entry: c, true-stale" c (c-true-stalep c))
    (c-assert (not (cmdead c))() "rethink entry cmdead ~a" c)
    (unless (c-true-stalep c)
      (return-from c-rethink))
    
    (when *rethink-deferred*
      (trc nil "bingo!!!!!! rethink deferring" c *cause*)
      (push (list c *cause*) *rethink-deferred*)
      (return-from c-rethink))
    
    (c-assert (not (cmdead c))() "rethink here?? cmdead ~a" c)
    
    ; looking ahead for interference avoids JIT detection, which where a dependency
    ; already exists causes re-entrance into the rule, which should calculate the same
    ; value twice and echo only once, but still seems like something to avoid since
    ; we do already have the technology.
    ;
    
    (bIf (interf (sw-detect-interference c nil))
      (progn
        (trc "!!!!!!!!!! rethink of " c :held-up-by interf)
        (c-pending-set c interf :interfered)
        #+dfdbg (when (trcp c)
                  (trc "!!!!!!!!!! rethink of " c :held-up-by interf)
                  #+nah (dump-stale-path interf)
                  )
        (return-from c-rethink))
      (when (sw-pending c)
        (trc nil "no interference now for " c)
        (c-pending-set c nil :dis-interfered)))
    
    (when (cmdead c)
      (trc nil "woohoo!!! interference checking finished model off" c)
      (return-from c-rethink))
    
    (unless (c-true-stalep c)
      (trc nil "woohoo!!! interference checking refreshed" c)
      (return-from c-rethink))
    
    (with-billing-adj
        (typecase c
          (c-ruled (c-calculate-and-set c))
      
          (synapse
           (trc nil "c-rethink > testing rethink of: syn,salv,valu" c salvage (c-value (syn-used c)))
           (if (funcall (syn-fire-p c) c (c-value (syn-used c)))
               (progn
                 (trc nil "c-rethink> decide yes on rethink on syn, valu" c (c-value (syn-used c)))
                 (c-rethink (syn-user c)))
             (trc nil "c-rethink> decide nooooo on rethink on synapse" c (syn-user c) salvage)))
          ))))

(defmacro def-c-echo (slotname
                      (&optional (selfarg 'self) (newvarg 'new-value)
                                 (oldvarg 'old-value) (oldvargboundp 'old-value-boundp))
                      &body echobody)
  ;;;(trc "echo body" echobody)
  `(progn
     (eval-when (:compile-toplevel :load-toplevel :execute)
       (setf (get ',slotname :echo-defined) t))
     ,(if (eql (last1 echobody) :test)
          (let ((temp1 (gensym))
                (loc-self (gensym)))
            `(defmethod c-echo-slot-name #-(or clisp cormanlisp) progn ((slotname (eql ',slotname)) ,selfarg ,newvarg ,oldvarg ,oldvargboundp)
               (let ((,temp1 (bump-echo-count ,slotname))
                     (,loc-self ,(if (listp selfarg)
                                     (car selfarg)
                                   selfarg)))
                 (when (and ,oldvargboundp ,oldvarg)
                   (format t "~&echo ~d (~a ~a) old: ~a" ,temp1 ',slotname ,loc-self ,oldvarg))
                 (format t "~&echo ~d (~a ~a) new: ~a" ,temp1 ',slotname ,loc-self ,newvarg))))
        `(defmethod c-echo-slot-name
             #-(or clisp cormanlisp) progn
           ((slotname (eql ',slotname)) ,selfarg ,newvarg ,oldvarg ,oldvargboundp)
           (declare (ignorable ,(etypecase selfarg
                                  (list (car selfarg))
                                  (atom selfarg))
                               ,(etypecase newvarg
                                  (list (car newvarg))
                                  (atom newvarg))
                               ,(etypecase oldvarg
                                  (list (car oldvarg))
                                  (atom oldvarg))
                               ,(etypecase oldvargboundp
                                  (list (car oldvargboundp))
                                  (atom oldvargboundp))))
           ,@echobody))))

(defmacro bump-echo-count (slotname) ;; pure test func
  `(if (get ',slotname :echos)
       (incf (get ',slotname :echos))
     (setf (get ',slotname :echos) 1)))

