;; -*- mode: Lisp; Syntax: Common-Lisp; Package: ffx; -*-
;;;
;;; Copyright  1995,2003 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(defpackage :ffi-extender
  (:nicknames :ffx)
  (:export
   #:ff-elt
   #:eltf #:eltd
   #:with-ff-array-elements
   #:make-ff-array
   ))

(in-package :ffx)

(defmacro make-ff-array (type &rest values)
  (let ((fv (gensym))(n (gensym))(vs (gensym)))
    `(let ((,fv (uffi:allocate-foreign-object ',type ,(length values)))
           (,vs (list ,@values)))
       (dotimes (,n ,(length values) ,fv)
         (setf (ff-elt ,fv ,type ,n)
           (coerce (nth ,n ,vs) ',(if (keywordp type)
                                     (intern (symbol-name type))
                                   (get type 'fficast))))))))

#+test
(uffi:allocate-foreign-object ':float 3)

#+test
(make-ff-array :float 1 2 3)

#+test
(ct::defctype floatarray (:single-float 3))
;--------- with-ff-array-elements ------------------------------------------


(defmacro with-ff-array-elements ((fa type &rest refs) &body body)
  `(let ,(let ((refn -1))
           (mapcar (lambda (ref)
                     `(,ref (uffi:deref-array ,fa '(:array ,type) ,(incf refn))))
             refs))
     ,@body))

#+test
(with-ff-array-elements ((make-ff-array :float 1 2 3) :float x1 x2 x3)
  (list x1 x2 x3))

;-------- ff-elt ---------------------------------------

#+test
(ct:defctype f1 (:single-float 1))

#+test
(let ((f3 (uffi:allocate-foreign-object :float 3)))
    (dotimes (n 3 (ct:cref (:single-float 0) f3 2))
        (print (uffi:deref-array f3 '(:array :float) 0))
        (setf (ct:cref (:single-float 0) f3 n) (* 1.0f0 n))))

#+test
(macroexpand '(uffi:deref-array f3 '(:array :float) 0))

#+test
(let ((f3 (uffi:allocate-foreign-object :float 3)))
    (dotimes (n 3 (ct:cref f1 f3 2))
        (setf (ct:cref f1 f3 n) (* 1.0f0 n))))

(defmacro ff-elt (v type n)
  `(uffi:deref-array ,v '(:array ,type) ,n))

#+test
(ff-elt (make-ff-array :float 1 2 3) :float 2)

(defun eltf (v n)
  (ff-elt v :float n))

#+test
(eltf (make-ff-array :float 1 2 3) 2)

(defun (setf eltf) (value v n)
  (setf (ff-elt v :float n) (coerce value 'float)))

(defun eltd (v n)
  (ff-elt v :double n))

(defun (setf eltd) (value v n)
  (setf (ff-elt v :double n) (coerce value 'double-float)))

