;; -*- mode: Lisp; Syntax: Common-Lisp; Package: ffx; -*-
;;;
;;; Copyright  1995,2003 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.


(defpackage :ffi-extender
  (:nicknames :ffx)
  (:export
   #:ff-register-callable
   #:ff-defun-callable
   #:ff-def-call
   #:ff-pointer-address
   ))


(in-package :ffx)

(defun ff-register-callable (callback-name)
  #+allegro
  (ff:register-foreign-callable callback-name)
  #+lispworks
  (fli:pointer-address
    (fli:make-pointer :symbol-name (symbol-name callback-name) ;; leak?
      :functionp t)))

(defmacro ff-defun-callable (call-convention result-type name args &body body)
  (declare (ignorable result-type))
#+lispworks
  `(fli:define-foreign-callable
    (,(symbol-name name) :result-type ,result-type :calling-convention ,call-convention)
    (,@args)
    ,@body)
#+allegro
  `(ff:defun-foreign-callable ,name ,args
     (declare (:convention ,(ecase call-convention
                              (:cdecl :c)
                              (:stdcall :stdcall))))
     ,@body))

#+dave
(fli:define-foreign-callable 
    (wndproc :result-type :long :calling-convention :stdcall)
    ((hwnd hwnd) (msg (:unsigned :long)) 
     (wparam (:unsigned :long)) (lparam (:unsigned :long)))
  (cond 
   ((= msg wm-paint) (wndproc-paint hwnd msg wparam lparam))
   #+console ((= msg wm-destroy) (PostQuitMessage 0) 0)
   (t (DefWindowProc hwnd msg wparam lparam))))

#+dave
(defun register-class ()
  (fli:with-foreign-string (cn-p ec bc :external-format *external-format*)
      *class-name*
    ;; Below - use with-dynamic... for automatic freeing. Make some pointers.
    (let ((wc-p (fli:allocate-foreign-object :type 'wndclass))
          (wp-p (fli:make-pointer :symbol-name "wndproc")))
      ...
      (setf (fli:foreign-slot-value wc-p 'lpfnWndProc)
            (fli:pointer-address wp-p)))))
      
#+test ;; lw-sample
(fli:define-foreign-callable
 ("square" :result-type :int)
 ((arg-1 :int)) (* arg-1 arg-1))

#+test
(ff-defun-callable :cdecl :int square ((arg-1 :int))
   (* arg-1 arg-1))

(defmacro ff-def-call ((module iname ename) args)
  #+cormanlisp
  (assert module () "Module (dll name, in fact) required for Corman Lisp")
  #+cormanlisp
  `(ct:defun-dll ,iname ,args
     :return-type :short
     :library-name ,module ;; required according Corman doc
     :entry-name ,ename
     :linkage-type :c) ;; ??

  #+allegro (declare (ignorable module))
  #+allegro
  `(ff:def-foreign-call (,iname ,ename) ,args)
  #+lispworks
  `(fli:define-foreign-function (,iname ,ename)
     ,(mapcar (lambda (arg) (if (listp (Cadr arg))
                                (list (car arg) (substitute :pointer '* (cadr arg)))
                              arg))
        args)
     :module ,module
     :result-type :int))

