;; -*- mode: Lisp; Syntax: Common-Lisp; Package: ogb; -*-
;;;
;;; Copyright  1995,2003 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(defpackage #:opengl-bindings
  (:nicknames #:ogb)
  (:use #:common-lisp #:uffi #:ffx)
  (:export "GLUT-GET-WINDOW" "GLUT-SET-WINDOW" "GLUT-POST-REDISPLAY"
     "WITH-MATRIX" "WITHIN-GL-BEGIN-END" "GL-PUSHM" "GL-POPM"
    "GLUT-CALLBACK-SET" "OGB-GLUT-INIT" "CLOSED-STREAM-P" "*SELECTING*"
    ))

(in-package #:ogb)

(defmacro defun-ogb (rtn module$ name$ (&rest type-args))
    `(defun-ffx ,rtn ,module$ ,name$ (,@type-args)
       (glec ',rtn)))

(defmacro with-matrix ((&optional load-identity-p) &body body)
  `(progn
     (gl-push-matrix)
     (glec :with-matrix-push)
     (unwind-protect
         (progn
           (gl-get-integerv GL_MODELVIEW_STACK_DEPTH *stack-depth*)
           (glec :get-stack-depth)
           (print `(with-matrix model matrix stack ,(aforef *stack-depth* 0)))
     
           ,(when load-identity-p
              `(gl-load-identity))
           (prog1
               (progn ,@body)
             (glec :with-matrix)))
       (gl-pop-matrix)
       (glec))))

(defvar *gl-begun*)
(defvar *gl-stop*)
(defmacro within-gl-begin-end ((what) &body body)
  `(progn
     (when *gl-begun*
       (setf *gl-stop* t)
       (break ":nestedbegin"))
     (let ((*gl-begun* t))
       (gl-begin ,what)
       ,@body
       (gl-end)
       (glec :with-gl-begun))))

(defun glec (&optional (id :anon))
  (unless *gl-begun*
    (let ((e (glGetError)))
      (if (zerop e)
          (unless t ;; (find id '(glutcheckloop glutgetwindow))
            (print `(cool ,id)))
        (if t ;; (null (find id '(glutInitDisplayMode glutInitWindowSize)))
            (unless *gl-stop*
              (setf *gl-stop* t)
              (format t "~&~%OGL error ~a at ID ~a" e id)
              ;(break "OGL error ~a at ID ~a" e id)
              )
          #+sigh (print `("OGL error ~a at ID ~a" ,e ,id)))))))

(defvar *stack-depth*
  (uffi:allocate-foreign-object :int 1))

(defun aforef (o n)
  (uffi:deref-array o '(:array :int) n))


(dft GLuint #-allegro-v5.0.1 :unsigned-int #+allegro-v5.0.1 :int integer)
(dft GLenum #-allegro-v5.0.1 :unsigned-int #+allegro-v5.0.1 :int integer)
(dft GLbitfield #-allegro-v5.0.1 :unsigned-int #+allegro-v5.0.1 :int integer)

(dft GLint :int integer)
(dft GLsizei :int integer)

(dft GLuint #-allegro-v5.0.1 :unsigned-int #+allegro-v5.0.1 :int integer)
(dft GLushort #-allegro-v5.0.1 :unsigned-int #+allegro-v5.0.1 :int integer)

(dft GLfloat #+lispworks :lisp-single-float #-lispworks :float single-float)
(dft GLclampf #+lispworks :lisp-single-float #-lispworks :float single-float)

(dft GLdouble :double double-float)
(dft GLclampd :double double-float)

(dft GLboolean :unsigned-byte #+allegro character #-allegro number)
(dft GLbyte :byte  #+allegro character #-allegro number) ;; typedef signed char     GLbyte; 
(dft GLvoid :void integer)

(dft GLshort #-allegro-v5.0.1 :short #+allegro-v5.0.1 :int integer)
(dft GLubyte :unsigned-byte  #+allegro character #-allegro number)



