;; -*- mode: Lisp; Syntax: Common-Lisp; Package: ogb; -*-
;;;
;;; Copyright  1995,2003 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.


(in-package :ogb)

(defparameter *glut-dll* nil)
(defvar *glut-lib-path*)

#+allegro
(defun fg-unload ()
  (dolist (dll (ff:list-all-foreign-libraries))
    (when (search "freeglut" (pathname-name dll))
      (print `(Unloading foreign library ,dll))
      (setf *glut-dll* nil)
      (ff:unload-foreign-library dll))))

#-lispworks
(defun ogb-glut-init (rerun-validator)
  (declare (ignorable load-oglfont-p))
  (setf *gl-begun* nil)

  (when *glut-dll*
    #+lispworks (return-from ogb-glut-init)
    (format t "~&yes on *glut-dll* loaded: ~a" *glut-dll*)
    (format t "~&glut state: ~a" (glutget (coerce GLUT_INIT_STATE 'integer)))
    (when (not (zerop (glutget (coerce GLUT_INIT_STATE 'integer))))
      (when rerun-validator
        (funcall rerun-validator))
      (print "about to fgDeinitialize")
      (fgDeinitialize)
      (print "glut de-initialized")))

  (when (not *glut-dll*)
    ;;#-lispworks
    (print "loading open GL/GLU")
    (uffi:load-foreign-library
     ;; "c:/winnt/system32/opengl32.dll"
     "c:/windows/system32/opengl32.dll"
     :module "OPENGL32")
    ;; -lispworks#-lispworks
    (uffi:load-foreign-library
     ;; "c:/winnt/system32/glu32.dll"
     "c:/windows/system32/glu32.dll"
     :module "GLU32")
    (print (list "loading GLUT" *glut-lib-path* (probe-file *glut-lib-path*)))
    (assert (setq *glut-dll* (uffi:load-foreign-library *glut-lib-path*
                               #+stable "c:/windows/system32/freeglut.dll"
                               :force-load #+lispworks nil #-lispworks t
                               :module "glut"))
      () "Unable to load GLUT from: ~a" *glut-lib-path* ))
  
  (format t "~&glut state 2 ~a" (glutget (coerce GLUT_INIT_STATE 'integer)))
  (print "about to initialize")
  
  
  (let ((argc (uffi:allocate-foreign-object :int 1)))
    (setf (uffi:deref-array argc '(:array :int) 0) 0)
    (unwind-protect
        (progn
          (ogb::glut-Init argc (uffi:make-null-pointer '(:array :cstring)))
          (print "glut initialised")
          
          )
      (uffi::free-foreign-object argc)))
  )


#+lispworks
(defun ogb-glut-init (rerun-validator)
  (declare (ignorable rerun-validator))
  (if *glut-dll*
      (progn
        (print (list "yes on *glut-dll* loaded" *glut-dll*))
        (print (list "glut state 2" (glutget (coerce GLUT_INIT_STATE 'integer))))
        #+not (when (not (zerop (glutget (coerce GLUT_INIT_STATE 'integer))))
          (when (and *sys* (kids *sys*))
            (error "windows from prior test remain extant. if not visible, evaluate: (setf (kids *sys*) nil)"))
          (return-from ogb-glut-init)))
    (progn
      (uffi:load-foreign-library "c:/windows/system32/opengl32.dll"
        :force-load t
        :module "OPENGL32")
      (uffi:load-foreign-library "c:/windows/system32/glu32.dll"
        :FORCE-LOAD t
        :module "GLU32")
      (setq *glut-dll* (uffi:load-foreign-library ogb::*glut-lib-path*
                         :force-load nil
                         :module "glut"))))

  (print (list  "glut state 1" (glutget (coerce GLUT_INIT_STATE 'integer))))
  (print (list  "about to initialize"))

  (let ((argc (uffi:allocate-foreign-object :int 1)))
    (setf (uffi:deref-array argc '(:array :int) 0) 0)
    (unwind-protect
        (progn
          (ogb::glut-Init argc (uffi:make-null-pointer '(:array :cstring)))
          (print (list  "glut initialised"))
          
          )
      (uffi::free-foreign-object argc)))
  )


(defvar *mdepth*)
(defvar *selecting*)

(defun closed-stream-p (w)
  (declare (ignore w))
  (let ((w (glutGetWindow)))
    (or (not (zerop (glGetError)))
      (zerop w))))

(defun mstack (tag)
  (let ((mm (uffi:allocate-foreign-object :int 1))
        (sd (uffi:allocate-foreign-object :int 1)))
    (glgetIntegerv gl_matrix_mode mm)
    (let ((mmi (uffi:deref-array mm '(:array :int) 0)))
      (cond
       ((eql mmi gl_modelview)
        (glgetIntegerv gl_modelview_stack_depth sd)
        (print (list "modelview stack depth" tag (uffi:deref-array sd '(:array :int) 0))))

       ((eql mmi gl_projection)
        (glgetIntegerv gl_projection_stack_depth sd)
        (print (list "projection stack depth" tag (uffi:deref-array sd '(:array :int) 0))))

       ((eql mmi gl_texture)
        (glgetIntegerv gl_texture_stack_depth sd)
        (print (list "texture stack depth" tag (uffi:deref-array sd '(:array :int) 0))))

       (t (print (list "mstack> unexpected matrix mode" tag mm)))))
    (uffi::free-foreign-object mm)
    (uffi::free-foreign-object sd)))



(defun gl-stack-depth (tag)
  (let ((sd (uffi:allocate-foreign-object :int 1)))
    (unwind-protect
        (ecase (cello-matrix-mode tag)
          (:modelview
           (glgetIntegerv gl_modelview_stack_depth sd)
           (uffi:deref-array sd '(:array :int) 0))
      
          (:projection
           (glgetIntegerv gl_projection_stack_depth sd)
           (uffi:deref-array sd '(:array :int) 0))
           
          (:texture
           (glgetIntegerv gl_texture_stack_depth sd)
           (uffi:deref-array sd '(:array :int) 0)))
      (uffi::free-foreign-object sd))))

(defun cello-matrix-mode (tag)
  (let ((mm (uffi:allocate-foreign-object :int 1))
        )
    (glgetIntegerv gl_matrix_mode mm)
    (let ((mmi (uffi:deref-array mm '(:array :int) 0)))
      (unwind-protect
          (cond
           ((eql mmi gl_modelview) :MODELVIEW)
           ((eql mmi gl_projection) :PROJECTION)
           ((eql mmi gl_texture) :TEXTURE)
           
           (t (break "gl-stack-depth> unexpected matrix mode ~a ~a" tag mmi)))
        (uffi::free-foreign-object mm)))))

(defun gl-pushm ()
  ;;(mstack :gl-pushm-entry)
  (gl-Push-Matrix))

(defun gl-popm ()
  ;;(c-assert (> (gl-stack-depth :gl-popm) 1))
  ;;(mstack :gl-popm-entry)
  (gl-pop-Matrix))

(defun glut-stroke-string (font string)
  (dotimes (n (length string))
    ;;(print `(stroke ,n ,(elt string n)))
    (glut-stroke-character font (coerce (char-code (elt string n)) 'integer))
    ))

(defun glut-bitmap-string (font string)
  (loop for c across string
        do (glut-bitmap-character font (char-code c))))

(defmacro callback-set (setter settee)
  `(When ,settee
     (,setter (ff-register-callable ,settee))))

(defun glut-callback-set (&key close
                           display
                           idle
                           keyboard
                           reshape
                           entry
                           visibility
                           timer-info
                           passive-motion
                           motion
                           mouse
                           menu-state
                           special
                           button-box
                           dials
                           menu-status
                           overlay-display
                           window-status
                           keyboard-up
                           special-up
                           joystick)
  (progn
    (callback-set glut-close-Func close)
    (callback-set glut-display-Func display)
    (callback-set glut-idle-Func idle)
    
    (callback-set glut-keyboard-Func keyboard)
    (callback-set glut-reshape-func reshape)
    (callback-set glut-entry-func entry)
    (callback-set glut-visibility-func visibility)
    (callback-set glut-passive-motion-func passive-motion)
    (callback-set glut-motion-func motion)
    (callback-set glut-mouse-func mouse)
    (callback-set glut-menu-state-func menu-state)
    (callback-set glut-special-func special)
    (callback-set glut-button-box-func button-box)
    (callback-set glut-dials-func dials)
    (callback-set glut-menu-status-func menu-status)
    (callback-set glut-overlay-display-func overlay-display)
    (callback-set glut-window-status-func window-status)
    (callback-set glut-keyboard-up-func keyboard-up)
    (callback-set glut-special-up-func special-up)
    (callback-set glut-joystick-func joystick)
    )
  (when timer-info
    (apply #'glut-timer-func timer-info)))