;; -*- mode: Lisp; Syntax: Common-Lisp; Package: ogb; -*-
;;;
;;; Copyright  1995,2003 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package #:opengl-bindings)

#| GLUT window sub-API. |#




#| Font Weights |#
(dfc FW_DONTCARE         0)
(dfc FW_THIN             100)
(dfc FW_EXTRALIGHT       200)
(dfc FW_LIGHT            300)
(dfc FW_NORMAL           400)
(dfc FW_MEDIUM           500)
(dfc FW_SEMIBOLD         600)
(dfc FW_BOLD             700)
(dfc FW_EXTRABOLD        800)
(dfc FW_HEAVY            900)

(dfc FW_ULTRALIGHT       FW_EXTRALIGHT)
(dfc FW_REGULAR          FW_NORMAL)
(dfc FW_DEMIBOLD         FW_SEMIBOLD)
(dfc FW_ULTRABOLD        FW_EXTRABOLD)
(dfc FW_BLACK            FW_HEAVY)


(dfc WGL_FONT_LINES      0)
(dfc WGL_FONT_POLYGONS   1)

(dfc GLUT_ACTION_EXIT                         0)
(dfc GLUT_ACTION_GLUTMAINLOOP_RETURNS         1)
(dfc GLUT_ACTION_CONTINUE_EXECUTION           2)
(dfc GLUT_ACTION_ON_WINDOW_CLOSE        #x01F9)

(defun-ffx :void "glut" "glutSetOption" (GLenum eWhat :int value))
(defun-ffx :void "glut" "glutWCurrencyAssert" ())
(defun-ffx :void "glut" "glutWCurrencySet" ())
(defun-ffx :void "glut" "glutWFill" (:float r :float g :float b :float alpha))
(defun-ffx :void "glut" "glutWFill2" (:float r :float g :float b :float alpha))
(defun-ffx :void "glut" "glutWClearColor" (:float r :float g :float b :float alpha))
(defun-ffx :void "glut" "glutWClear" ())

(defun-ffx :int "glut" "glutCreateWindow" (:cstring title))
(defun-ffx :int "glut" "glutCreateSubWindow" (:int win :int x :int y :int width :int height))
(defun-ffx :void "glut" "glutDestroyWindow" (:int win))
(defun-ffx :void "glut" "fgDeinitialize" ())

(ff-defun-callable :cdecl :void mgwclose ()
  (print "closing callback entered"))

(defpackage #:opengl-bindings
  (:nicknames #:ogb)
  (:use)
  (:export mgwclose "NATE" "FREEG" "GLUT-BITMAP-STRING"))

(defun nate ())
(defun freeg () t)

(defun-ffx :void "glut" "glutPostRedisplay" ())
(defun-ffx :void "glut" "glutPostWindowRedisplay" (:int win))
(defun-ffx :void "glut" "glutSwapBuffers" ())
(defun-ffx :int "glut" "glutGetWindow" ())
(defun-ffx :int "glut" "glutDestroyPending" ())
(defun-ffx :void "glut" "glutSetWindow" (:int win))
(defun-ffx :void "glut" "glutSetWindowTitle" (:cstring title))
(defun-ffx :void "glut" "glutSetIconTitle" (:cstring title))
(defun-ffx :void "glut" "glutPositionWindow" (:int x :int y))
(defun-ffx :void "glut" "glutReshapeWindow" (:int width :int height))
(defun-ffx :void "glut" "glutPopWindow" ())
(defun-ffx :void "glut" "glutPushWindow" ())
(defun-ffx :void "glut" "glutIconifyWindow" ())
(defun-ffx :void "glut" "glutShowWindow" ())
(defun-ffx :void "glut" "glutHideWindow" ())
(defun-ffx :void "glut" "glutFullScreen" ())
(defun-ffx :void "glut" "glutSetCursor" (:int cursor))
(defun-ffx :void "glut" "glutWarpPointer" (:int x :int y))

;;;(defun-ffx :void "glut" "glutInit" (integer argc integer argv)) no dice

#-cormanlisp
(ff-def-call ("glut" glut-init "glutInit")
                      ((argc (* :int))
                       (argv (* :void))))

#+cormanlisp
(ff-def-call ("glut" glut-init "glutInit")
                      ((argc (:int *))
                       (argv (:void *))))

(defun-ffx :void "glut" "glutInitDisplayMode" (GLuint mode)) ;; .h said "unsigned int"

(defun-ffx :void "glut" "glutInitWindowPosition" (GLint x GLint y)) ;; .h said ints
(defun-ffx :void "glut" "glutInitWindowSize" (GLint width GLint height)) ;; .h said ints
(defun-ffx :void "glut" "glutInitDisplayString" (:cstring string))
(defun-ffx :void "glut" "glutLeaveMainLoop" ())
(defun-ffx :void "glut" "glutMainLoop" ())
(defun-ffx :void "glut" "glutCheckLoop" ())
(defun-ffx :void "glut" "glutMainLoopEvent" ())


#| GLUT window callback sub-API. |#

(defun-ffx :void "glut" "glutCloseFunc" (:long lpfn)) ;; (*func)(void)
(defun-ffx :void "glut" "glutWMCloseFunc" (:long lpfn)) ;; (*func)(void) ;; same
(defun-ffx :void "glut" "glutDisplayFunc" (:long lpfn)) ;; (*func)(void)
(defun-ffx :void "glut" "glutReshapeFunc" (:long lpfn)) ;; (*func)(int width (), int height));
(defun-ffx :void "glut" "glutKeyboardFunc" (:long lpfn)) ;; (*func)(unsigned char key, int x, int y));
(defun-ffx :void "glut" "glutMouseFunc" (:long lpfn)) ;; (*func)(int button, int state, int x, int y));
(defun-ffx :void "glut" "glutMotionFunc" (:long lpfn)) ;; (*func)(int x, int y));
(defun-ffx :void "glut" "glutPassiveMotionFunc" (:long lpfn)) ;; (*func)(int x, int y));
(defun-ffx :void "glut" "glutEntryFunc" (:long lpfn)) ;; (*func)(int state));
(defun-ffx :void "glut" "glutVisibilityFunc" (:long lpfn)) ;; (*func)(int state));
(defun-ffx :void "glut" "glutIdleFunc" (:long lpfn)) ;; (*func)(void));
(defun-ffx :void "glut" "glutTimerFunc" (GLuint millis :long lpfn)) ;; (int value), int value);
(defun-ffx :void "glut" "glutMenuStateFunc" (:long lpfn)) ;; (*func)(int state));

(defun-ffx :void "glut" "glutSpecialFunc" (:long lpfn)) ;; (*func)(int key, int x, int y));
(defun-ffx :void "glut" "glutSpaceballMotionFunc" (:long lpfn)) ;; (*func)(int x, int y, int z));
(defun-ffx :void "glut" "glutSpaceballRotateFunc" (:long lpfn)) ;; (*func)(int x, int y, int z));
(defun-ffx :void "glut" "glutSpaceballButtonFunc" (:long lpfn)) ;; (*func)(int button, int state));
(defun-ffx :void "glut" "glutButtonBoxFunc" (:long lpfn)) ;; (*func)(int button, int state));
(defun-ffx :void "glut" "glutDialsFunc" (:long lpfn)) ;; (*func)(int dial, int value));
(defun-ffx :void "glut" "glutTabletMotionFunc" (:long lpfn)) ;; (*func)(int x, int y));
(defun-ffx :void "glut" "glutTabletButtonFunc" (:long lpfn)) ;; (*func)(int button, int state, int x, int y));

(defun-ffx :void "glut" "glutMenuStatusFunc" (:long lpfn)) ;; (*func)(int status, int x, int y));
(defun-ffx :void "glut" "glutOverlayDisplayFunc" (:long lpfn)) ;; (*func)(void));

(defun-ffx :void "glut" "glutWindowStatusFunc" (:long lpfn)) ;; (*func)(int state));

(defun-ffx :void "glut" "glutKeyboardUpFunc" (:long lpfn)) ;; (*func)(unsigned char key, int x, int y));
(defun-ffx :void "glut" "glutSpecialUpFunc" (:long lpfn)) ;; (*func)(int key, int x, int y));
(defun-ffx :void "glut" "glutJoystickFunc" (:long lpfn)) ;; (*func)(unsigned int buttonMask, int x, int y, int z), int pollInterval);

#| GLUT color index sub-API. |#
(defun-ffx :void "glut" "glutSetColor" (:int cell GLfloat red GLfloat green GLfloat blue))
(defun-ffx GLfloat "glut" "glutGetColor" (:int ndx :int component))
(defun-ffx :void "glut" "glutCopyColormap" (:int win))

#| GLUT state retrieval sub-API. |#
(defun-ffx :int "glut" "glutGet" (GLenum eWhat))
(defun-ffx :int "glut" "glutDeviceGet" (GLenum type))

#| GLUT extension support sub-API |#
(defun-ffx :int "glut" "glutExtensionSupported" (:cstring name))
(defun-ffx :int "glut" "glutGetModifiers" ())
(defun-ffx :int "glut" "glutLayerGet" (GLenum type))

(defun-ffx :void "glut" "glutBitmapCharacter" (:void *font :int character))

(defun-ffx :int "glut" "glutBitmapWidth" (:void *font :int character))
(defun-ffx :int "glut" "glutBitmapHeight" (:void *font))
(defun-ffx GLfloat "glut" "glutBitmapXOrig" (:void *font))
(defun-ffx GLfloat "glut" "glutBitmapYOrig" (:void *font))

(defun-ffx :void "glut" "glutStrokeCharacter" (:void *font :int character))
(defun-ffx GLfloat "glut" "glutStrokeDescent" (:void *font))

#+Test
(list
 (glut-bitmap-height GLUT_BITMAP_TIMES_ROMAN_24)
 (glut-bitmap-width GLUT_BITMAP_TIMES_ROMAN_24 (char-code #\H)))

(defun-ffx :int "glut" "glutStrokeWidth" (:void *font :int character))
(defun-ffx GLfloat "glut" "glutStrokeHeight" (:void *font))

#+test
(list
 (glut-stroke-height glut_stroke_mono_roman)
 (glut-stroke-width GLUT_stroke_ROMAN (char-code #\H)))

(defun-ffx :int "glut" "glutBitmapLength" (:void *font :cstring string))
(defun-ffx :int "glut" "glutStrokeLength" (:void *font :cstring string))

#| GLUT pre-built models sub-API |#

(defun-ffx :void "glut" "glutWireSphere" (GLdouble radius GLint slices GLint stacks))
(defun-ffx :void "glut" "glutSolidSphere" (GLdouble radius GLint slices GLint stacks))
(defun-ffx :void "glut" "glutWireCone" (GLdouble base GLdouble height GLint slices GLint stacks))
(defun-ffx :void "glut" "glutSolidCone" (GLdouble base GLdouble height GLint slices GLint stacks))
(defun-ffx :void "glut" "glutWireCube" (GLdouble size))
(defun-ffx :void "glut" "glutSolidCube" (GLdouble size))
(defun-ffx :void "glut" "glutWireTorus" (GLdouble innerRadius GLdouble outerRadius GLint sides GLint rings))
(defun-ffx :void "glut" "glutSolidTorus" (GLdouble innerRadius GLdouble outerRadius GLint sides GLint rings))
(defun-ffx :void "glut" "glutWireDodecahedron" ())
(defun-ffx :void "glut" "glutSolidDodecahedron" ())
(defun-ffx :void "glut" "glutWireTeapot" (GLdouble size))
(defun-ffx :void "glut" "glutSolidTeapot" (GLdouble size))
(defun-ffx :void "glut" "glutWireOctahedron" ())
(defun-ffx :void "glut" "glutSolidOctahedron" ())
(defun-ffx :void "glut" "glutWireTetrahedron" ())
(defun-ffx :void "glut" "glutSolidTetrahedron" ())
(defun-ffx :void "glut" "glutWireIcosahedron" ())
(defun-ffx :void "glut" "glutSolidIcosahedron" ())
